! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module closserrbars_mod
!  VNY 02.11.2004 : a function for rigorous computation
!  of the error bars for the coincidence-loss corrected count rates
!  is added:
!
!  cLossErrBars(frameTime, timeDel, observedCountRate, verbFlag, deadFrac)
!
! The function returnes the rms-error estimated with the use of the rigourous algorithm
! for error-bars (S.Rosen, 2004). 
!
! Input parameters:
! ----------------
! frameTime - real(kind=double): frame time in seconds
! timeDel - real: time of the observation, in seconds (it could be a bin-time or the e
! xposure time)
! deadFrac - real(kind=double): dead fraction in units of the fraction of the frame time
! observedCountRate - real : observed count rate (which is affecetd by the coincidence losses)
!
! Output parameters:
!-------------------
! trueCountRate - real: the estimated count rate (corrected for the coincidence losses); 
! this value can be used for checking purporses only
!
!
   use dal
   use cal
   use param
   use errorHandling
   use frameTime_mod
   use datafilter
   use sasdal

   implicit none

   private calc_ln_fact, initRandomSequence

   public correctedCountRate
   public coincLossAsymErr
   public cLossErrBars

contains


!----------------------------------------------------------------------------
! Using Simon's algorithm for CL-correction and estimation of rms
!----------------------------------------------------------------------------
! The function returnes the rms-error for the estimated CL-corrected counts 
! 
!
! Input parameters:
! ----------------
! frameTime - real(kind=double): frame time in seconds
! timeDel - real: time of the observation, in seconds (it could be a bin-time or 
! the exposure time)
! deadFrac - real(kind=double): dead fraction in units of the fraction of the frame time
! observedCountRate - real : observed count rate (which is affecetd by the 
! coincidence losses)
!
! Output parameters:
!-------------------
! trueCountRate - real: the estimated count rate (corrected for the coincidence losses); 
! this value can be used for checking purporses only
!
!
real function cLossErrBars(frameTime, timeDel, & 
     observedCountRate, trueCountRate, deadFrac)
  ! frameTime is in seconds
  real(kind=double), intent(in) :: frameTime
  real, intent(in) :: timeDel
  real, intent(out):: trueCountRate ! counts per second
  real, intent(in) :: observedCountRate
  real(kind=double) :: deadFrac
  real :: dispersionLeft
  real :: dispersionRight
  real :: rmsValue, rmsValue1, rmsValue2
  !integer :: nFrames
  integer :: verbFlag2

!*****************************************************************
  real(kind=double) :: maxCountsPerFrame, sourceErr, corrRateErrMinus, corrRateErrPlus
  integer :: maxCounts, i, minCounts
  real :: sim_frameTime, sim_timeBin
  !real :: lossCountRate, lossCountRate1, lossCountRate2
  real :: lossCountRate1, lossCountRate2
  !real :: lossCountsArray2(2000), dispLeftArray(2000), dispRightArray(2000), rmsArray(2000)
  real(kind=double) :: source, corrRate
  logical :: extend
  real :: deltaLoss

      if (frameTime .gt. 0.0) then
         maxCountsPerFrame=0.99/frameTime
      else
         maxCountsPerFrame=200.0
      endif
      maxCounts=6*maxCountsPerFrame
      if (maxCounts .gt. 2000) maxCounts=2000

      sourceErr=0.0
      corrRateErrMinus=0.0
      corrRateErrPlus=0.0
      extend=.false.

         source=observedCountRate
         sourceErr=sqrt(source)
         corrRate = CorrectedCountRate(dble(source), sourceErr, &
              frameTime, deadFrac, extend, corrRateErrMinus, &
              corrRateErrPlus)
      sim_timeBin=timeDel
      sim_frameTime=frameTime

! Check the values of corrRate from corrRate-10% to maxCounts
!------------------------------------------------------------
      minCounts=corrRate-0.1*corrRate
      do i=minCounts,maxCounts-1
         trueCountRate=real(i)
         verbFlag2=0
         call coincLossAsymErr(sim_frameTime, sim_timeBin, trueCountRate, &
              lossCountRate1, dispersionLeft, &
              dispersionRight, rmsValue1, verbFlag2)
         trueCountRate=real(i+1)
         call coincLossAsymErr(sim_frameTime, sim_timeBin, trueCountRate, &
              lossCountRate2, dispersionLeft, &
              dispersionRight, rmsValue2, verbFlag2)
         if (lossCountRate1 .le. observedCountRate .and. &
              lossCountRate2 .gt. observedCountRate) exit  
      enddo
      deltaLoss=lossCountRate2-lossCountRate1
      if (abs(deltaLoss) .gt. 1.0e-5) then
         rmsValue=rmsValue1+(rmsValue2-rmsValue1)*(lossCountRAte2-observedCountRate)/deltaLoss
      else
         rmsValue=rmsValue1
      endif
      cLossErrBars=rmsValue
end function cLossErrBars


!*********************************************************
! Function CorrectedCountRate
! Corrects the raw count rate for coincidence losses, etc  
!*********************************************************
real(kind=double) function CorrectedCountRate(rawCountRate, & 
                  countRateError, frameTime, deadFrac, extended, &
                  countRateErrorMinus, countRateErrorPlus)
    real(kind=double), intent(in)    :: rawCountRate
    real(kind=double), intent(in) :: countRateError
    real(kind=double), intent(out)    :: countRateErrorMinus
    real(kind=double), intent(out)    :: countRateErrorPlus
 
    real(kind=double), intent(in) :: frameTime
    real(kind=double), intent(in) :: deadFrac
    logical, intent(in) :: extended
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: correctedRate

    rawRate = ValueAsymmetricErrorT(rawCountRate, dble(countRateError), dble(countRateError))
    correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, 0.0d0, extended)

    !****************************************************************
    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    !*****************************************************************
    if (CorrectedRate%value .lt. 1e10) then
        CorrectedCountRate = CorrectedRate%value
        countRateErrorMinus = correctedRate%errorMinus
        countRateErrorPlus = correctedRate%errorPlus
    else
       !***************************************************************
       !---VNY 01.11.2001 For the case of CAL_omPhotoNatural would fail
       !     the count rate are left without correction
       !***************************************************************
       if (rawCountRate .lt. 1e10) then
          CorrectedCountRate = rawCountRate
          countRateErrorMinus = countRateError
          countRateErrorPlus = countRateError
       else
          CorrectedCountRate = -1.0d0
          countRateErrorMinus = -1.0d0
          countRateErrorPlus = -1.0d0
       end if
    end if
end function CorrectedCountRate


!------------------------------------------------
! Simulate count rate
! 25.03.2004
!------------------------------------------------
!subroutine simulateCountRate(frameTime, timeBin, trueCountRate, &
!     lossCountRate, simulatedCountRate, dispersionLeft, dispersionRight)
!  real, intent(in) :: frameTime
!  real, intent(in) :: timeBin
!  real, intent(in) :: trueCountRate ! counts per second
!  real, intent(out) :: lossCountRate
!  real, intent(out) :: simulatedCountRate
!  real, intent(out) :: dispersionLeft
!  real, intent(out) :: dispersionRight!

!  real :: trueFluxPerFrame
!  integer :: nFrames
!  integer :: seedSize
!  integer, allocatable :: seedArray(:)
!  integer, allocatable :: trueCountsArray(:)
!  integer, allocatable :: lossCountsArray(:)
!  integer :: i, ival
!  logical first
!  real :: trueMean, lossMean, nMean
!  real :: nMeanLeft, nMeanRight
!  real :: diff
!  real :: trueRate, lossRate!

!  nFrames=int(timeBin/frameTime)
!  allocate(trueCountsArray(nFrames))
!  allocate(lossCountsArray(nFrames))


!!  write(*,*) " L2356 nFrames=", nFrames
!  ! VNY 26,03,2004
!  ! test Poissonian generator
!!  write(*,*) " Test Poissonian generator"
!!  write(*,*) "--------------------------"
!!  trueFluxPerFrame=100
!!  first=.true.
!!  ival=random_Poisson(trueFluxPerFrame, first)
!!  first=.false.
!!  trueCountsArray=0
!!  trueMean=0.0
!!  nMean=0.0
!!  do i=1, nFrames
!!     ival=random_Poisson(trueFluxPerFrame, first)
!!     trueCountsArray(i)=ival
!!     trueMean=trueMean+ival
!!     nMean=nMean+1.0
!!  enddo
!!  trueRate=trueMean/nMean
!!  dispersionLeft=0.0
!!  nMeanLeft=0.0
!!  do i=1, nFrames
!!     diff=trueCountsArray(i)-trueRate
!!     dispersionLeft=dispersionLeft+diff*diff
!!     nMeanLeft=nMeanLeft+1.0
!!  enddo
!!  dispersionLeft=sqrt(dispersionLeft/(nMeanLeft-1))
!!  write(*,*) " nFrames=", nFrames, " meanRate=", trueRate, " meanDisp=", dispersionLeft
!!  write(*,*) "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
!!----
!  trueCountsArray=0
!  lossCountsArray=0
!  trueMean=0.0
!  lossMean=0.0
!  nMean=0.0
!  trueFluxPerFrame=trueCountRate*frameTime
!  first=.true.
!  ival=random_Poisson(trueFluxPerFrame, first)
!  first=.false. 
!  do i=1, nFrames
!     ival=random_Poisson(trueFluxPerFrame, first)
 !    trueCountsArray(i)=ival
!     trueMean=trueMean+real(ival)
!     if (ival .gt. 0) then
!        lossCountsArray(i)=1
!        lossMean=lossMean+1.0
!     endif
!     nMean=nMean+1.0
!  enddo
!  trueRate=trueMean/timeBin
!  lossRate=lossMean/timeBin
!!  write(*,*) " L2387 trueRate=", trueRate, " lossRate=", lossRate 
!  if (nMean .gt. 0.0) then
!     trueMean=trueMean/nMean
!     lossMean=lossMean/nMean
!  endif
!!  write(*,*) " L2392 nMean=", nMean, " trueMean=", trueMean, &
!!       " lossMean=", lossMean
!!  simulatedCountRate=trueMean
!!  lossCountRate=lossMean
!  simulatedCountRate=trueRate
!  lossCountRate=lossRate
!  dispersionLeft=0.0
!  dispersionRight=0.0
!  nMeanLeft=0.0
!  nMeanRight=0.0
!  ! compute dispersion (to the left and to the right from the maximum
!  do i=1, nFrames
!     diff=trueCountsArray(i)/frameTime-trueRate
!     !write(*,*) " i=", i, " true=", trueCountsArray(i)/frameTime, &
!     !     " loss=", lossCountsArray(i)/frameTime, " diff=", diff
!     if (diff .lt. 0.0) then
!        dispersionLeft=dispersionLeft+diff*diff
!        nMeanLeft=nMeanLeft+1.0
!     endif
!     if (diff .gt. 0.0) then
!        dispersionRight=dispersionRight+diff*diff
!        nMeanRight=nMeanRight+1.0
!     endif
!  enddo
!  if (nMeanLeft .gt. 0.0) then
!     dispersionLeft=sqrt(dispersionLeft/(nMeanLeft-1.0))
!  else
!     dispersionLeft=sqrt(dispersionLeft)
!  endif
!  if (nMeanRight .gt. 0.0) then
!     dispersionRight=sqrt(dispersionRight/(nMeanRight-1.0))
!  else
!     dispersionRight=sqrt(dispersionRight)
!  endif!
!
!  deallocate(trueCountsArray)
!  deallocate(lossCountsArray)!

!end subroutine simulateCountRate

!---------------------------------------
! Using Simon's coincidence loss test routine
! 27.05.2004
!---------------------------------------
subroutine coincLossAsymErr(frameTime, timeBin, trueCountRate, &
     lossCountRate, dispersionLeft, dispersionRight, rmsValue, verbFlag)
  real, intent(in) :: frameTime
  real, intent(in) :: timeBin
  real, intent(in) :: trueCountRate ! counts per second
  real, intent(out) :: lossCountRate
  real, intent(out) :: dispersionLeft
  real, intent(out) :: dispersionRight
  real, intent(out) :: rmsValue
  integer, intent(in) :: verbFlag
  integer :: nFrames

!*****************************************************************
!  character(len=50) fout

  !real(kind=double) :: pi, prob_m_k_mu, mu, mu_eff, ln_p, p, p_last, max_p
  real(kind=double) :: pi, mu, mu_eff, ln_p, p
  real(kind=double) :: ln_k_fact, ln_m_fact, ln_k_m_fact, sum_p
  !real(kind=double) :: p_poiss, ln_p_poiss, sum_p_poiss, mu_poiss
  real(kind=double) :: mu_poiss
  !real(kind=double) :: rms, mean, r2, errl, errh, devl, devh, cl_factor
  real(kind=double) :: rms, mean, r2, errl, errh, devl, devh
  !real(kind=double) :: cl_formula, mean_poiss, sigma_poiss, sum_p_l, sum_p_h
  real(kind=double) :: cl_formula, sum_p_l, sum_p_h
  !real(kind=double) :: ln_fixed_term, mean1, r2_1, crfrr
  real(kind=double) :: ln_fixed_term, crfrr
  real(kind=double) :: ec(5), emp_corr, fco

  !integer(kind=int32) :: k, m, i, max
  integer(kind=int32) :: k, m, i

  real(kind=double) :: ci, t ! incident CR (cts/s) and frame rate (s)
  real(kind=double) :: co_best, diff, ci_pred_best
  !real(kind=double) :: ci_low, ci_hi, co_low, co_hi
 
  real(kind=double) :: co, ci_pred, muo
  !logical found_max, stop, use_empirical
  logical use_empirical

! VNY 14.06.2004 from Simon's code:
! insert empirical coincidence loss coefficients, from 
! ~/sasccf/OM_PHOTTONAT_0003.CCF. Coefficients work on the co*t (raw CRFR 
! ratio).
  ec(1)=1.0d0
  ec(2)= 0.0756730d0
  ec(3)=-0.1439860d0
  ec(4)= 0.6384970d0
  ec(5)=-0.5701840d0

  pi=4.d0*atan(1.d0)

  nFrames=int(timeBin/frameTime)

  ci=trueCountRate
  t=frameTime

  mu_poiss=ci ! incident counts during one second
  mu=ci*t ! incident counts during one frame
  crfrr=ci*t ! INCIDENT CRFR ratio
  cl_formula=(1.d0-exp(-mu))/frameTime !t
  k=nFrames ! total number of frames per exposure time (timeBin)


! There is a PROBLEM to use the empirical curve directly here in this 
! test code because the empirical correction uses a polynomial based on 
! the observed, not incident count rate so have to iteratively find what 
! observed CR gives incident count rate provided by user. Quick solution 
! here where we have ci, not co, initially, is to simply compute ci for 
! a range of co*T (=muo =CRFRobserved values) and find value of co which 
! gives Ci (empirical) closest to input value

      diff=1.e10
      ci_pred_best=-1.0

      do i=1,100000

         muo=(i-1)/100000.0
         co=muo/t

         emp_corr=ec(1) + ec(2)*muo + ec(2)*muo*muo + ec(3)*muo**3 + & 
          ec(4)*muo**4 + ec(5)*muo**5


         ci_pred=-log(1.d0-muo)/t
         ci_pred=ci_pred*emp_corr


         if(abs(ci_pred - ci).lt.diff)then
            diff=abs(ci_pred-ci)
            co_best=co
            ci_pred_best=ci_pred
         end if

      end do

! compute correction factor f(Co) that is thus used to convert this Co to Ci
! Will be used lower down if empirical correction is applied. 

      muo=co_best*t
      fco=ec(1) + ec(2)*muo + ec(2)*muo*muo + ec(3)*muo**3 + & 
        ec(4)*muo**4 + ec(5)*muo**5
      !if (verbFlag .eq. 1) then 
      !   write(*,'(/,"Based on empirical CL function, trials show that an &
      !        &observed rate ",/," of Co= ",f10.3," would be observed for an &
      !        &incident rate of Ci= ",f10.3,/," compared to the user specified &
      !        &rate ci(input) = ",f10.3)')co_best,ci_pred_best,ci
      !   if(ci_pred_best.lt.0.0)then
      !      write(*,*)' NO BEST EMPIRICAL SOLN FOUND'
      !      !call exit
      !   end if
      !endif

! First compute the probability distribution of the incident count rate, for 
! reference
!
!!  sum_p_poiss=0.d0
!!  mean_poiss=0.d0
!!  sigma_poiss=0.d0

!!  max=int(5*Ci)
!!  if (ci.lt.10) max=30
!!  do i=0,max
!!     m=i
!!     ln_m_fact=calc_ln_fact(m)
!!     ln_p_poiss = m*log(mu_poiss) - mu_poiss - ln_m_fact
!!     p_poiss = exp(ln_p_poiss)
!!     if(p_poiss.lt.1.d-30)p_poiss=0.d0
!!     sum_p_poiss = sum_p_poiss + p_poiss
!!     mean_poiss=mean_poiss+p_poiss*m
!!  end do
!!  if (verbFlag .eq. 1) write(*,*) " L2556 mean_poiss=", mean_poiss
!
! As a check, compute sigma using sqrt[sum(p*(xi-X)**2] rather than 
! sqrt[bar(r**2) - (bar(r))**2]
!
  devl=0.d0
  devh=0.d0
  sum_p_l=0.d0
  sum_p_h=0.d0

!!  do i=0,max
!!     m=i
!!     ln_m_fact=calc_ln_fact(m)
!!     ln_p_poiss = m*log(mu_poiss) - mu_poiss - ln_m_fact
!!     p_poiss = exp(ln_p_poiss)
!!     if(p_poiss.lt.1.d-30)p_poiss=0.d0
!!     sigma_poiss=sigma_poiss + p_poiss * (1.d0*m-mean_poiss)**2
!!     if(m.le.mean_poiss)then
!!        devl=devl + p_poiss*(m-mean_poiss)**2
!!        sum_p_l=sum_p_l+p_poiss
!!     end if
!!     if(m.gt.mean_poiss)then
!!        devh=devh + p_poiss*(m-mean_poiss)**2
!!        sum_p_h=sum_p_h+p_poiss
!!     end if
!!  end do

!      write(1,'(//)')

!  if (verbFlag .eq. 1) then
!     write(*,*) " For input Poisson distribution "
!     write(*,*) " ------------------------------ "
!     write(*,*) " L2587 mean_poiss=", mean_poiss, " sqrt(sigma_poiss)=", &
!          sqrt(sigma_poiss)
!     write(*,*) " Sum of P(n,mu)=", sum_p_poiss, " (should be 1.0)"
!     write(*,*) " devl=", devl, " devh=", devh, " sum_p_l=", sum_p_l, &
!          " sum_p_h=", sum_p_h
!     write(*,*) " sqrt(devl/sum_p_l)=", sqrt(devl/sum_p_l), &
!          " sqrt(devh/sum_p_h)=", sqrt(devh/sum_p_h)
!  endif

!  compute the probability of getting m=i counts if we run k frames with an 
!  incident rate of Ci counts/s and a frame time of t seconds. This 
!  probability is based on kCm * P(0)**(k-m) * P(>=1)**m 
!
!  where
!
!  kCm = k!/(m!)(k-m)!
!
!

!
!  switch for  using empirical correction 
!
  use_empirical=.true.

  sum_p=0.d0
  rms=0.d0
  r2=0.d0
  mean=0.d0

!----------- 15.06.2004
  mu_eff=mu
  if(use_empirical)then
     mu_eff=mu/fco   !compute effective incident mu
     cl_formula=co_best
  end if
  ln_fixed_term=log(1.d0-exp(-mu_eff)) !this term if fixed for each k below
!-----------

  ln_k_fact=calc_ln_fact(k)        ! constant for given no of frames

  do i=0,k  ! m can not exceed k
     m=i        
     ln_m_fact=calc_ln_fact(m)
     ln_k_m_fact=calc_ln_fact(k-m)
     ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact
!     ln_p = ln_p - (k-m)*mu + m*log(1.d0-exp(-mu))
     ln_p = ln_p - (k-m)*mu_eff + m*ln_fixed_term 
     p=exp(ln_p)
     if(p.lt.1.d-30)p=0.d0    ! set to zero if P is very small. 
     mean=mean+p*m
     r2=r2 + p*m*m
     sum_p=sum_p+p
  end do

!--- 15.06.2004
! rounding errors and Stirlings approx give rise to value of sum_p which is 
! very slightly > 1.000. Thus, normalise sum(r2) amd mean by sum_p as if to 
! make them equivalent to having come from a distribution with sum_p=1.0. 
! This appears to remove potential problem of sum(r2)-mean*mean being -ve. 

      r2=r2/sum_p         
      mean=mean/sum_p     

      rms=r2-mean*mean    ! rms (square of)
      !if (verbFlag .eq. 1) then
      !   if(use_empirical)then
      !      write(*,'(" For the coincidence loss affected (truncated) &
      !           &distribution (WITH EMPIRICAL CORRECTION INCLUDED)",/)')
      !   else
      !      write(*,'(" For the coincidence loss affected (truncated) &
      !           &distribution",/)')
      !   end if!
      ! 
      !   write(*,'("Sum of P(k,m,mu) (should be 1.0) &
      !        & = ",f12.8)')sum_p
      !   write(*,'("Mean of observed cts = ",f12.4," cf cl formula &
      !        &mean = ",f12.4)')mean, cl_formula*k*t
      !endif
!----

!!  rms=r2-mean*mean    ! rms (square of)
!!  if(rms.ge.0.d0)then
!!     rms=sqrt(rms)
!!  else
!!     if (verbFlag .eq. 1) write(*,*)'r2-mean*mean < 0 !'
!!  end if
!!  cl_factor=ci*frameTime*k/mean
!
! now compute asymmetric errors - For asymmetric errors need to know 
! what mean is first, so must compute above loop first and then repeat to 
! get deviations from mean, weighted by P
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  errl=0.d0
  errh=0.d0
  devl=0.d0
  devh=0.d0
  rms=0.d0
  do i=0,k     ! m can not exceed k
     m=i
     ln_m_fact=calc_ln_fact(m)
     ln_k_m_fact=calc_ln_fact(k-m)
     ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact
     ln_p = ln_p - (k-m)*mu + m*log(1.d0-exp(-mu))
     p=exp(ln_p)
     if (p.lt.1.d-30) p=0.d0
     rms=rms+p*(m-mean)**2
     if (m.le.mean) devl=devl + p*(m-mean)**2
     if (m.gt.mean) devh=devh + p*(m-mean)**2
  end do
  errl=sqrt(devl)
  errh=sqrt(devh)
  !if (verbFlag .eq. 1) write(*,*) "L2765  devl=", devl, "errl=", errl, " devh=", devh, " errh=", errh, " rms=", rms
  if (rms .gt. 0.d0) then
     rms=sqrt(rms)
  else
     if (verbFlag .eq. 1) write(*,*) " L2771 rms is negative"
  endif


  rmsValue=rms/timeBin
  dispersionLeft=errl/timeBin
  dispersionRight=errh/timeBin
  lossCountRate=mean/timeBin
end subroutine coincLossAsymErr

!-------------------------------------
! log of factorial
!-------------------------------------
real(kind=double) function calc_ln_fact(i)      
  integer(kind=int32) :: i, j
  real(kind=double) :: ln_i_fact, pi, i_fact
  pi=4.d0*atan(1.d0)    !3.141592654d0
  if(i.le.50)then
     j=1
     i_fact=1.d0      ! set for 0 factorial
     do while(j.le.i)
        i_fact=i_fact*j
        j=j+1
     end do
     ln_i_fact=log(i_fact)   ! convert to a log
  else   !use Stirling's approximation
     ln_i_fact=0.5d0*log(2.d0*pi*i) + 1.d0*i*(log(1.d0*i)-1.d0)
  end if
  calc_ln_fact=ln_i_fact
  return
end function calc_ln_fact
! End of Simon's code
!*****************************************************************

!--------------------------------------
! Initialise the random number generator
!--------------------------------------
subroutine initRandomSequence(initNumber)
  integer, intent(in) :: initNumber
  integer :: seedSize, i
  integer, allocatable :: seedArray(:)
  call random_seed(SIZE=seedSize)
  allocate(seedArray(seedSize))
  do i=1,seedSize
     seedArray(i)=i*initNumber
  enddo
  call random_seed(PUT=seedArray)
  deallocate(seedArray)
end subroutine initRandomSequence

end module closserrbars_mod

