! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: omlcbuild_mod
!
! Authors:  Tim Harries - MSSL, Vladimir Yershov - MSSL
!          
! Description: Constructs an OM OSW FITS source timeseries from source and background rates files produced by using evselect
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!
! History:
! 
!    15.05.2001 VNY -  timeStart parameter was corrected (TIMEZERO replaced with TSTART)
!    15.05.2001 VNY -  the output table title has been changed from RATES to RATE (necessary for lcplot programme) 
!    04.06.2001 VNY  - wdxSet input parameter is added (it is necessary 
!                      for count rates correction )
!    05.06.2001 VNY - sourceList input parameter is added (for count rates correction)
!    06.06.2001 VNY - makes correction for count losses
!    26.06.2001 VNY - CorrectedCountRate function has been changed
!    06.07.2001 VNY - Corrected the background estimation
!    08.08.2001 VNY - correction of the counts detected within the source radius is made
!                     now taking into account possible lost of photons at the edge of the OSW
!                     and using CAL_omGetPSFmap routine 
!    13.08.2001 VNY - Test with just wider PSF : ((radius-6)**2)/36
!    15.08.2001 VNY - the use of Tracking history file in order to compensate shifts of the source
!                     during the exposure (new parameter is used with the name of the TH-file: thxSet)
!    28.08.2001 VNY - various test with the Tracking History file correcting shifts within one exposure
!    29.09.2001 VNY - Correcting the source count rates when it is near the edge of the field of view
!    30.08.2001 VNY - Average magnitude is inserted to the header of the output data set file istead of
!                     the magnitude column 
!    05.09.2001 VNY - Averaging TH shifts instead of interpolation
!    06.09.2001 VNY - PSF_FRAC parameter has been added to the output table attributes (means the fraction
!                     of the image, which lies within the OSW, just to indicate whether or not the extracted
!                     light curve is affected by the proximity of the source to the edge of the field of view
!    19.09.2001 VNY - checking the background counts for extremely high values
!    20.09.2001 VNY - checkin the source counts for extremely high values
!    21.09.2001 VNY - median filtration of the output time series
!    24.09.2001 VNY - The use of the Tracking History file for the FOV-proximity correction has been removed.
!                     Respectively, the txhset input parameter has been removed as well.
!    18.10.2001 VNY - unused variables removed 
!    25.10.2001 VNY - initialisation of background rate errors array has been introduced (SPR-2611)
!    29.10.2001 VNY - initialisation of countRateErrorMinus and countrateErrorPlus parameters 
!                     has been introduced to the CorrectedCountRate function
!    30.10.2001     - rectified the calculation of rateSourceErrOutArray
!                     (rateBackOutArray was replaced with rateBAckErrOutArray)
!    31.10.2001 VNY - rectified the correction for the PSF fraction outside the OSW
!    01.11.2001 VNY - rectified count rates for the case if the COINCIDENCE table missed in OM_PHOTONAT_XXXX.CCF
!                     (to avoid the Illegal Count Rate messages)
!    13.11.2001 VNY - background is being removed from the source counts before applying the PSF fraction
!                     correction. Then it is restored in order to calculate correctly the 
!                     Coincidence Lost Correction. And after that, the background is 
!                     subtracted again, but at this time from the area defined by CAL_Aperture_Radius.
!                     (SPR-2629) 
!    19.11.2001 VNY - 3 additional keywords have been added to the binary table extension (SRRADIUS, CLRADIUS
!                     and PHRADIUS, the source extraction radius, the coincidence loss correction radius
!                     and the photometry radius). 
!                     Also the parameter VIGNET has been replaced with PSF_FRAC keyword, because vignetting
!                     is defined in OGIP as a geometrical parameter (fraction of the area), but the task
!                     calculates the fraction of the PSF, not of the area. Respectively, PSF_CAPP keyword
!                     has been added, which means the the correction for PSF_FRAC was applied
!   06.12.2001 VNY -  background counts scaling has been corrected in order to get uniform lightcurve
!                     level for different lightcurve extraction areas (fixing SSC-SPR-2668).
!                     New parameter (binBpe) has been added to getFrameTime routine, according to
!                     the changes made by CB in this routine
!   07.12.2001 VNY -  VIGNET parameter is set to 1.0 (no vignetting) in order to permit lcurve to build 
!                     properly the graph
!   10.12.2001 VNY -  improved background counts scaling in order to get uniform lightcurve level
!                     for different lightcurve extraction areas (fixing SSC-SPR-2668).
!                     New parameter (binBpe) has been added to the getFrameTime call,
!                     in accordance with recent changes made by CB in the getFrameTime routine.
!                     VIGNET keyword in the output file header is set to 1.0 (no
!                     vignetting) in order to permit the proper use of this file by LCURVE task.   
!   11.12.2001 VNY -  fine agjustment of the PSF fraction calculation has been made: avoiding circles 
!                     from the neighbour sources are taken into account in accordance with recent changes
!                     in the OMREGION task. 
!   12.12.2001 VNY -  the avoiding circle is variable (because its radius it taken from omdetect). This
!                     could cause a variability of the source area from exposure to exposure. This
!                     could noticably affect the output light curve.
!   17.12.2001 VNY -  TIME column has been added to the output light curve FITS file
!   07.01.2002 SRR -  changed TIMEZERO value to 0.d0 in order to get values of time in the TIME column
!                     referred directly to MJDREF.
!   17.01.2002 VNY -  protected against the absence of data in the background RATE file
!   23.01.2002 VNY -  allocation of timeBack and timeSource arrays has been changed to satisfy dlocate routine:
!                     in dlocate these arrays are supposed to be starting from 1, but a 0 value is 
!                     also read by it, that caused a subscript crash when the array has only one component
!   25.01.2002 VNY -  Scaling background counts to the source area has been improved. The source area 
!                     is not circular for the sources near the OSW edge. Now the real shape and size of this 
!                     area is taken into account before scaling and subtracting background from the 
!                     source counts (SSC-SPR-2715).   
!   05.02.2002 VNY -  Error section in the description document file has been updated (SSC-SPR-2730)
!
!   08.02.2002 VNY - Negative background counts are zeroed. These could appear as the result of 
!                    interpolation between input and output time bins when there are only a few background
!                    counts in the input data (SSC-SPR-2738). In this case the error bars for the zeroed counts 
!                    are deduced from the background median 
!
!  13.02.2002 VNY -  Background calculation has been improved: Fraction of PSF within the background extraction
!                    area has been taken into account. Background counts are corrected for this PSF fraction before
!                    subtracting background from the source counts. This could be essential for small source
!                    extraction radii (SSC-SPR-2745).
!
!  22.03.2002 VNY - Memory allocation/deallocation has been corrected (SPR-2814) 
!
!  26.03.2002 (VNY) DISTRIBUTION has been changed from ssc to world
!  27.03.2002 (VNY) Protected against 0-size of the input source rate table
!  28.03.2002 (VNY) Additional protection against 0-size of the input source rate table 
!  28.03.2002 (VNY) Corrected a bug in memory deallocation introduced on 22.03.2002
!  29.03.2002 (VNY) Default values of integers are forced to be int32-kind
!  30.03.2002 (VNY) Check of the input source rates table is introduced
!
!  18.04.2002 (VNY) Correction of scaling factors, which makes the task more robust in calculation
!                   of count rates for faint objects (SSC-SPR-2826). 
!                   
!  24.04.2002 (VNY)  Further improvement of the source/background scaling factors:
!                    the algorithm for estimation of the source extraction area
!                    is made consistent with that used in evselect (SSC-SPR-2826).
!	             The source counts error bars are protected against possible negative
!                    values of the errors appearing in the CAL_omPhotonatural function.
!
!  09.05.2002 (The Victory Day) A small shift to the psf-map is applied when calculating the
!                    scaling areas and coefficients. This makes the psf-map coinsiding with
!                    the CCD-grid, and the scaling factors are calculated more rogorousely.
!	             A bug, which restricted the timebinsize parameter, has been fixed
!                    (SSC-SPR-2864)  
!
! 22.05.2002 VNY -   RA and DEC keywords have been replaced with RA_OBJ and DEC_OBJ in order to give
!                    the user an easier possibility to run BARYCEN on the timeseries.  
! 12.06.2002 VNY -   TIME column is arranged in accordance with OGIP recommendations (time 
!                    value now corresponds to the centre of the time-bin interval)
! 13.06.2002 VNY -   Last time-bin interval has been scaled taking into account the 
!                    posiibility of that it is shorter than al other time-bins
! 17.09.2002 VNY -   Adjusted to couple with the modified SRCLIST-file. The omdetect task now doesn't
!                    produce the MAG-column. Thus, omlcbuild picks up this column only if exists.
! 17.09.2002 VNY -   Calculation of mean magnitude is made after averaging the count rates.   
!        
! 25.09.2002 VNY -   Source list table is arraged to be similar to the OMICHAIN. The comments to the 
!                    column names in the output RATES table have been added. The mean magnitude goes 
!                    to the SWSLI and TIMESR files
!
! 08.10.2002 VNY -   Check of the existence of CORR_BACKGROUND_RATE columnd introduced to cope with 
!                    possible problems with this column (SSC-SPR-2995)
!
! 25.10.2002 VNY -   Corrected a bug, which was introduced at the end of September and which 
!                    resulted in the impossibility to find the CORR_RATE_ERR column in the source 
!                    list file (SSC-SPR-3004)
!
! 29.10.2002 VNY -   Protection against the absence of the COINCLOSS_CORR column is introduced (SPR-3007)
!
! 07.04.2003 VNY - Inspecting the code: it seems that the last time-bin is assigned a larger 
!                  error-bar
! 08.04.2003 VNY - The increased error-bar of the last time bin is correct: this happens just 
!                  because the photons for this last time bin were collected during a shorter 
!                  period of time, and the accuracy of measuring the count-rate for this 
!                  time-bin is less than for the rest of the data.
!                  Further inspecting the code: it seems that the source and background 
!                  light-curves are exagerratedly anticorrelated. This is probably caused by the 
!                  used algorithm of the background estimation with the system of two equations. 
! 08.04.2003 VNY   A small bug has been found in the interpolation of the last two timebins
!                  of the background timeseries. An SPR (no.3103) raised.
!                  The bug is corrected.
! 16.06.2003 VNY   A small bug has been found (and fixed), which was causing a subscript error
!                  in the array of the background timeseries (SSC-SPR-3131). Apparently, the 
!                  bug was introduced while correcting the previous problem. 
! 02.03.2004 VNY   The programme is protected against too high counts that cannot be corrected
!                  for the coincidence-loss. When such high count-rate occure, the CAL_omPhotoNatiral
!                  produces a warnig message and does not corrects the counts.
!                  To avoid possible sudden jumps in the count rates for the bright sources
!                  the high rates are set to their maximal corrigeable value (SSC-SPR-3259)        
! 25.03.2004 VNY   Test with simulations of the coincidence losses
! 15.06.2004 VNY   Introduction of the rigourous estimation of the error-bars (S.Rosen's algorithm)
!                  (fixing SSC-SPR-3337)
! 31.01.2006 VNY   Introducing the time-dependent sensitivity correction
!
! 10.02.2006 VNY   Fixed SPR-3580 (time-dependent sensitivity correction)
! 16.02.2006 VNY   Putting the coeeficient for the time-dependent sensitivity correction
!                  to the header of the timeseries file
! 17.02.2006 VNY   Found a small bug in the assignment of the input table columns 
!                  (mag was assigned to rates but only in the case when the rates column did not
!                  exist in the input file; so, this bug was unlikely to reveal itself in the 
!                  normal conditions) - the bug is fixed
! 21.02.2006 VNY   The input source-list file is modified to make it adjusted to the values 
!                  computed here (source and background count rates, pixel areas, significance) 
! 06.03.2007 VNY   Introducing the check of the QFLAG column type in the source list
!                  table (which could be int8 or int16, depending on the version of omdetect).
!                  Automatically switching in between these two data types.
! 07.06.2007 VNY   Suppressing the background subtraction for mod-8 corrupted images (to fix
!                  SSC-SPR-3144) because the bright corrupted sources occupy practically
!                  the whole Fast window.
! 13.02.2008 VNY   Introducing the usage of the CAL routine for computing the time-dependent 
!                  sensitivity correction factor (SSC-SPR-3706)
! 25.04.2008 VNY   The use of the theoretical calibration curve for coincidence-loss correction 
!                  is replaced by empirical polynomial coefficients (i.e. the CAL routine 
!                  CAL_omPhotoNatural is now called with the logical parameter "empirical" set to 
!                  true instead of false). Currently, the difference between the theoretial 
!                  calibration curve and the polynomial function is small but in the future the 
!                  calibration files might change (SSC-SPR-6456).
! 05.06.2008 VNY   Moifying the raw count rates for the source and background in the source list
!                  table produced by omdetect (these are not consistent with the count rates
!                  computed by omlcbuild).
! 11.03.2009 VNY   Introducing the variability tests
!
! 30.04.2009 VNY   Fixing a bug related to an unallocated pointer (SPR-6525)
! 14.05.2009 VNY   Removing the debugging test messages left from the previous upload (SPR-6530) 
! 18.05.2009 VNY   Issuing a warning message in the case of finding an unrealistically large
!                  number of sources                 
! 28.05.2009 VNY   Restoring the limit of 2000 sources for profucing light curves
!                  (since this limit avoids craches due to lack of the computer memory)
!                  SSC-SPR-6538
! 08.09.2009 VNY   random_number module is removed as obsolete
! 07.10.2009 VNY   test of omPhotoNatural
! 10.10.2009 VNY   Introducing new parameter mod8corrupted to indicate that the target source 
!                  is too bright and affected by the modulo-8 noise
! 27.11.2009 VNY   Introduce the parameter imageset to calculate the background level
!                  by using the imaging data;
!
!                  Introduced the parameter "dtcorrection" in order to correct the overcorrected
!                  coincidence-losses for bright sources (which are happen to be overcorrected
!                  when the source count rates exceed 60 counts per second)
!
!                  Introduced the parameter "bkgfromimage=yes/no" to force the task to use the
!                  imaging-data-based background
! 26.01.2010 VNY   Removing the parameter "dtcorrection" as not yet approved by testing
!                  on a sufficiently large variety of data sets
! 31.01.2010 VNY   Protected against the crash because of the possibility of missing block OMIMI1 in the
!                  accompanying imaging data file.
! 20.02.2010 VNY   Fixed a bug related to the wrong calculation of the background level when using
!                  the option bkgfromimage=yes   
! 16.03.2010 VNY   Corrected a subtle mistake related to a 0.5-pixel
!                  miscalculation of the source PSF position with respect to the edge of the
!                  Fast window (SSC-SPR-6602). This mistake was causing about 10% variability 
!                  of the source count rates when the source was touching the edge 
!                  of the window, its position varying with respect to this 
!                  edge by about 1 or 2 pixels. 	
!
! 18-02-2011 VNY  For some reason the above-mentioned 0.5-pixel correction of the PSF position (SSC-SPR-6602)
!                 was commented (not working). So the correction has been restored
! 19-02-2011 VNY  The 0.5-pixel correction is modified once again as the previous modification was not complete.
! 12-11-2012 VNY  Protected against possible cycling in the subroutine sort 
!                 if its input contains NaN values
! 01-06-2014 VNY  fixing the problem reported in SPR-0007207: measuring the background from the imaging window
!                 fails if the image is binned; also a small bug resulting in smaller output count rates
!                 for the last time bin is fixed (the size of the time bin was almost twice of what it
!                 should be)

module omlcbuild_mod

  use dal
  use cal
  use omskybackground_mod
  use ominterpolate_mod
  use omsmoothimage_mod
  use OmSourceParam_mod        ! Source definitions, etc
  use OmPhotometry_mod         ! Does aperture photometry
  use OmConstants_mod
  use param  
  use errorHandling
  use frameTime_mod
  use datafilter
  use sasdal
!  use utility_mod
!  use random_mod
  use ommagnitude_mod
  use sla
  !use allocation_mod 

  implicit none 

  type test
     real(kind=single) :: Stat , Prob
  end type test

  type data_tests
     real(kind=single) :: Mean_Count , Variance, Max_Deviation
     type(test) :: Chi_Square , K_S
  end type data_tests

  !****************************************
  ! A structure to store pixel information
  !****************************************
  Type PixelStruct
     integer(int16) :: x = 0
     integer(int16) :: y = 0
     integer(int16) :: number = 0
     real(single) :: counts = 0.0
     real(single) :: backgroundCounts = 0.0
     integer(int16) :: noNeighbours = 0
  end Type PixelStruct
  
  private 

  public omlcbuild



contains 
  
  !---------------------------------------------------------------------
  subroutine omlcbuild
  !---------------------------------------------------------------------

   ! variables
   ! a temp DAL dataset to find area of regions
   type(DataSetT) :: tempSet
   type(TableT) :: tempTable
   type(ColumnT) :: tempColX, tempColY

   integer(Kind=int32), pointer :: xArray(:), yArray(:)
   integer(kind=int32) :: npts, nPix
   integer(kind=int32) :: nx, ny
   integer(kind=int32) :: i,j,k

   real(kind=double) :: areaIn, areaOut, areaIn2, areaOut2, areaIn3, areaOut3
   real(kind=double) :: areaBkg
   real(kind=single) :: bkgExtractionArea, srcExtractionArea
   logical :: outOfAvoid
   real(kind=single) :: dist
   integer(kind=int32) :: ia
   integer(kind=int32) :: nBackAvoid
   real (kind=single), dimension(10) :: xBackAvoid
   real (kind=single), dimension(10) :: yBackAvoid
   real (kind=single), dimension(10) :: rBackAvoid

   ! region and lc file filenames
   character(len=1024) :: sourceReg, backReg, expression
   character(len=1024) :: backRates, sourceRates, sourceList, wdxSet

   ! region file variables
   real(kind=single) :: sourceArea, backArea, detectedArea
   type(DataSetT) :: sourceSet, backSet, sourceRegSet, backRegSet
   !type(DataSetT) :: imagingDataSet
   character(len=1024) :: imagingDataSet
   type(TableT) :: sourceTable, backTable
   type(TableT) :: regionTable
   type(ColumnT) :: sourceRadiusCol, sourceXPosCol, sourceYPosCol, sourceShapeCol
   type(TableT) :: backRegionTable
   type(ColumnT) :: backRadiusCol, backXPosCol, backYPosCol, backShapeCol

   ! Priority Window Data dataset handle
   type(DataSetT) :: wdxDataSetH
   ! Input table handle
   type(TableT) :: inTableH
   !---VNY 08.10.02 noColumns changed from 16 to 17
   !---VNY 20.02.2006 noColumns changed from 20 to 23
   integer, parameter :: noColumns = 28! 23
   type(ColumnT), dimension(noColumns) :: inColumnH
   integer(kind=int8), dimension(noColumns) :: inColumnFlag

   ! Input dataset handle
   type(DataSetT) :: inDataSetH
   real(kind=double)              :: detectionLimitCountRate         ! The detection limit count rate (counts/second)
   real(kind=double)              :: detectionLimitMagnitude         ! The detection limit magnitude
   real(kind=double) :: error = 0.0 
   real(kind=single) :: calApertureRadius = 12.0          ! Radius used for coincidence losses   
   real(kind=single) :: photometryApertureRadius          ! Radius used for converting counts to magnitudes
 
   character(len=8) :: dataMode                                      ! Input file data mode.
   character(len=9) :: filter                                        ! Keywords.
   real(kind=single) :: exposure
   real(kind=double) :: frameTime                                    ! OM CCD frame time.
   real(kind=double) :: deadFrac                                     ! Dead time in units of CCD frame time.
   real(kind=single) :: deadFracCorrection
   integer(kind=int32) ccdMode                                       ! OM CCD data mode.
   integer(kind=int32) :: filterIdInt

   logical :: scaleCounts = .false.
   logical :: vignApp
   logical :: backApp
   logical :: deadApp
   logical :: psfApp
   logical :: bkgRateExists   ! Indicates that background rate table exists in the file
   logical :: bkgRateId      ! Indicates that background and source rates are of the same EXP_ID
   logical :: srcRateExists ! Indicate that source rate table does exist and is not of zero size

   real(kind=single) :: countsRatio = 1.0   
   real(kind=single) :: vignet =1.0 ! Vignetting parameter
   real(kind=single) :: psfFrac =0.0
   real(kind=double) :: corrRate                                     ! Corrected count rate.
   real(kind=double) :: corrRateErrPlus                              ! Corrected rate error + (asymmetric).
   real(kind=double) :: corrRateErrMinus                             ! Corrected rate error - (asymmetric).
   real(kind=double) :: corrRateErrPlus1                              ! Corrected rate error + (asymmetric).
   real(kind=double) :: corrRateErrMinus1                             ! Corrected rate error - (asymmetric).
   real(kind=double) :: photcal                          ! ratio between the PSF encircled by the photometry and cal radii
   real(kind=double) :: caldet                           ! ratio between the PSF encircled by the cal and detected radii
   real(kind=double) :: sumIn2, sumIn3, sumOut2, sumOut3
   real(kind=double) :: sumBkg, testSumBkg    ! PSF-fraction within background extraction area
   real(kind=double) :: testSumSrc
   real(kind=double) :: corrMag
   ! Corrected magnitude
   real(kind=double) :: corrMagErrPlus                               ! Corrected magnitude error + (asymmetric).
   real(kind=double) :: corrMagErrMinus                              ! Corrected magnitude error - (asymmetric).
   logical :: extend
   logical :: empirical
   real(kind=single) :: countFrameRatio = 0.5
   real(kind=single), dimension(:), pointer :: rate                  ! Rates pointer.
   real(kind=single), dimension(:), pointer :: rateErr               ! Rates error pointer.
   real(kind=single), dimension(:), pointer :: corr_rate                  ! Rates pointer.
   real(kind=single), dimension(:), pointer :: corr_rate_err               ! Rates error pointer.
   real(kind=single), dimension(:), pointer :: corr_background_rate
   real(kind=single), dimension(:), pointer :: spb_coiloss
   real(kind=single) :: nSpb_coiloss, spb_coiloss_max, spb_coiloss_min
   real(kind=single), dimension(:), pointer :: bk_coiloss
   real(kind=single) :: nBk_coiloss
   integer(kind=int32), dimension(:), pointer :: nopixels 
   real(kind=single), dimension(:), pointer :: significance

   real(kind=double), dimension(:), pointer :: sourceRA              ! Source Right Ascension
   real(kind=double), dimension(:), pointer :: sourceDEC              ! Source Declination
   real(kind=double), dimension(:), pointer :: sourceLII              ! Source Galactic Longitude
   real(kind=double), dimension(:), pointer :: sourceBII              ! Source Galactic Lattitude
   integer(kind=int8), dimension(:), pointer :: sourceQF8              ! Source quality flag
   integer(kind=int16), dimension(:), pointer :: sourceQF16              ! Source quality flag
   integer(kind=int16) :: columnDataTypeFlag
   integer(kind=int8), dimension(:), pointer :: sourceCF              ! Source confusion flag

   real(kind=single), dimension(:), pointer :: chi2                  ! chi-square statistics pointer.
   real(kind=single), dimension(:), pointer :: chi2prob   ! chi-square probability pointer
   real(kind=single), dimension(:), pointer :: ksStat     ! Kolmogorov-Smirnov statistics pointer
   real(kind=single), dimension(:), pointer :: ksStatProb ! Kolmogorov-Smirnov probability pointer
   real(kind=single), dimension(:), pointer :: maxDev     ! max deviation in the number of sigmas

   real(kind = double), parameter :: PI = 3.141592653589793_double  
!****************************************************
! 07-06-2001 CB Need to read in the source coordinates 
! and aperture radius columns
!****************************************************
   real(kind=single), dimension(:), pointer :: sourceXCoord          ! Source x-coordinate.
   real(kind=single), dimension(:), pointer :: sourceYCoord          ! Source y-coordinate. 
   real(kind=single), dimension(:), pointer  :: sourceApertureRadius  ! Source aperture radius
   real(kind=single), dimension(:), pointer :: backGroundRate        ! Source background rate pointer.
   real(kind=single), allocatable, dimension(:)  :: backGroundRateSafe 
   real(kind=single), dimension(:), pointer :: mag                   ! Magnitude pointer.
   real(kind=single), dimension(:), pointer :: magErr                ! Magnitude error pointer.
   integer(kind=int8), dimension(:), pointer :: extended
   integer :: iLoop

   real(kind=single) :: magnitude
   real(kind=single) :: magnitudeErr
   real(kind=single) :: magnitudeBack
   real(kind=single) :: magnitudeBackErr
!   integer(kind=int8) :: extendedFlag
   integer(kind=int32) :: nSources                                   ! Number of sources.
   integer(kind=int32) :: nBins, nSourceRateRows                   ! Number of counts in the RATE table 
   integer(kind=int32) :: nNegative                                  ! Number of -ve count rates.

   ! output variables
   type(DataSetT) :: outSet
   type(TableT) :: outTable
   character(len=1024) :: outFile
 
   ! timeline variables
   real(kind=double), allocatable :: timeSource(:)
   real(kind=double), allocatable :: timeBack(:)
   real(kind=double), allocatable :: timeOut(:)

   ! rates file columns
   type(ColumnT) :: rateSourceCol, rateBackCol
   type(ColumnT) :: rateSourceOutCol
   type(ColumnT) :: rateSourceErrOutCol, rateBackOutCol
   type(ColumnT) :: rateBackErrOutCol
   type(ColumnT) :: timeSourceOutCol

   ! rates file pointers
   real(kind=Single), allocatable :: rateSourceArray(:), rateBackArray(:)
   real(kind=Single), allocatable :: SourceArrayError(:), BackArrayError(:)
   integer(kind=int32), pointer :: iSourceArray(:), iBackArray(:)
   integer(kind=int8) :: bgsubtracted

   real(kind=Single), pointer :: rateSourceOutArray(:)
   real(kind=Single), pointer :: rateSourceErrOutArray(:)
   real(kind=Single), pointer :: rateBackOutArray(:)
   real(kind=Single), pointer :: rateBackErrOutArray(:)
   real(kind=Double), pointer :: timeSourceOutArray(:)

   real(kind=double) :: source, background, bkg_safe, source_safe
   real(kind=double) :: timeStart, bkgTimeStart
   real(kind=double) :: timeEnd, bkgTimeEnd
   real(kind=double) :: timeDel, bkgTimeDel
   real(kind=double) :: timeExposure
   real(kind=double) :: lastTimeDel, bkgLastTimeDel

   real(kind=double) :: mgAver
   real(kind=double) :: nmgAver
   real(kind=double) :: mgMin
   real(kind=double) :: mgMax
   real(kind=double) :: rateAver

   ! VNY 17.02.2006 for test purposes
   real(kind=double) :: rateAver2, nRateAver2, rateErrAver2

   ! VNY 28.09.2007 fr test purposes
   real(kind=double) :: rateAver3row, nRateAver3row
   real(kind=double) :: bkgRateAver3row, nbkgRateAver3row

   real(kind=double) :: rateMin
   real(kind=double) :: rateMax
   real(kind=double) :: mgBackAver
   real(kind=double) :: nmgBackAver
   real(kind=double) :: rateBackAver
   real(kind=double) :: rateBackAverIm ! for the background rate from the imaging window
   real(kind=double) :: backErrAver
   real(kind=double) :: rateErrAver
   
   !---VNY 21.09.2001 array for the median filtration
   real(kind=single), dimension(9) :: rateAuxArray 
   real(kind=single) :: rateAux

   integer(kind=int32) :: nBack, nRegion, nBackRegion
   integer(kind=int32) :: srcNo
   real(kind=double) :: sourceErr
   real(kind=double) :: backErr

   integer :: nAvoidingSources
   real(kind=Single) :: avoidX(1:10)
   real(kind=Single) :: avoidY(1:10)
   real(kind=Single) :: avoidR(1:10)
   logical :: outsideOfAvoidingCircle

   real(kind=Single), pointer :: radiusDetectedArray(:)
   real(kind=Single), pointer :: sourceXPosArray(:)
   real(kind=Single), pointer :: sourceYPosArray(:)

   real(kind=Single), pointer :: sourceRateArray(:)
   real(kind=Single) :: rate0
   real(kind=Single), pointer :: sourceSignificanceArray(:)
   real(kind=Single) :: significance0

   real(kind=Single), pointer :: backRadiusArray(:)
   real(kind=Single), pointer :: backXPosArray(:)
   real(kind=Single), pointer :: backYPosArray(:)

   character(len=20) :: sourceShape
   character(len=20) :: backShape
   character(len=80) :: sourceShapeComment
   character(len=80) :: backShapeComment
   character(len=80) :: clcComment1, clcComment2

   !--- VNY 14.05.2001 - origin of the OSW
    character(len=255) :: window0                                   ! Origin of the OSW
    integer(kind=int32) :: windowX0, windowX1, windowY0, windowY1   ! Coordinates of the OSW origin
    character(len=1) :: ch1, ch2, ch3, ch4, ch5
    !---VNY 9.05.2002 - psf shift within a pixel
    real(kind=single) :: psfShiftX, psfShiftY

    !   Variables for checking PSF-mapping routine
    real(kind=single), dimension(:, :), pointer :: psf ! To hold the point-spread function
    integer(kind=int32) :: xRes, yRes
    real(kind=double) :: xPos0, yPos0
    integer(kind=int32) :: ix, iy
    real(kind=single) :: dxLeft, dxRight, dyDown, dyUp
    real(kind=single) psfRadius, xyDistance, xc, yc, sumIn, sumOut, radiusDetected
!    real(kind=double) :: PsfFractionOfTotalCounts
!    real(kind=single) :: radiusTest
    logical :: binBpe
    integer(kind=int32) :: backMax
    integer(kind=int32), allocatable :: backDist(:)
    real(kind=double) :: backSum, backSumFound, medianBkg, medianBkgError
    logical :: medianFound
    real(kind=double) :: prop
    integer(kind=int32) :: ja
    real(kind=double) :: ratioBS
    !real(kind=double) :: testBkg, testInt, testSrc, testZnam
    real(kind=double) :: testBkg, testSrc, testZnam
    real(kind=double) :: sumInB, sumOutB, sumIn2B, sumIn3B, sumOut2B, sumOut3B
    real(kind=double) :: areaInB, areaOutB, areaIn2B, areaOut2B, areaIn3B, areaOut3B
    real(kind=double) :: areaBkgB, sumBkgB
    real(kind=double) :: corrRateErr
!    integer :: colNum
    integer :: position

    real(kind=single),  dimension(:), pointer :: psf1Correction          ! psf1 correction
    real(kind=single),  dimension(:), pointer :: psf2Correction          ! psf2 correction
    real(kind=single) :: nullValue = -999.999  
    integer :: columnDim(1) =(/1/)
!---VNY 26.02.2004 test file
!    integer(kind=int16) :: posPoint   
!    character(len=1024) :: textOutFile, outFile2 !+ 
    real(kind=single) :: corrToSource 
    integer(kind=int32) :: nHighCoincidenceLoss, nCountsOverLimit
    real(kind=double) :: maxCountsPerFrame
    integer(kind=int16) :: percentage
    integer(kind=int16) :: mod8corrupted
    integer(kind=int16) :: dtCorrection
    logical :: hasOmimiBlock
    real(kind=single) :: imxPos, imyPos
    integer :: imiPos, imjPos

!
!--- for random_Poisson
!    real(kind=single) :: mu
!    logical :: first
!    integer(kind=int16) :: ival
!    real(kind=single) :: rnd_s, rnd_d, rnd_p, rnd_q, rnd_p0
!    logical :: rnd_full_init
!    integer(kind=int16) :: rnd_l, rnd_m
!    real(kind=single) :: rnd_pp(35)
!    integer :: seedSize
!    integer, allocatable :: seedArray(:)
!    integer :: simDistributionArray(0:1000), lossDistributionArray(0:1000)
!    real :: simCountsArray(1000), lossCountsArray(1000)
!    real :: pMean, pMeanLoss, nMean, dispLeft, dispRight, diff
!    real :: nMeanLeft, nMeanRight
!    integer :: nFrames, iFlux , iFluxFrame
!    real :: obsDuration , frameDuration
    real :: sim_frameTime, sim_timeBin, trueCountRate, lossCountRate
!    real :: simulatedCountRate, dispersionLeft, dispersionRight
    real :: dispersionLeft, dispersionRight
!    real :: averSim, averLoss, dispLeftLoss, dispRightLoss
!    integer :: sampleSize, indexSim, indexLoss
!-----
    real :: lossCountsArray2(2000), dispLeftArray(2000), dispRightArray(2000), rmsArray(2000)
    real :: loss2corr(2000), loss2corrErrLeft(2000), loss2corrErrRight(2000)
    real :: loss2corrRMS(2000)
    real :: maxLoss, rmsValue
    integer :: maxCounts, nAux, iloss, verbFlag

    real(double) :: mjd
    real(single) :: tdpCorrFactor

    logical :: degradationCorrection
    integer(kind=int32) :: mod8flag

    !-------------------------------------
    ! Variables for the variability tests
    !-------------------------------------
    real(kind=single) :: Bin_Width
    integer :: Binning_Factor
    real(kind=single),dimension(:),pointer :: Time_Vect
    real(kind=single),dimension(:,:),pointer :: Net_Rate, Net_Err 
    real(kind=single),dimension(:,:),pointer :: Bkg_Rate, Bkg_Err
    real(kind=single),dimension(:,:),pointer :: Net_Count, Net_Total
    logical,dimension(:),pointer :: Good_Bin
    integer :: Nlight , Nbin
    character(len=12) :: Which_Tests
    real(kind=single) :: Exposure_Time
    real(kind=single) :: Time_Zero
    logical(boolean),dimension(:),pointer  :: gotnullvals
    integer :: iaux
    type(data_tests),dimension(1) :: Light_Curve
    logical :: useImagingFile

    !--------------------------------------------------------
    !! Following are the variables for saving testimage77.fit
    character(len=255) :: outSetCheck
    integer(int32) :: outDimCheck(2) 
    type(DataSetT) :: outDataSetCheckH 
    type(ArrayT) :: outArrayCheckH 
    real(single), dimension(:, :), pointer :: outOSWcheckP  
    !--------------------------------------------------------

    !----------------------------------------------------
    ! The following will be used for the imaging data set
    type(DataSetT) :: imDataSetH  
    type(ArrayT) :: imArrayH  
    real(kind=single)          :: imexposureTime
    logical(kind=Bool) :: imbinBPE
    integer  :: dim(2), imnx, imny     ! size of the image
    integer(kind=int32) :: imwindowX0, imwindowX1, imwindowY0, imwindowY1 
    integer(kind=int32) :: imwindowDX, imwindowDY
    character(len=6) :: imFilter
    integer(kind=int16), pointer :: imImage1(:,:)
    real(kind=single), pointer :: imImage2(:,:)
    real(kind=single), pointer :: imImage3(:,:)
    integer(kind=int32), pointer :: imImage(:,:)  
    integer(kind=int32) :: allocErr   
    integer :: i1, j1, is,ie, js,je
    real(kind=single) :: imAver, nimAver, imagingBkg, imagingBkgRate, imagingBkgBin
    real(kind=single) :: imagingBkgRateErr, imagingErr, imagingBkgBinErr
    logical(kind=bool) :: imagingBkgOk 
    logical(kind=bool) :: bkgFromImage, ignoreMod8noise, subtractBkg
    logical :: existFile

    real(kind=single) :: snoCorrAver, sclCorrAver, nsCorrAver
    real(kind=single) :: sourceSubtracted

   integer(kind=int32) :: imbinAx1                               
   integer(kind=int32) :: imbinAx2 

   integer(kind=int32) :: imbinX
   integer(kind=int32) :: imbinY



   ! 2012-10-18 modifying to create the same background image 
   ! as in omdetect
   real(single), pointer :: backgroundimage(:, :) => null()
   real(single), pointer   :: backgroundsigmaImage(:, :) => null() 
   logical(boolean), pointer   :: qualArray(:, :) => Null() 
   real(single) :: averageBkg, averageSigma
   logical(boolean) :: useImagingCoordinates
   type(sourceParameters), pointer :: sourceData(:)  
   integer(int16), pointer :: levelImage(:, :) => null()
   integer(int8) :: backgroundMethod
   real(single) :: apertureRadius  
   integer(int16) :: apnx, apny ! for aperture photometry module
   integer(int16)  :: nPSources

   real(single) :: im2bkg, im2bkgErr, im2Aperture, im2BkgArea
   !real(single) :: avSrcExtrArea, avBkgExtrArea
   real(single) :: avSrcCLcorr, navSrcCLcorr, avBkgCLcorr, navBkgCLcorr
   real(single) :: avMagCLcorr, navMagCLcorr

   backgroundMethod = 6 ! The background method for aperture photometry
   imagingBkg=0.0
   imagingBkgOk=.false.
   imagingBkgRate=0.0
   imagingBkgRateErr=1.0
   imagingBkgBin=0.0
   imagingBkgBinErr=1.0
   bkgFromImage=.false.
   ignoreMod8noise=.false.
   subtractBkg=.true.
   im2bkg=-1.0
   im2bkgErr=-1.0
   im2Aperture=-1.0
   im2BkgArea=-1.0

   Which_Tests="both" ! Kolmmogorov-Smirnov and Chi-squared tests

   !---------------------------
   !!    !real(single) :: coeffs(2)
   !    real(double), pointer :: coeffs(:)=> null()
   !---------------------------
   degradationCorrection = .true.
   
   lossCountsArray2=0.0
   dispLeftArray=0.0
   dispRightArray=0.0
   
   ! get parameters
   sourceReg = stringParameter("srcregionset")
   backReg = stringParameter("bkgregionset")
   sourceRates = stringParameter("srcrateset")
   backRates = stringParameter("bkgrateset")
   sourceList = stringParameter("sourcelistset")
   wdxSet = stringParameter("wdxset")
   outFile = stringParameter("outset")
   mod8corrupted=intParameter("mod8corrupted")

   !    ignoreMod8noise = booleanParameter("ignoremod8noise")
   

   !    dtCorrection=intParameter("dtcorrection")
   dtcorrection=0

   bkgFromImage = booleanParameter("bkgfromimage")
   subtractBkg = booleanParameter("subtractbkg")
   
   imagingDataSet=""
   useImagingFile=.false.
   useImagingCoordinates=.false.
   imagingDataSet = stringParameter("imageset")
   
   !write(*,*) "L660 dtCorrection=", dtcorrection
   !write(*,*) "L661 mod8corrupted=", mod8corrupted
   !write(*,*) "L662 ignoreMod8noise=", ignoreMod8noise
   !write(*,*) "L663 imagingDataSet=", trim(imagingDataSet)
   iaux=len(trim(imagingDataSet))
   !write(*,*) "L665 iaux=", iaux
   if (iaux .gt. 1) useImagingFile=.true.
   !write(*,*) "L667 useImagingFile=", useImagingFile
   !write(*,*) "L668 bkgFromImage=", bkgFromImage
   !write(*,*) "L670 subtractbkg=", subtractbkg
   
   ! Check whether the requested imaging file does exist
   existFile=.false.
   if (.not. bkgFromImage) useImagingFile=.false.

   if (useImagingFile) then
      write(errstr,'(a,a)') "Requested the use of the imaging-mode data: ",trim(imagingDataSet)
      call message(AppMsg, 2)
      !write(*,*) "L703 using the imaging-mode file for the background calculation"
      inquire (FILE=trim(imagingDataSet), EXIST=existFile)
   endif
   if (.not. existFile .and. useImagingFile) then 
      call warning('OMLCBUILD','Could not find imaging-mode data')
      !write(*,*) "L709 imaging-mode file does not exist"
      useImagingFile=.false.
      bkgFromImage=.false. ! The imaging background cannot be determined
   !else 
   !   write(*,*) "L711 imagingDataSet=", trim(imagingDataSet)
   endif

   if (useImagingFile) then
      
      hasOmimiBlock=.false.
      
      imDataSetH = DataSet(imagingDataSet, read) 
      if (hasBlock(imDataSetH, "OMIMI1")) then
         hasOmimiBlock=.true.
      endif
      if (hasOmimiBlock) then
         imArrayH = array(imDataSetH, "OMIMI1", read)
         imexposureTime = real32Attribute(imArrayH, "EXPOSURE") ! Image exposure time
         imbinBPE = booleanAttribute(imArrayH, "BINBPE")
         imFilter = stringAttribute(imArrayH, "FILTER")
         imbinAx1 = int32Attribute(imArrayH, "BINAX1")
         imbinAx2 = int32Attribute(imArrayH, "BINAX2")
      else
         imArrayH = array(imDataSetH, "PRIMARY", read)
         imexposureTime = real32Attribute(imDataSetH, "EXPOSURE") ! Image exposure time
         imbinBPE = booleanAttribute(imDataSetH, "BINBPE")
         imFilter = stringAttribute(imDataSetH, "FILTER")
         imbinAx1 = int32Attribute(imDataSetH, "BINAX1")
         imbinAx2 = int32Attribute(imDataSetH, "BINAX2")
      endif
      
      dim = dimensions(imArrayH)
      !----
      imbinX = imbinAx1
      if (imbinX .lt. 1.0) imbinX = 1.0
      imbinY = imbinAx2
      if (imbinY .lt. 1.0) imbinY = 1.0
      
      binx=imbinX
      biny=imbinY
      !write(*,*) "L813 dim1=", dim(1), " dim2=", dim(2)
      if (hasOmimiBlock) then
         !write(*,*) "L771 the image has OMIMI block"
         if (hasAttribute(imArrayH,"WINDOWX0")) then
            !write(*,*) "L773 the image has keyword WINDOWX0"
            imwindowX0= int32Attribute(imArrayH, "WINDOWX0")
            imwindowY0= int32Attribute(imArrayH, "WINDOWY0")
            imwindowDX= int32Attribute(imArrayH, "WINDOWDX")
            imwindowDY= int32Attribute(imArrayH, "WINDOWDY")
            imwindowX1=imwindowX0+imwindowDX-1.0
            imwindowY1=imwindowY0+imwindowDY-1.0
            useImagingCoordinates=.true.
            ! if the lower left corner of the image is defined
            ! then the coordinates of the source can be calculated
            ! for the imaging-mode window
         else
            !write(*,*) "L773 the image does not have keyword WINDOWX0"
            imwindowX0= 0
            imwindowY0= 0
            imwindowDX= dim(1)
            imwindowDY= dim(2)
            imwindowX1=imwindowX0+imwindowDX-1.0
            imwindowY1=imwindowY0+imwindowDY-1.0
         endif
      else
         !write(*,*) "L838 the image does not have OMIMI block \n"
         imwindowX0= 0
         imwindowY0= 0
         imwindowDX= dim(1)
         imwindowDY= dim(2)
         imwindowX1=imwindowX0+imwindowDX-1.0
         imwindowY1=imwindowY0+imwindowDY-1.0
      endif
      !write(*,*) "L790 imwindowX0=", imwindowX0, " imwindowY0=", imwindowY0, &
      !     " useImagingCoordinates=", useImagingCoordinates

      if (hasOmimiBlock) then
         imImage => int32array2data(imArrayH)
         dim = shape(imImage)
         imnx = dim(1)
         imny = dim(2)   
      else
         imImage1 => int16array2data(imArrayH)
         dim = shape(imImage1)
         imnx = dim(1)
         imny = dim(2)   
      endif

      ! Allocate arrays for the backround image 
      allocate (imImage2(0 : imnx-1, 0 : imny-1), stat = allocErr)
      imImage2=-1.0

      allocate (imImage3(0 : imnx-1, 0 : imny-1), stat = allocErr)
      imImage2=-1.0

      allocate (qualArray(0 : imnx-1, 0 : imny-1), stat = allocErr)
      qualarray=.false.
      
      if (hasOmimiBlock) then
         do i=0,imnx-1
            do j=0,imny-1
               imImage2(i,j)=imImage(i,j)
               imImage3(i,j)=imImage(i,j)
            enddo
         enddo
      else
         do i=0,imnx-1
            do j=0,imny-1
               imImage2(i,j)=imImage1(i,j)
               imImage3(i,j)=imImage1(i,j)
            enddo
         enddo
      endif
      
      !----------------------
      !write(*,*) "L799 creating testimage77.fit"
      !outSetCheck="testimage77.fit"
      !outDimCheck(1)=imWindowX1-imWindowX0+1
      !outDimCheck(2)=imWindowY1-imWindowY0+1
      !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
      !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
      !     real32, outDimCheck, "Counts", "IMAGE")!             
      !outOSWcheckP => real32Array2Data(outArrayCheckH)
      !outOSWcheckP = imImage2 
      !! Release handles on data.
      !call release(outArrayCheckH)        
      !call release(outDataSetCheckH) 
      ! 
      !write(*,*) "L812 Test image 77 saved:", trim(outSetCheck)!
      !--------------------------------


       !write(*,*) "L818 trying to create the background image" 
       !call createBackgroundImage2(imImage2, qualArray, 10000_int16, imbinX, &
      call createBackgroundImage2(imImage3, qualArray, 60_int16, imbinX, &
           averageBkg, averageSigma, levelimage=levelimage, &
           backgroundimage=backgroundimage, backgroundsigmaimage=backgroundsigmaImage)
       !write(*,*) "L848 finished creating background image averageBkg=", averageBkg, &
       !     " averageSigma=", averageSigma

       write(errstr,'(a,f6.3)') "finished creating background image, average background: ", averageBkg 
       call message(AppMsg, 2)

       !write(*,*) "---"
       !do i=119,123
       !   do j=108,112
       !      write(*,*) "L881 i=", i, " j=", j, " imimage3=", (imImage3(i,j)
       !   enddo
       !enddo
       !write(*,*) "---"

       !----------------------
       !write(*,*) "L846 creating testbkgimage77.fit"
       !outSetCheck="testbkgimage77.fit"
       !outDimCheck(1)=imWindowX1-imWindowX0+1
       !outDimCheck(2)=imWindowY1-imWindowY0+1
       !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
       !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
       !     real32, outDimCheck, "Counts", "IMAGE")!             
       !outOSWcheckP => real32Array2Data(outArrayCheckH)
       !outOSWcheckP = backgroundimage  
       !! Release handles on data.
       !call release(outArrayCheckH)        
       !call release(outDataSetCheckH) 
       ! 
       !write(*,*) "L859 testbkgimage77 saved:", trim(outSetCheck)!
       !--------------------------------


    endif ! useImagingFile

    !************************************************************
    !************************************************************

!--- VNY test 26.02.2004
!   posPoint=scan(trim(outFile), '.', .true.)
!   outFile2=outFile
!   if (posPoint .ne. 0) then
!      outFile2(posPoint:1024)=' '
!      write(textOutFile, '(a,a)') trim(outFile2), ".txt"
!   else
!      textOutFile='omlcbuild_out.txt'
!   endif
!   open(unit = 2, file = textOutFile, status = 'unknown')
!   write(2,*) "Testing coincidence loss correction in OMLCBUILD 26.02.2004"
!---

   sourceRegSet = dataSet(TRIM(sourceReg), read)
   backRegSet = dataSet(TRIM(backReg), read)

   ! Check of the length of the source rate table
   ! get a handle on the source rates file
   sourceSet = dataSet(TRIM(sourceRates), read)   
   srcRateExists=.false.
   nSourceRateRows=0
   if (hasBlock(sourceSet, "RATE")) then
      sourceTable = table(sourceSet,"RATE")
      nSourceRateRows = numberOfRows(sourceTable) ! Number of points in the light curve 
      timeDel = real64Attribute(sourceTable, "TIMEDEL")
      write(errstr,*) "Length of the Source Rates Table:  ", nSourceRateRows
      call message(AppMsg, 2)
      call release(sourceTable)
   endif

   call release(sourceSet)

   ! Read in the OM OSW Source List file
   inDataSetH=dataSet(sourceList, modify)
   ! Protection against the absence of the data-block SRCLIST
   if (hasBlock(inDataSetH, "SRCLIST")) then
      !**************************************************************
      ! Read in the detection limit count rate from the PRIMARY header
      !**************************************************************
      detectionLimitCountRate = real(real32Attribute(inDataSetH, "CRLIMIT"))
      !write(*,*) "L923 detectionLimitCountRate=", detectionLimitCountRate
      write(errstr,'(a,f6.3)') "detection limit count rate: ",detectionLimitCountRate
      call message(AppMsg, 2)

      ! Get a handle on the source list binary table extension.
      inTableH = table(inDataSetH, "SRCLIST")

      ! Read in the source list table extension keywords.
      dataMode = stringAttribute(inDataSetH, "DATAMODE")
      filter = stringAttribute(inDataSetH, "FILTER")
      exposure = real32Attribute(inDataSetH, "EXPOSURE")
      binBpe= booleanAttribute(inDataSetH, "BINBPE")

      
      !*************************************************************************
      ! Need the modified Julian date for time-dependent sensitivity correction
      !************************************************************************
      mjd = ModifiedJulianDate(trim(stringattribute(inDataSetH, "DATE-OBS")))

      !*********************************************************
      ! Time-dependent sensitivity correction-factor
      !*********************************************************
      ! VNY 13.02.2008 using the CAL-routine for computing the time-dependent
      ! sensitivity correction factor
      !tdpCorrFactor = TDPCorrectionFactorOld(mjd, filter)
      !write(*,*) "L585 tdpCorrectionFactor=", tdpCorrFactor

      
      !********************************************************
      ! Read in the OM ODF Priority Window Data Auxiliary file.
      !********************************************************
      wdxDataSetH = dataSet(wdxSet, read)

      
      !************************************************************************
      ! Get the frame time from the OM ODF Priority Window Data Auxiliary file.
      !************************************************************************
      ! New parameter added to the subroutine getFrameTime: binBpe
      call getFrameTime(wdxDataSetH, frameTime, deadFrac, binBpe)  ! Frame time is in milli-seconds 

      !**************************************************************
      ! Computing a new empirical (experimental) dead fraction correction
      !**************************************************************
      !deadFracCorrection=1.0+2.8*(deadFrac-0.017)
      deadFracCorrection=1.0+1.0*(deadFrac-0.017)
!      write(*,*) "L793 deadFracCorrection=", deadFracCorrection
!      if (dtCorrection .gt. 0) then
!         write(errstr,'(a,f8.5)') "Applying an extra dead-time correction factor: ", &
!              deadFracCorrection
!         call message(AppMsg, 2)
!      endif


      frameTime = frameTime / 1000.0    ! Convert frame time to seconds

      
      if (frameTime .gt. 0.0) then
         maxCountsPerFrame=0.99/frameTime
      else
         maxCountsPerFrame=200.0
      endif
      maxCounts=6*maxCountsPerFrame
      if (maxCounts .gt. 2000) maxCounts=2000

! VNY 25.03.2004 tests
      
!      sourceErr=0.0
!      corrRateErrMinus=0.0
!      corrRateErrPlus=0.0
!      extend=.false.
!      do i=1,200
!         source=real(i)
!!      source=78.0
!!      sourceErr=sqrt(source)
!!      corrRate = CorrectedCountRate(dble(source), sourceErr, &
!!           frameTime, deadFrac, extend, corrRateErrMinus, &
!!           corrRateErrPlus)
      !write(*,*) " L552 sourceRate=",source, " corrRate=", corrRate 
      !write(*,'(a, i5, f9.2, f9.2, f9.2, f9.2, f9.2)') " i=", i, &
      !     corrRate, corrRate/real(i), sourceErr, &
      !     corrRateErrMinus, corrRateErrMinus/sourceErr
!      enddo


!
      !write(*,*) "****************************************"
!      write(*,*) " 25.03.2004 Trying random_Poisson "
!      obsDuration=10.0 ! seconds (timeBin)
      sim_timeBin=timeDel
!      frameDuration=0.01 ! seconds (frameTime)
      sim_frameTime=frameTime !0.01
!      iFlux=120 ! counts per second (trueCountRate)

      !write(*,*) " L644 sim_frameTime=", sim_frameTime, &
      !" sim_timeBin=", sim_timeBin, " trueCountRate=", trueCountRate
      !write(*,*) " L703 maxCounts=", maxCounts
      do i=1,maxCounts
         trueCountRate=i
         verbFlag=0
         !if (i .eq. 20) verbFlag=1
         call coincLossAsymErr(sim_frameTime, sim_timeBin, trueCountRate, &
              lossCountRate, dispersionLeft, &
              dispersionRight, rmsValue, verbFlag)
         lossCountsArray2(i)=lossCountRate
         dispLeftArray(i)=dispersionLeft
         dispRightArray(i)=dispersionRight
         rmsArray(i)=rmsValue
         !write(*,'(a,f5.1,a,f5.1,a,f5.1,a,f5.1,a,f5.1,a,f5.1,a,f5.1)') &
         !     " L671 true=", trueCountRate,  " loss=", lossCountRate, &
         !     " disp= (", dispersionLeft, ",", dispersionRight, ")"
      enddo
      ! Reorganise the array into lost counts depending on true counts
      maxLoss=0.0
      do i=1,maxCounts
         if (lossCountsArray2(i) .gt. maxLoss) maxLoss=lossCountsArray2(i)
      enddo ! i
      nAux=int(maxLoss)
      loss2corr=0.0
      loss2corrErrLeft=0.0
      loss2corrErrRight=0.0
      do i=1,nAux
         do iloss=1,maxCounts-1
            if (lossCountsArray2(iloss) .lt. real(i) .and. &
                 lossCountsArray2(iloss+1) .ge. real(i)) then
               loss2corr(i)=real(iloss)
               loss2corrErrLeft(i)=dispLeftArray(iloss)
               loss2corrErrRight(i)=dispRightArray(iloss)
               loss2corrRMS(i)=rmsArray(iloss)
            endif
         enddo ! iloss
      enddo ! i
      !write(*,*) "----------- loss2corr array --------"
      !do i=1,nAux
         !write(*,*) " i=", i, " loss2corr(i)=", loss2corr(i), &
         !     " lossErrLeft=", loss2corrErrLeft(i), " lossErrRight=", loss2corrErrRight(i), &
         !     " lossRMS=", loss2corrRMS(i)
      !enddo
      !write(*,*) "------------------------------------"
!!!!!!!
      
!      write(*,*) " Using the subroutine simulateCountRate"
!      write(*,*) "****************************************"
!      call initRandomSequence(2)
!      sampleSize=500
!      do trueCountRate=250,350
!         averSim=0.0
!         nMean=0.0
!         averLoss=0.0
!         simCountsArray=0.0
!         lossCountsArray=0.0
!         simDistributionArray=0
!         lossDistributionArray=0
!         do i=1,sampleSize
!            call simulateCountRate(sim_frameTime, sim_timeBin, trueCountRate, &
!                 lossCountRate, simulatedCountRate, dispersionLeft, &
!                 dispersionRight)
!            !write(*,*) "i=", i, " simulatedCountRate=", simulatedCountRate, &
!            !  " lossCountRate=", lossCountRate
!            simCountsArray(i)=simulatedCountRate
!            lossCountsArray(i)=lossCountRate
!            averSim=averSim+simulatedCountRate
!            averLoss=averLoss+lossCountRate
!            nMean=nMean+1.0
!            indexSim=int(simulatedCountRate)
!            indexLoss=int(lossCountRate)
!            if (indexSim .lt. 1000) then
!               simDistributionArray(indexSim)=simDistributionArray(indexSim)+1
!            endif
!            if (indexLoss .lt. 1000) then
!               lossDistributionArray(indexLoss)=lossDistributionArray(indexLoss)+1
!            endif
!         enddo
!         !write(*,*) " ***************************************"
!         !do i=0, 200
!         !   write(*,*) " i=", i, " sim=", simDistributionArray(i), &
!         !        " loss=", lossDistributionArray(i)
!         !enddo
!         !write(*,*) "****************************************"
!         averSim=averSim/nMean
!         averLoss=averLoss/nMean
!         !write(*,*) " averSim=", averSim, " averLoss=", averLoss
!         ! compute dispersions
!         dispLeft=0.0
!         dispRight=0.0
!         nMeanLeft=0.0
!         nMeanRight=0.0
!         do i=1,sampleSize
!            diff=simCountsArray(i)-averSim
!            if (diff .lt. 0.0) then
!               dispLeft=dispLeft+diff*diff
!               nMeanLeft=nMeanLeft+1.0
!            endif
!            if (diff .gt. 0.0) then
!               dispRight=dispRight+diff*diff
!               nMeanRight=nMeanRight+1.0
!            endif
!         enddo
!         if (nMeanLeft .gt. 1.0) then
!            dispLeft=sqrt(dispLeft/(nMeanLeft-1.0))
!         endif
!         if (nMeanRight .gt. 1.0) then
!            dispRight=sqrt(dispRight/(nMeanRight-1.0))
!         endif
!         !write(*,*) " dispersions for averSim: left=", dispLeft, " right=", dispRight
!         ! Dispersionns for lossAver
!         dispLeftLoss=0.0
!         dispRightLoss=0.0
!         do i=1,sampleSize
!            diff=lossCountsArray(i)-averLoss
!            if (diff .lt. 0.0) then
!               dispLeftLoss=dispLeftLoss+diff*diff
!               nMeanLeft=nMeanLeft+1.0
!            endif
!            if (diff .gt. 0.0) then
!               dispRightLoss=dispRightLoss+diff*diff
!               nMeanRight=nMeanRight+1.0
!            endif
!         enddo
!         if (nMeanLeft .gt. 1.0) then
!            dispLeftLoss=sqrt(dispLeftLoss/(nMeanLeft-1.0))
!         endif
!         if (nMeanRight .gt. 1.0) then
!            dispRightLoss=sqrt(dispRightLoss/(nMeanRight-1.0))
!         endif
!         !write(*,*) " dispersions for averLoss: left=", dispLeftLoss, &
!         !     " right=", dispRightLoss
!         !write(*,*) "-----------------------"
!         write(*,'(a,f5.1,a,f5.1,a,f5.1,a,f5.1,a,f5.1,a,f5.1,a,f5.1)') &
!              " true=", trueCountRate, " sim=", averSim, " disp= (", &
!              dispLeft, ",", dispRight, ")  loss=", averLoss, &
!              " disp= (", dispLeftLoss, ",", dispRightLoss
!      enddo
!-------------------------------

      !********************
      ! Get the ccd mode.
      !********************
      if (trim(dataMode) .eq. "IMAGING") then
         ccdMode = IMAGE
      else
         ccdMode = FAST
      end if

      !**************************************
      ! Get the filter Id integer value
      !**************************************
      filterIdInt = filterIdToInt(filter)
      
      !*******************
      ! Set the CAL state
      !********************
      call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdInt)
      

      ! VNY 13.02.2008 checking the time degradation coefficients from CAL

      !call CAL_omGetDegradationCoeffs( coeffs )
      !write(*,*) "L865 coeffs=", coeffs

      tdpCorrFactor = TDPCorrectionFactor(mjd, filter)
      !write(*,*) "L868 tdpCorrFactor=", tdpCorrFactor, " (from ommagnitude)"

      !*************************************************************
      ! 17th May 2001 - Due to the changes to the photometry in
      ! omdetect, now need to scale the source counts from the 12
      ! pixel aperture radius to that specified by the CAL.
      ! For U, B and V there will be no correction.
      !************************************************************

      photometryApertureRadius = CAL_omGetApertureRadius()
!      write(*,*) "L1054 photometryApertureRadius=", photometryApertureRadius

      if (photometryApertureRadius .gt. calApertureRadius) then
         !**********************************************
         ! Compute the correction factor needed to scale  
         ! the source counts from an aperture radius of
         ! 12.0 pixels to the apertureRadius
         !**********************************************
         scaleCounts = .true.
      end if

      !**************************
      ! Release handles on data.
      !**************************
      call release(wdxDataSetH)

      !******************************************************
      ! Calculate the instrumental limiting magnitude from
      ! the detection limit count rate
      !******************************************************
      if (detectionLimitCountRate .gt. 0.0) then      
         corrRate = CorrectedCountRate(detectionLimitCountRate, error, frameTime, &
            deadFrac, .false., .false., corrRateErrMinus, corrRateErrPlus)
         !write(*,*) "L900 corrRate=", corrRate, " detectionLimitCountRate=", &
         !     detectionLimitCountRate 
         if (corrRate .gt. 0.0) then
            !**********************************************************
            ! May need to scale the counts from 12 to 35 pixels for uv
            !*********************************************************
            if (scaleCounts) then
               corrRate = corrRate * (photometryApertureRadius / calApertureRadius) ** 2
            end if
            call CAL_omPhotoMagnitude(corrRate, detectionLimitMagnitude, corrRateErrMinus, &
                                   corrRateErrMinus, corrMagErrMinus, corrMagErrPlus)
         else
            detectionLimitMagnitude = 0.0d0
         end if
      else
         detectionLimitmagnitude = 0.0
      end if
      call setAttribute(inDataSetH, "MAGLIMIT", real(detectionLimitMagnitude), "", "detection limiting magnitude")

      !******************************************************************
      ! Get the number of rows in the source list binary table extension.
      !******************************************************************
      nSources = numberOfRows(inTableH)
      !---VNY 26.03.2002 Setting a limit to the input rows to 1000 sources
      !---VNY 15.06.2004 Setting a limit to the input rows to 2000 sources
      !
      if (nSources .gt. 0 .and. nSources .lt. 2000) then
      !if (nSources .gt. 0 .and. nSources .lt. 2) then
!*********************************************************************
! VNY 2009.0518
!*********************************************************************
         ! Get handles and pointers on the source list binary table extension colums.
         inColumnFlag=1
         !---VNY 25.10.2002 Check all the columns of the source list file
         if (hasColumn(inTableH, "RATE")) then
            inColumnH(1) = column(inTableH, "RATE", modify)
            rate => real32Data(inColumnH(1))
         else
            inColumnFlag(1)=0
            allocate(rate(0:nSources-1))
         endif
         
         if (hasColumn(inTableH, "RATE_ERR")) then
            inColumnH(2) = column(inTableH, "RATE_ERR", modify)
            rateErr => real32Data(inColumnH(2))
         else
            inColumnFlag(2)=0
            allocate(rateErr(0:nSources-1))
         endif

         if (hasColumn(inTableH, "MAG")) then
            inColumnH(3) = column(inTableH, "MAG", modify)
            mag => real32Data(inColumnH(3))
         else
            inColumnFlag(3)=0
            allocate(mag(0:nSources-1))
         endif

         if (hasColumn(inTableH, "MAGERR")) then
            inColumnH(4) = column(inTableH, "MAGERR", modify)
            magErr => real32Data(inColumnH(4))
         else
            inColumnFlag(4)=0
            allocate(magErr(0:nSources-1))
         endif

         !---VNY 08.10.2002 Check if has the column CORR_BACKGROUND_RATE
         if (hasColumn(inTableH, "CORR_BACKGROUND_RATE")) then
            inColumnH(17)=column(inTableH, "CORR_BACKGROUND_RATE", modify)
            corr_background_rate => real32Data(inColumnH(17))
         else
            inColumnFlag(17)=0
            allocate(corr_background_rate(0:nSources-1))
         endif

         if (hasColumn(inTableH, "EFLAG")) then
            inColumnH(5) = column(inTableH, "EFLAG", modify)
            extended => int8Data(inColumnH(5))
         else
            inColumnFlag(5)=0
            allocate(extended(0:nSources-1))
         endif

         if (hasColumn(inTableH, "BACKGROUND_RATE")) then
            inColumnH(6) = column(inTableH, "BACKGROUND_RATE", modify)
            backGroundRate => real32Data(inColumnH(6))
         else
            inColumnFlag(6)=0
            allocate(backGroundRate(0:nSources-1))
         endif

         allocate(backGroundRateSafe(0:nSources-1))
         backgroundRateSafe=backgroundRate

         if (hasColumn(inTableH, "XPOS")) then
            inColumnH(7) = column(inTableH, "XPOS", read)
            sourceXCoord => real32Data(inColumnH(7))
         else
            inColumnFlag(7)=0
            allocate(sourceXCoord(0:nSources-1))
         endif

         if (hasColumn(inTableH, "YPOS")) then
            inColumnH(8) = column(inTableH, "YPOS", read)
            sourceYCoord => real32Data(inColumnH(8))
         else
            inColumnFlag(8)=0
            allocate(sourceYCoord(0:nSources-1))
         endif

         if (hasColumn(inTableH, "APERTURE")) then
            inColumnH(9) = column(inTableH, "APERTURE", read)
            sourceApertureRadius => real32Data(inColumnH(9))
         else
            inColumnFlag(9)=0
            allocate(sourceApertureRadius(0:nSources-1))
         endif

         if (hasColumn(inTableH, "RA")) then
            inColumnH(10) = column(inTableH, "RA", read)
            sourceRA => real64Data(inColumnH(10))
         else
            inColumnFlag(10)=0
            allocate(sourceRA(0:nSources-1))
         endif

         if (hasColumn(inTableH, "DEC")) then
            inColumnH(11) = column(inTableH, "DEC", read)
            sourceDEC => real64Data(inColumnH(11))
         else
            inColumnFlag(11)=0
            allocate(sourceDEC(0:nSources-1))
         endif

         if (hasColumn(inTableH, "LII")) then
            inColumnH(12) = column(inTableH, "LII", read)
            sourceLII => real64Data(inColumnH(12))
         else
            inColumnFlag(12)=0
            allocate(sourceLII(0:nSources-1))
         endif
 
         if (hasColumn(inTableH, "BII")) then
            inColumnH(13) = column(inTableH, "BII", read)
            sourceBII => real64Data(inColumnH(13))
         else
            inColumnFlag(13)=0
            allocate(sourceBII(0:nSources-1))
         endif    
   
         columnDataTypeFlag=0
         if (hasColumn(inTableH, "QFLAG")) then
            inColumnH(14) = column(inTableH, "QFLAG", read)
            columnDataTypeFlag=columnDataType(inColumnH(14))
           if (columnDataTypeFlag .eq. 2) then
              sourceQF8 => int8Data(inColumnH(14))
           else
              sourceQF16 => int16Data(inColumnH(14))
           endif
         else
            inColumnFlag(14)=0
            allocate(sourceQF16(0:nSources-1))
         endif      

         if (hasColumn(inTableH, "CFLAG")) then
            inColumnH(15) = column(inTableH, "CFLAG", read)
            sourceCF => int8Data(inColumnH(15))
         else
            inColumnFlag(15)=0
            allocate(sourceCF(0:nSources-1))
         endif     

         

         if (hasColumn(inTableH, "SIGNIFICANCE")) then
            inColumnH(18) = column(inTableH, "SIGNIFICANCE", modify)
            significance => real32Data(inColumnH(18))
         else
            inColumnFlag(18)=0
            allocate(significance(0:nSources-1))
         endif    
         !if (hasColumn(inTableH, "SIGNIFICANCE")) then
         !   call  deleteColumnolumn(inTableH, "SIGNIFICANCE")
         !   inColumnFlag(18)=0
         !endif     


         if (hasColumn(inTableH, "CORR_RATE")) then
            inColumnH(19) = column(inTableH, "CORR_RATE", modify)
            corr_rate => real32Data(inColumnH(19))
         else
            inColumnFlag(19)=0
            allocate(corr_rate(0:nSources-1))
         endif     
         ! VNY 20.02.2006
         !if (hasColumn(inTableH, "CORR_RATE")) then
         !   call deleteColumn(inTableH, "CORR_RATE")
         !   inColumnFlag(19)=0
         !endif     

         if (hasColumn(inTableH, "CORR_RATE_ERR")) then
            inColumnH(20) = column(inTableH, "CORR_RATE_ERR", modify)
            corr_rate_err => real32Data(inColumnH(20))
         else
            inColumnFlag(20)=0
            allocate(corr_rate_err(0:nSources-1))
         endif     
         !if (hasColumn(inTableH, "CORR_RATE_ERR")) then
         !   call deleteColumn(inTableH, "CORR_RATE_ERR")
         !   inColumnFlag(20)=0
         !endif    

         if (hasColumn(inTableH, "SPB_COILOSS")) then
            inColumnH(21) = column(inTableH, "SPB_COILOSS", modify)
            spb_coiloss => real32Data(inColumnH(21))
         else
            inColumnFlag(21)=0
            allocate(spb_coiloss(0:nSources-1))
         endif      

         if (hasColumn(inTableH, "BK_COILOSS")) then
            inColumnH(22) = column(inTableH, "BK_COILOSS", modify)
            bk_coiloss => real32Data(inColumnH(22))
         else
            inColumnFlag(22)=0
            allocate(bk_coiloss(0:nSources-1))
         endif    

         if (hasColumn(inTableH, "NOPIXELS")) then
            inColumnH(23) = column(inTableH, "NOPIXELS", modify)
            nopixels => int32Data(inColumnH(23))
         else
            inColumnFlag(23)=0
            allocate(nopixels(0:nSources-1))
         endif      

         if (hasColumn(inTableH, "PSF1_CORR")) then
            inColumnH(16) = column(inTableH, "PSF1_CORR", modify)
            psf1Correction => real32Data(inColumnH(16))
         else
            inColumnFlag(16)=0
            allocate(psf1Correction(0:nSources-1))
            psf1Correction = 0.0 !nullValue
         endif

!----- VNY 2009-03-11 --- adding the variability data to the source list table
         if (hasColumn(inTableH, "CHI2")) then            
            inColumnH(24) = column(inTableH, "CHI2", modify)
            chi2 => real32Data(inColumnH(24))
         else            
            inColumnFlag(24)=0
            allocate(chi2(0:nSources-1))
            chi2=0.0
         endif
         

         if (hasColumn(inTableH, "CHI2_PROB")) then            
            inColumnH(25) = column(inTableH, "CHI2_PROB", modify)
            chi2prob => real32Data(inColumnH(25))
         else            
            inColumnFlag(25)=0
            allocate(chi2prob(0:nSources-1))
            chi2prob=0.0
         endif


         if (hasColumn(inTableH, "K-S")) then            
            inColumnH(26) = column(inTableH, "K-S", modify)
            ksStat => real32Data(inColumnH(26))
         else   
            inColumnFlag(26)=0
            allocate(ksStat(0:nSources-1))
            ksStat=0.0
         endif


         if (hasColumn(inTableH, "K-S_PROB")) then            
            inColumnH(27) = column(inTableH, "K-S_PROB", modify)
            ksStatProb => real32Data(inColumnH(27))
         else           
            inColumnFlag(27)=0
            allocate(ksStatProb(0:nSources-1))
            ksStatProb=0.0
         endif


         if (hasColumn(inTableH, "MAXDEV")) then            
            inColumnH(28) = column(inTableH, "MAXDEV", modify)
            maxDev => real32Data(inColumnH(28))
         else    
            inColumnFlag(28)=0
            allocate(maxDev(0:nSources-1))
            maxDev=0.0
         endif


!------------------------- 2009-03-11 ---------------

         allocate(psf2Correction(0:nSources-1))    
         psf2Correction = 0.0 !nullValue
   
         
         !******************************************************************
         ! Loop through all the sources  - 
         ! Correct each count-rate for the detector's 
         ! lifetime and ageing, then calculate the magnitude.
         !******************************************************************   
         nNegative = 0   
         !!iLoop=0
         !iLoop=srcNo-1
         !if (iLoop .lt. 0) iLoop=0
         !write(*,*) "L1056 iLoop=", iLoop, " srcNo=", srcNo
         !if (extended(iLoop) .eq. 0.0) then
         !   extend = .false.
         !else
         !   extend = .true.
         !end if
         ! set up a large temporary event-list of 2048x2048 elements
         nx = 2048
         ny = 2048
         tempSet = dataSet("tmp.fts", temp)
         tempTable = addTable(tempSet,"EVENTLIST", nx*ny)
         tempColX = addColumn(tempTable, "X", integer32)
         tempColY = addColumn(tempTable, "Y", integer32)
         xArray => int32data(tempColX)
         yArray => int32data(tempColY)
         npts = 0
         do i = 0, nx-1
            do j = 0, ny-1
               xArray(npts) = i
               yArray(npts) = j
               npts = npts + 1
            enddo
         enddo

         ! filter the temp event-list by the source region, which
         ! gives the number of pixels contained within the source area
         write(expression,'(a,a,a)') "region(",trim(sourceReg),", X,Y)"
         !
         call initFilter(expression)
         nPix = doFilter(tempTable) 
         sourceArea = real(nPix)

         write(errstr,'(a,f6.1)') "Source region area (pixels): ",sourceArea
         call message(AppMsg, 2)

         call release(tempColY)
         call release(tempColX)
         call release(tempTable)
         call release(tempSet)

         ! do the same for the background
         tempSet = dataSet("tmp.fts", temp)
         tempTable = addTable(tempSet,"EVENTLIST", nx*ny)
         tempColX = addColumn(tempTable, "X", integer32)
         tempColY = addColumn(tempTable, "Y", integer32)
         xArray => int32data(tempColX)
         yArray => int32data(tempColY)
         npts = 0
         do i = 0, nx-1
            do j = 0, ny-1
               xArray(npts) = i
               yArray(npts) = j
               npts = npts + 1
            enddo
         enddo
         write(expression,'(a,a,a)') "region(",trim(backReg),", X,Y)"
         call initFilter(expression)
         nPix = doFilter(tempTable)
         backArea = real(nPix)

         write(errstr,'(a,f9.1)') "Background region area (pixels): ",backArea
         call message(AppMsg, 2)

         call release(tempColY)
         call release(tempColX)
         call release(tempTable)
         call release(tempSet)

         srcNo = int32Attribute(sourceRegSet, "SRCNO")
         timeExposure = real64Attribute(sourceRegSet, "EXPOSURE") 
         iLoop=srcNo-1
         if (iLoop .lt. 0) iLoop=0
         if (iLoop .gt. nSources-1) iLoop=nSources-1
         if (extended(iLoop) .eq. 0.0) then
            extend = .false.
         else
            extend = .true.
         end if
         

         ! get a handle on the source rates file
         sourceSet = dataSet(TRIM(sourceRates), read)   

         !--- OSW origin
         window0 = stringAttribute(sourceSet, "DETSEC")
         read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
              ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
         !write(*,*) "L1548 x0=", windowX0, " x1=", windowX1, " y0=", windowY0, &
         !     " y1=", windowY1
         !write(*,*) "L1550 timeExposure=", timeExposure

         !***********************************************
         ! Use  the imaging file to calculate the background level
         !***********************************************
         ! First exclude the Fast window from consideration (by masking it) 
         !------------------------------------------------

          if (useImagingFile) then
             imAver=0.0
             nimAver=0.0
             is=windowX0-10
             ie=windowX1+10
             js=windowY0-10
             je=windowY1+10

             !write(*,*) "L1630 imwindowX0=", imwindowX0, " imwindowY0=", imwindowY0, &
             !     " useImagingCoordinates=", useImagingCoordinates

             do i=is,ie
                do j=js,je
                   i1=(i-imwindowX0)/imbinX
                   j1=(j-imwindowY0)/imbinX
                   if (i1 .ge. 0 .and. i1 .lt. imnx) then
                      if (j1 .ge. 0 .and. j1 .lt. imny) then
                         imImage2(i1,j1)=-2.0
                      endif
                   endif
                enddo
             enddo
             ! Compute the first iteration of the average of the rest of the imaging window
             do i=0, imnx-1
                do j=0,imny-1
                   if (imImage2(i,j) .gt. 0.0) then
                      imAver=imAver+imImage2(i,j)
                      nimAver=nimAver+1.0
                   endif
                enddo
             enddo
             if (nimAver .gt. 1.0) imAver=imAver/nimAver
             !write(*,*) "L1654 first iteration  imAver=", imAver
             ! Mask the bright pixels (>1.5*imAver) and compute the new imAver
             do i=0, imnx-1
                do j=0,imny-1
                   if (imImage2(i,j) .gt. 1.5* imAver) then ! 16.03.2010 was 2.0*imVver
                      imImage2(i,j)=-1
                   endif
                enddo
             enddo   
             imAver=0.0
             nimAver=0.0
             do i=0, imnx-1
                do j=0,imny-1
                   if (imImage2(i,j) .gt. 0.0) then
                      imAver=imAver+imImage2(i,j)
                      nimAver=nimAver+1.0
                   endif
                enddo
             enddo
             if (nimAver .gt. 1.0) imAver=imAver/nimAver
             !write(*,*) "L1674 second iteration imAver=", imAver

             !******************************************************************************
             ! Compute the background rate error at the standard deviation from the average
             !******************************************************************************
             imagingErr=0.0
             nimAver=0.0
             do i=0, imnx-1
                do j=0,imny-1
                   if (imImage2(i,j) .gt. 0.0) then
                      imagingErr=imagingErr+(imImage2(i,j)-imAver)*(imImage2(i,j)-imAver)
                      nimAver=nimAver+1.0
                   endif
                enddo
             enddo

             if (nimAver .gt. 1.0) then
                imagingErr=imagingErr/(nimAver-1.0)
             endif
             imagingErr=sqrt(imagingErr)

             if (nimAver .gt. 1.0) then
                imagingErr=imagingErr/sqrt(nimAver)
             endif


             ! Take into account binning of the image (to use further unbinned pixels)
             rateAux=real(imbinX)*real(imbinX)
             imAver=imAver/rateAux
             imagingErr=imagingErr/rateAux
 
             if (imAver .gt. 0.0) then
                imagingBkg=imAver
                imagingBkgOk=.true.
                if (timeExposure .gt. 1.0) then
                   imagingBkgRate=imagingBkg/timeExposure
                   imagingBkgRateErr=imagingErr/timeExposure
                   imagingErr=imagingErr/timeExposure                   
                endif
                ! Convert the imaging background count rates (per pixel) to 
                ! the number of background counts during the time binning interval
                imagingBkgBin=imagingBkgRate*timeDel
                imagingBkgBinErr=imagingBkgRateErr*sqrt(timeDel)

                !imagingBkgRate=imagingBkgRate*timeDel
                !imagingErr=imagingErr*sqrt(timeDel)
                
             endif
             !write(*,*) "L1719 imagingBkgRate=", imagingBkgRate, " imagingErr=", &
             !     imagingErr, " imagingBkgRateErr=", imagingBkgRateErr

             !write(*,*) "L1722 from the smoothed image averageBkg=", averageBkg, " averageSigma=", averageSigma

             !!!!!!----------------------
             !write(*,*) "L1626 creating testimage77.fit"
             !outSetCheck="testimage77.fit"
             !outDimCheck(1)=(imWindowX1-imWindowX0+1)/imbinX
             !outDimCheck(2)=(imWindowY1-imWindowY0+1)/imbinY
             !write(*,*) "L1630 imWindowX1=", imWindowX1, " imWindowX0=", imWindowX0
             !write(*,*) "L1631 outDimCheck(1)=",outDimCheck(1)
             !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
             !
             !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
             !     real32, outDimCheck, "Counts", "IMAGE")!  
             !      
             !outOSWcheckP => real32Array2Data(outArrayCheckH)
             ! 
             !outOSWcheckP = imImage2 
             ! 
             ! Release handles on data.
             !call release(outArrayCheckH)        
             !call release(outDataSetCheckH) 
             ! 
             !write(*,*) "L1644 Test image 77 saved:", trim(outSetCheck)!
             !!!!!!--------------------------------
             !deallocate (imImage2, stat=allocErr)
             !deallocate (qualArray, stat=allocErr)
       
          endif ! useImagingFile
          !*********************************************
          srcRateExists=.false.
          if (hasBlock(sourceSet, "RATE")) then
             sourceTable = table(sourceSet,"RATE")
             nBins = numberOfRows(sourceTable) ! Number of points in the light curve
             !            write(*,*) "L1589 nBins=", nBins
             !---Protection against large values of nBins
             if (nBins .gt. 10000) nBins=9999
             if (nBins .lt. 0) nBins=-nBins
             
             !************************************
             write (errstr, *) "Source RATE table size in the SCRATE file:  ", nBins
             call message(AppMsg, 2)
             
             !---protection against possible zero size of the source rates table
             if (nBins .gt. 0) then 
                srcRateExists=.true.
             else
                !---additional protection against zero size of source rates table
                call warning("srcRATE", "Source RATE table is of zero size (SCRATE  file)")
                !---Stop if nBins is less or equals to zero
                if (nSourceRateRows .gt. 0) then
                   nBins=nSourceRateRows
                   call warning("corrSrcRate", " Correcting source RATE table length")
                   srcRateExists=.true.
                else
                   call fatal("dataERR", " Impossible to get the correct source rates table length")
                endif
             endif
             
             rateSourceCol = column(sourceTable,"COUNTS", read)
             iSourceArray => int32data(rateSourceCol)
             
             !--- reading the source aperture radius detected by OMREGION
             regionTable = table(sourceSet, "REGION")
             sourceShapeCol = column(regionTable, "SHAPE", read)
             sourceRadiusCol = column(regionTable, "R", read)
             sourceXPosCol = column(regionTable, "X", read)
             sourceYPosCol = column(regionTable, "Y", read)
             nRegion = numberOfRows(regionTable)

             radiusDetectedArray => real32data(sourceRadiusCol)
             sourceXPosArray => real32data(sourceXPosCol)
             sourceYPosArray => real32data(sourceYPosCol)
             
             !---fixing the number of avoiding sources 
             nAvoidingSources=0
             if (nRegion .gt. 1) then 
                nAvoidingSources=nRegion-1
             end if
             
             ! Limiting the number of avoiding sources
             if (nAvoidingSources .gt. 10) then 
                nAvoidingSources=10
             end if
             
             ! Preparing an array of avoiding source coordinates relative to
             ! the centre of the target source
             do i=1,nAvoidingSources
                k=i*50 ! each source is a sub-array of 50 elements
                avoidX(i)=sourceXPosArray(k)-sourceXPosArray(0)
                avoidY(i)=sourceYPosArray(k)-sourceYPosArray(0)
                avoidR(i)=radiusDetectedArray(k)
             end do
             !write(*,*) "L1355 nAvoidingSources=", nAvoidingSources
             !do i=1, nAvoidingSources
             !   write(*,*) "L1213 i=", i, " avoidX=", avoidX(i), " avoidY=", &
             !        avoidY(i), " avoidR=", avoidR(i)
             !enddo
             ! Prepare comment on the source region
             sourceShapeComment=" "
             if (nRegion .gt. 0) then
                sourceShape = stringCell(sourceShapeCol,0)
                ! Calculate the source shape word length              
                wordlength: do i=1,20
                   if (ichar(sourceShape(i:i)) .eq. 0) then
                      exit wordlength
                   end if
                end do wordlength
                
                write(sourceShapeComment, '(a, a, a, f6.1, a, f6.1, a, f5.1, a)') "source ", &
                     sourceShape(1:i-1), &
                     "(", sourceXPosArray(0),",", sourceYPosArray(0), ",", &
                     radiusDetectedArray(0), ")"            
             end if
             
             allocate(rateSourceArray(0:nBins-1))
             rateSourceArray=0.0               
             rateSourceArray = real(iSourceArray)
             allocate(timeSource(0:nBins))
             timeStart = real64Attribute(sourceTable, "TSTART")
             
             timeEnd = real64Attribute(sourceTable, "TSTOP")
             !write(*,*) "L1890 nBins=", nBins, " nBack=", nBack
             do i = 0, nBins
                
                timeSource(i) = timeStart + dble(i-1)*timeDel
                !if (i .lt. 6 .or. i .gt. nBins-6) then
                !   write(*,*) "L1893 i=", i, " timeSource=", timeSource(i)
                !endif
             enddo
             ! timeEnd corresponds to the index i=nBins 

             !write(*,*) "L1894 timeStart=", timeStart, " timeEnd=", timeEnd, " timeSource(nBins)=", timeSource(nBins), &
             !     " diff=", timeEnd-timeSource(nBins)
             lastTimeDel=timeEnd-timeSource(nBins)
            
             ! get a handle on the background rates file
             backSet = dataSet(TRIM(backRates), read)   
             
             !---Protection against the non-existence of the block RATE: bkgRateExists
             bkgRateExists = .false.
             if ( hasBlock(backSet, "RATE")) then
                bkgRateExists = .true.
             else
                call warning("bkgRATE", "Block RATE does not exist in the background rates file")
             endif
             
             medianBkgError=1.0
             medianBkg=0.0
             backMax=0
             nBack = 0
             nRegion = 0
             
             if (bkgRateExists) then
                
                backTable = table(backSet,"RATE")
                nBack = numberOfRows(backTable)
                !write(*,*) "L1925 nBack=", nBack

                ! --- protection against too large values of nBack
                if (nBack .gt. 10000) nBack=9999
                
                if (nBack .lt. 0) nBack=-nBack
                
                rateBackCol = column(backTable,"COUNTS", read)
                iBackArray => int32data(rateBackCol)
                
                ! Calculating the maximal background rate
                backMax=0
                do i=0, nBack-1
                   if (iBackArray(i) .gt. backMax) backMax=iBackArray(i)
                enddo
                
                ! Build the background distribution
                allocate(backDist(1:backMax+1))
                backDist=0
                do i=0,nBack-1
                   do j=0,backMax
                      if (iBackArray(i) .eq. j) backDist(j+1)=backDist(j+1)+1
                   enddo
                enddo
                
                ! Estimate the background median
                medianFound=.false.
                backSum=0.0
                backSumFound=0.0
                j=0
                do i=1,backMax+1
                   backSum=backSum+real(backDist(i))
                   if (real(nBack)/2.0 .le. backSum .and. .not. medianFound) then
                      medianFound=.true.
                      j=i
                      backSumFound=backSum
                   endif
                enddo
                medianBkg=1.0
                if (backSumFound .gt. 0.0) medianBkg=real(j)*real(nBack)/2.0/backSumFound
                medianBkgError=sqrt(medianBkg)
                deallocate(backDist)
             endif ! bkgRateExists

             !--- reading the background aperture radius detected by OMREGION
             !--- using backRegSet instead of backSet (in case of the absense 
             !    of the latter)
             backRegionTable = table(backRegSet, "REGION")
             backShapeCol = column(backRegionTable, "SHAPE", read)
             backRadiusCol = column(backRegionTable, "R", read)
             backXPosCol = column(backRegionTable, "X", read)
             backYPosCol = column(backRegionTable, "Y", read)
             nBackRegion = numberOfRows(backRegionTable)
             backRadiusArray => real32data(backRadiusCol)
             backXPosArray => real32data(backXPosCol)
             backYPosArray => real32data(backYPosCol)
             
             nBackAvoid=0
             do ia=1,nBackRegion
                backShape = stringCell(backShapeCol,ia-1)
                if (TRIM(backShape) .eq. "!CIRCLE") then
                   nBackAvoid=nBackAvoid+1
                   if (nBackAvoid .le. 10) then
                      xBackAvoid(nBackAvoid)= backXPosArray((ia-1)*50)
                      yBackAvoid(nBackAvoid)=backYPosArray((ia-1)*50)
                      rBackAvoid(nBackAvoid)=backRadiusArray((ia-1)*50)
                   endif
                endif
             end do
            if (nBackAvoid .gt. 10) nBackAvoid=10

            ! Protection agains an empty background rates table
            if (nBack .lt. 1) then
               
               nBack=0
               bkgRateExists = .false.
               call warning ("RATE", "Input background rates table is empty")
            endif

            ! Prepare a comment on the background region
            backShapeComment=" "
            if (nBackRegion .gt. 0) then
               backShape = stringCell(backShapeCol,0)
               ! Calculate the background shape word length              
               wordlength2: do i=1,20
                  if (ichar(backShape(i:i)) .eq. 0) then
                     exit wordlength2
                  end if
               end do wordlength2
 
               write(backShapeComment, '(a, a, a, f6.1, a, f6.1, a, f5.1, a, f5.1, a)') &
                    "background ", backShape(1:i-1), "(", backXPosArray(0),",", backYPosArray(0), &
                    ",", backRadiusArray(0), ",", backRadiusArray(1), ")"        
            end if
            mod8flag=0
            if (hasAttribute(sourceSet, "MOD8FLAG")) then
               mod8flag=int32attribute(sourceSet, "MOD8FLAG")
            endif

            bkgRateId = .true.
            if (bkgRateExists) then
               if (nBack .gt. 2) then
                  allocate(rateBackArray(0:nBack-1))
                  rateBackArray = real(iBackArray)
                  allocate(timeBack(0:nBack))
               else
                  allocate(rateBackArray(0:nBins-1))
                  rateBackArray=0.0
                  allocate(timeBack(0:nBins))
               endif
                  bkgTimeStart = real64Attribute(backTable, "TSTART")
                  bkgTimeDel = real64Attribute(backTable, "TIMEDEL")
                  bkgTimeEnd = real64Attribute(backTable, "TSTOP")

                  !write(*,*) "L2039 bkgTimeStart=",  bkgTimeStart , " bkgTimeDel=", bkgTimeDel, &
                  !     " bkgTimeEnd=", bkgTimeEnd

               if (stringAttribute(backSet,"EXP_ID") /= &
                    stringAttribute(sourceSet,"EXP_ID") ) then
                  bkgRateId = .false.
                  call warning("diffExp","Source and background rates files from different exposures")
               endif

               if (stringAttribute(backRegSet,"EXP_ID") /= &
                    stringAttribute(backSet, "EXP_ID")) then
                  bkgRateId= .false.
                  call warning("diffExp", "Background rates and region files have different EXP_IDs")
               endif

               if (stringAttribute(sourceRegSet,"EXP_ID") /= &
                    stringAttribute(sourceSet, "EXP_ID")) then
                  bkgRateId = .false.
                  call warning("diffExp", "Source rates and region files have different EXP_IDs")
               endif
               !write(*,*) "L2050 nBack=", nBack, " nBins=", nBin
               !write(*,*) "L2051 bkgTimeDel=", bkgTimeDel, " bkgTimeEnd=", bkgTimeEnd, &
               !     " timeBack(nBack)=", timeBack(nBack) 
               if (nBack .gt. 2) then
                  do i = 0, nBack
                     timeBack(i) = bkgTimeStart + dble(i-1)*bkgTimeDel
                     !if (i .lt. 5 .or. i .gt. nBack-6) then
                     !   write(*,*) "L2066 i=", i, " timeBack=", timeBack(i)
                     !endif
                  enddo
                  !write(*,*) "---"
                  
                       
                  bkgLastTimeDel=bkgTimeEnd-timeBack(nBack)
                  !write(*,*) "L2070 bkgTimeStart=", bkgTimeStart, " bkgTimeEnd=", bkgTimeEnd, &
                  !     " bkgLastTimeDel=", bkgLastTimeDel
               else
                  do i = 0, nBins
                     timeBack(i) = bkgTimeStart + dble(i-1)*timeDel
                  enddo
                  bkgLastTimeDel=bkgTimeEnd-timeBack(nBins)
               endif

               ! the two time lines might be very different (if there is few or little background counts)
               ! the master time line is taken from the source 
               timeStart=timeSource(1)
               timeEnd=timeSource(nBins)
               !write(*,*) "L2069 timeStart=", timeStart, " timeEnd=", timeEnd
            endif ! bkgRateExists
            allocate(timeOut(0:nBins-1))
            do i = 0, nBins-1
               !--- Shifting times by a half of the timeDel-interval ahead
               timeOut(i) = timeStart + dble(i)*timeDel + timeDel/2.0
            enddo
!            

            allocate(sourceArrayError(0:nBins-1))
            if (nBack .gt. 2) then
               allocate(backArrayError(0:nBack-1))
            else
               allocate(backArrayError(0:nBins-1))
            endif
            sourceArrayError=0.0
            backArrayError=0.0

            ! create the output dataset
            outSet = dataSet(outFile, create)

            call copyAllAttributes(attributable(sourceSet), attributable(outSet))
            call setAttribute(outSet, "CONTENT", "OM OSW SOURCE TIMESERIES")
            call setAttribute(outSet, "SRCNO", srcNo, "n", "Source number")

            
            outTable = addTable(outSet, "RATE", nBins, "Table with the source and background count rates")

            !---SRR (19/12/01) changed TIMEZERO value to 0.d0 (this should reflect the offset from the 
            !   TSTART parameter which gets propagated into the headers from obtstart I think). 
            call copyAllAttributes(attributable(sourceTable), attributable(outTable))
            call setAttribute(outTable, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
            call setAttribute(outTable, "TIMEDEL", timeDel, "s", "time binning interval")

            !--- Adding time column to the output table
            timeSourceOutCol = addColumn(outTable, "TIME", real64, "S", columnDim, " Time ralative to TIMEZERO")
            timeSourceOutArray => real64data(timeSourceOutCol)

            rateSourceOutCol = addColumn(outTable,"RATE", real32, "COUNTS/S", columnDim, "source counts rate")
            rateSourceOutArray => real32data(rateSourceOutCol)

            rateSourceErrOutCol = addColumn(outTable,"ERROR", real32,"COUNTS/S", &
                 columnDim, "source counts rate error")
            rateSourceErrOutArray => real32data(rateSourceErrOutCol)

            rateBackOutCol = addColumn(outTable,"BACKV", real32, "COUNTS/S", &
                 columnDim, " background counts rate")
            rateBackOutArray => real32data(rateBackOutCol)

            rateBackErrOutCol = addColumn(outTable,"BACKE", real32, "COUNTS/S", &
                 columnDim, " background counts rate error")
            rateBackErrOutArray => real32data(rateBackErrOutCol)

            ! Using medianBkgError as the estimate for the error bars when 
            ! source or background is zero.
            sourceArrayError=medianBkgError
            if (srcRateExists) then
               do i=0,nBins-1
                  if (rateSourceArray(i) .gt. 0.0) sourceArrayError(i) = sqrt(rateSourceArray(i))
               enddo
            endif

            ! If there is no background counts at all, the background error bars are set to 0
            if (backMax .gt. 0) then
               backArrayError=medianBkgError
            else
               backArrayError=0.0
            endif

            if (bkgRateExists) then
               if (nBack .gt. 2) then
                  do i=0,nBack-1
                     if (rateBackArray(i) .gt. 0.0) backArrayError(i) = &
                          max(backArrayError(i), sqrt(rateBackArray(i)))
                  enddo
               else
                  do i=0,nBins-1
                     if (rateBackArray(i) .gt. 0.0) backArrayError(i) = &
                          max(backArrayError(i), sqrt(rateBackArray(i)))
                  enddo
               endif
            endif ! bkgRateExists

            !--- Applying  the PSF-mapping function
            !******************************************************************************

            ! xPos0 and yPos0 are the coordinates measured from the CCD origin
            xPos0=dble(sourceXPosArray(0))
            yPos0=dble(sourceYPosArray(0))


            !*****************************************************************
            ! Checking the possibility to calculate background in the same way
            ! as it is done in omichain (imaging photometry)
            !*****************************************************************
            !write(*,*) "L2195 windowX0=", windowX0, " windowY0=", windowY0, &
            !     " xPos0=", xPos0, " yPos0=", yPos0, " timeExposure=", timeExposure
            imxPos=10.0
            imyPos=10.0

            !write(*,*) "L2138 sourceXcoord(0)=", sourceXcoord(0), " sourceYcoord(0)=", sourceYcoord(0), &
            !     " rate(0)=", rate(0), " significance(0)=", significance(0)
            apnx=int(imwindowX1-imwindowX0+1, int16)
            apny=int(imwindowY1-imwindowY0+1, int16)
            if (apnx .lt. 40 .or. apny .lt. 40) then
               ! the image is too small: setting the flag useImagingFile to false
               !write(*,*) "L2214 the image is too small: skip using the imaging data "
               write(errstr,'(a)') "The background image is too small (ignoring) "
               call message(AppMsg, 2)
               useImagingFile=.false.
            endif
            if (useImagingFile) then
               nullify(sourceData)
               allocate(sourceData( 1 : 1))

               !sourceData(1)%xPos=xPos0
               !sourceData(1)%yPos=yPos0
               sourceData(1)%exposure=timeExposure
               !write(*,*) "L2209 imwindowX0=", imwindowX0, " imwindowX1=", imwindowX1
               !write(*,*) "L2210 imwindowY0=", imwindowY0, " imwindowY1=", imwindowy1
               apertureRadius = real(12 / imbinX) 
               !apertureRadius = real(6 / imbinX) 

               !*****************************************************************************
               ! 2014-06-01 fixing the bug reported in SPR-7206 (taking into account the binning factor)
               !******************************************************************************
               apnx=int(imwindowX1-imwindowX0+1, int16)
               apny=int(imwindowY1-imwindowY0+1, int16)
               apnx=apnx/imbinX
               apny=apny/imbinY
               
               nPSources=1
               !write(*, *) "L2215 apnx=", apnx, " apny=", apny

               !*****************************************************************************
               ! 2014-06-01 fixing the bug reported in SPR-7206 (taking into account the binning factor)
               !******************************************************************************
               !imxPos=xPos0-imwindowX0+1.0
               !imyPos=yPos0-imwindowY0+1.0
               imxPos=(xPos0-imwindowX0+1.0)/imbinX
               imyPos=(yPos0-imwindowY0+1.0)/imbinY

               !write(*,*) "L2219 imxPos=", imxPos, " imyPos=", imyPos
               sourceData(1)%xPos=imxPos
               sourceData(1)%yPos=imyPos
               sourceData(1)%iMin=max(0,int(imxPos-apertureRadius/4.0+0.5))
               sourceData(1)%iMax=min(apnx-1,int(imxPos+apertureRadius/4.0+0.5)) 
               sourceData(1)%jMin=max(0,int(imyPos-apertureRadius/4.0+0.5))
               sourceData(1)%jMax=min(apny-1,int(imyPos+apertureRadius/4.0+0.5)) 
               sourceData(1)%rawCountRate=rate(0)
               sourceData(1)%significance=significance(0)

               imiPos=int(imxPos)
               imjPos=int(imyPos)
               !write(*,*) "L2231 iMin=",sourceData(1)%iMin, " iMax=",  sourceData(1)%iMax
               !write(*,*) "L2232 jMin=",sourceData(1)%jMin, " jMax=",  sourceData(1)%jMax
               do i=sourceData(1)%iMin, sourceData(1)%iMax
                  do j=sourceData(1)%jMin, sourceData(1)%jMax
                     !levelImage(imiPos, imjPos)=1
                     !write(*,*) "L2172 i=", i, " j=", j, " imImage3=",imImage3(i,j), " bkg=",backgroundImage(i,j)  
                     levelImage(i, j)=1
                  enddo
               enddo

               !write(*,*) "L2126 implementing aperture photometry at the imaging data set apnx=", apnx, &
               !     " apny=", apny, " nPSources=", nPSources
               write(errstr,'(a)') "implementing aperture photometry at the imaging data set "
               call message(AppMsg, 2)

               !do i=1, npSources
               !   write(*,*) "L2156 before  x=", sourceData(i)%xPos, " y=", sourceData(i)%yPos, &
               !        " exposure=", sourceData(i)%exposure, " signif=", sourceData(i)%significance, &
               !        " raw=", sourceData(i)%rawCountRate, " corr=", sourceData(i)%corrCountRate, &
               !        " mag=", sourceData(i)%magnitude, " rawbkg=", sourceData(i)%rawBackgroundCountRate, &
               !        " apRadius=", sourceData(i)%apertureRadius, " innerBkg=", sourceData(i)%innerBckRadius, &
               !        " outerBkg=", sourceData(i)%outerBckRadius, " iMin=", sourceData(i)%iMin, &
               !        " iMax=", sourceData(i)%iMax, " jMin=", sourceData(i)%jMin, " jMax=", sourceData(i)%jMax
               !enddo

               call AperturePhotometry(apnx, apny, imImage3, backgroundImage, levelImage, qualArray,  & 
                    apertureRadius, nPSources, sourceData)

               !do i=1, npSources
               !   write(*,*) "L2169 after x=", sourceData(i)%xPos, " y=", sourceData(i)%yPos, &
               !        " exposure=", sourceData(i)%exposure, " signif=", sourceData(i)%significance, &
               !        " raw=", sourceData(i)%rawCountRate, " corr=", sourceData(i)%corrCountRate, &
               !        " mag=", sourceData(i)%magnitude, " rawbkg=", sourceData(i)%rawBackgroundCountRate, &
               !        " apRadius=", sourceData(i)%apertureRadius, " innerBkg=", sourceData(i)%innerBckRadius, &
               !        " outerBkg=", sourceData(i)%outerBckRadius, " iMin=", sourceData(i)%iMin, &
               !        " iMax=", sourceData(i)%iMax, " jMin=", sourceData(i)%jMin, " jMax=", sourceData(i)%jMax
               !enddo

               call AperturePhotometry(sourceData, int(npSources), imImage3, levelImage, &
                    qualArray, real(6 / imbinX), 10.0) ! was 10

               !do i=1, npSources
               !   write(*,*) "L2299 after2 x=", sourceData(i)%xPos, " y=", sourceData(i)%yPos, &
               !        " exposure=", sourceData(i)%exposure, " signif=", sourceData(i)%significance, &
               !        " raw=", sourceData(i)%rawCountRate, " corr=", sourceData(i)%corrCountRate, &
               !        " mag=", sourceData(i)%magnitude, " rawbkg=", sourceData(i)%rawBackgroundCountRate, &
               !        " CLcorrBkg=", sourceData(i)%corrBackgroundCountRate, " err=", sourceData(i)%corrBackgroundCountRateErr, &
               !        " apRadius=", sourceData(i)%apertureRadius, " innerBkg=", sourceData(i)%innerBckRadius, &
               !        " outerBkg=", sourceData(i)%outerBckRadius, " iMin=", sourceData(i)%iMin, &
               !        " iMax=", sourceData(i)%iMax, " jMin=", sourceData(i)%jMin, " jMax=", sourceData(i)%jMax
               !enddo

               im2bkg=sourceData(1)%rawBackgroundCountRate
               im2bkgErr=sourceData(1)%rawBackGroundCountRateErr
               im2Aperture=sourceData(1)%apertureRadius

               deallocate(sourceData)
               !write(*,*) "L2119 deallocated sourceData"

               deallocate (imImage2, stat=allocErr)
               deallocate (imImage3, stat=allocErr)
               deallocate (qualArray, stat=allocErr)
            endif ! useImagingFile

            !write(*,*) "L2250 im2bkg=",im2bkg, " im2bkgErr=", im2bkgErr, " im2Aperture=", im2Aperture

            !---Calculation of the psf-shift with respect to the ccd-grid
            psfShiftX=xPos0-real(int(xPos0))-0.5
            psfShiftY=yPos0-real(int(yPos0))-0.5
            xRes=1
            yRes=1
            radiusDetected=radiusDetectedArray(0)
            !write(*,*) "L2039 radiusDetected=", radiusDetected, " photometryApertureRAdius=", &
            !     photometryApertureRadius, " backR1R2=", backRadiusArray(0), backRadiusArray(1)
            
            !*********************************************************************
            ! 2009-10-11 VNY: In the case of a modulo-8 corrupted image
            ! the extraction of counts was made by evselect from the entire emage,
            ! but we have to assume that these extracted photons belong to a smaller
            ! region, say, of 6-pixel radius. Otherwise, the coincidence-loss correction
            ! will not work properly
            !***********************************************************************
            !
            
            if (mod8corrupted .gt. 0) then
               radiusDetected=12.0
            endif
            !
            !********************************************************************** 

            
            ! selection of the radius for PSF-mapping
            psfRadius=max(photometryApertureRadius,radiusDetected, real(windowX1-windowX0+1), &
                 real(windowY1-windowY0+1))
            psfRadius=max(psfRadius, backRadiusArray(0), backRadiusArray(1))
            !write(*,*) "L2065 psfRadius=", psfRadius

            detectedArea = PI * (radiusDetected) * (radiusDetected)
            !write(*,*) "L2284 detectedArea=", detectedArea, " radiusDetected=", radiusDetected
            !---Taking into account the 0.5-pixel shift of the coordinate system
            !dxLeft=xPos0-windowX0 - 0.5 
            !dxRight=windowX1-xPos0 - 0.5
            !dyDown=yPos0-windowY0 - 0.5
            !dyUp=windowY1-yPos0 - 0.5

            ! The 0.5-pixel shifts are corrected 2010-03-16 (SPR-6602)
            !dxLeft=xPos0-windowX0 - 0.5 
            !dxRight=windowX1-xPos0 + 0.5
            !dyDown=yPos0-windowY0 - 0.5
            !dyUp=windowY1-yPos0 + 0.5

            ! The 0.5-pixel shifts for the smaller source-to-edge distance
            dxLeft=xPos0-windowX0 
            dxRight=windowX1-xPos0
            dyDown=yPos0-windowY0
            dyUp=windowY1-yPos0
            if (dxLeft .lt. dxRight) then
               dxLeft=dxLeft-0.25
               dxRight=dxRight+1.25
            else
               dxLeft=dxLeft+1.25
               dxRight=dxRight-0.25
            endif
            
            if (dyDown .lt. dyUp) then
               dyDown=dyDown-0.25
               dyUp=dyUp+1.25
            else
               dyDown=dyDown+1.25
               dyUp=dyUp-0.25
            endif

            !*********************************************************************
            ! 2011-02-18 more correct calculation of dxLeft and dxRight fractions
            !**********************************************************************
            !dxLeft=xPos0-windowX0
            !dxRight=windowX1-xPos0
            !dyDown=yPos0-windowY0
            !dyUp=windowY1-yPos0

            !dxLeft=dxLeft+0.5
            !dxRight=dxRight+0.5

            !dyDown=dyDown+0.5
            !dyUp=dyUp+0.5




            !write(*,*) "L2093 xPos0=", xPos0, " windowX0=", windowX0, " windowX1=", windowX1
            !write(*,*) "L2094 dxLeft=", dxLeft, " dxRight=", dxRight 
            !write(*,*) "L2095 yPos0=", yPos0, " windowY0=", windowY0, " windowY1=", windowY1
            !write(*,*) "L2096 dyDown=", dyDown, " dyUp=", dyUp 
            call CAL_omGetPSFmap(xPos0,yPos0, int(psfRadius), int(psfRadius), int(psfRadius), &
                 int(psfRadius), xRes, yRes, dble(countFrameRatio), psf)

            !write(*,*) "L2083 Point-spread function:"
            !write(*,*) "----------------------------"
            !iy=psfRadius
            !do ix=0,2*psfRadius
            !   write(*,*) "{", ix-psfRadius, ", ", psf(ix,iy), "}," 
            !enddo
!+++++++
!--- VNY 28.08.2001 Temporarely using a wider PSF : sqrt(((radius-6)**2)/100)
!              do iy=0,2*psfRadius
!                 do ix=0,2*psfRadius
!                    xc=real(ix-psfRadius)
!                    yc=real(iy-psfRadius)
!                    xyDistance=sqrt(xc*xc+yc*yc)
!                    if (abs(xyDistance) .le. 6.0) then 
!                       psf(ix,iy)=psf(ix,iy)+sqrt((xyDistance-6.0)*(xyDistance-6.0)/100.0)
!                    end if
!                 end do
!              end do
!--- VNY 28.08.2001 When the PSF will be corrected in CAL_omGetPSFmap, the above code should be removed
!++++++
!+++++++
!--- VNY 16.03.2010 Test flattening the PSF
!            ix=psfRadius+4.0
!            iy=psfRadius
!            rateAux=psf(ix,iy)
!            write(*,*) "L2096 psf(2)=", rateAux
!              do iy=0,2*psfRadius
!                 do ix=0,2*psfRadius
!                    xc=real(ix-psfRadius)
!                    yc=real(iy-psfRadius)
!                    xyDistance=sqrt(xc*xc+yc*yc)
!                    if (abs(xyDistance) .le. 4.0) then 
!                       psf(ix,iy)=rateAux 
!                    end if
!                 end do
!              end do
!--- 
!++++++

            ! Calculation of the PSF fractions within different areas
            sumIn=0.  ! for PSF integrated over the area of the source inside the window 
            sumOut=0. ! the same, but outside the window
            sumIn2=0. ! PSF integrated within CALapertureRadius
            sumOut2=0. 
            sumIn3=0.
            sumOut3=0.
            sumBkg=0.0
            testSumBkg=0.0
            testSumSrc=0.0

            areaIn=0 ! Geometrical areas 
            areaOut=0
            areaIn2=0
            areaOut2=0
            areaIn3=0
            areaOut3=0
            areaBkg=0
               
            ! Calculate the number of pixels within the background region
            ! that are not masked by the avoiding circle
            bkgExtractionArea=0
            srcExtractionArea=0
            do ix=windowX0, windowX1
               do iy=windowY0,windowY1
                  outOfAvoid=.true.
                  if (nBackAvoid .gt. 0) then
                     do ia=1,nBackAvoid
                        dist=(real(ix)-xBackAvoid(ia))*(real(ix)-xBackAvoid(ia))+ &
                             (real(iy)-yBackAvoid(ia))*(real(iy)-yBackAvoid(ia))
                        dist=sqrt(dist)
                        if (dist .lt. rBackAvoid(ia)) then
                           outOfAvoid=.false.
                        endif
                     enddo
                  endif
                  if (outOfAvoid) then
                     dist=(real(ix)-backXPosArray(0))*(real(ix)-backXPosArray(0)) +  & 
                          (real(iy)-backYPosArray(0))*(real(iy)-backYPosArray(0))
                     dist=sqrt(dist)
                     !if (dist .ge. backRadiusArray(0) .and. dist .le. backRadiusArray(1)) then
                     if (dist .ge. backRadiusArray(0) .and. dist .lt. backRadiusArray(1)) then
                        bkgExtractionArea=bkgExtractionArea+1
                     endif
                     !---I the distance within the detection radius then use a 
                     !   different option for the source area
                     if (dist .le. radiusDetected) then
                        srcExtractionArea=srcExtractionArea+1
                     endif
                  endif
               enddo
            enddo
            !write(*,*) "L2433 srcExtractionArea=", srcExtractionArea, &
            !     " bkgExtractionArea=", bkgExtractionArea
            !write(*,*) "L2145 nAvoidingSources=", nAvoidingSources, " nBackAvoid=", nBackAvoid
            !write(*,*) "L2146 psfShiftX=", psfShiftX, " psfShiftY=", psfShiftY
            !write(*,*) "L2147 dxRight=", dxRight, " dxLeft=", dxLeft, " psfRadius=", psfRadius
            ! Calculate psf-fractions inside and outside OSW 
            do iy=0,2*int(psfRadius)
               do ix=0,2*int(psfRadius)
                  !---Checking whether ix and iy are inside of the avoiding circle
                  outsideOfAvoidingCircle = .true.
                  if (nAvoidingSources .gt. 0) then
                     ! Calculate the distance from the avoiding centre
                     do i = 1, nAvoidingSources
                        !---Applying the psf-shift with respect to the ccd-grid
                        xc=real(ix)-psfRadius-avoidX(i)+psfShiftX
                        yc=real(iy)-psfRadius-avoidY(i)+psfShiftY
                        xyDistance=sqrt(xc*xc+yc*yc)
                        if (xyDistance .lt. avoidR(i)) then
                           outsideOfAvoidingCircle = .false.                            
                        end if
                     end do
                  end if
                  if (nBackAvoid .gt. 0) then
                     do i=1,nBackAvoid
                        dist=(real(ix)-xBackAvoid(i))*(real(ix)-xBackAvoid(i))+ &
                             (real(iy)-yBackAvoid(i))*(real(iy)-yBackAvoid(i))
                        dist=sqrt(dist)
                        if (dist .lt. rBackAvoid(i)) then
                           outsideOfAvoidingCircle=.false.
                        endif
                     enddo
                  endif ! nBackAvoid .gt. 0
                  !---Applying the psf-shift with respect to the ccd-grid
                  xc=real(ix)-psfRadius+psfShiftX
                  yc=real(iy)-psfRadius+psfShiftY
                  xyDistance=xc*xc+yc*yc
                  xyDistance=sqrt(xyDistance)
                    
                  ! calculation of the psf fractions within  
                  ! and outside the OM science window
                  if (xyDistance .le. radiusDetected) then     
                     if (xc .le. dxRight .and. xc .ge. -dxLeft &
                          !  An additional condition is necessary here in order to take into account 
                          !  the counts falling into the avoiding area of the neighbour source 
                          .and. yc .le. dyUp .and. yc .ge. -dyDown &
                          .and. outsideOfAvoidingCircle) then
                        sumIn=sumIn+psf(ix,iy)
                        areaIn=areaIn+1
                     else                          
                        sumOut=sumOut+psf(ix,iy)
                        areaOut=areaOut+1
                     end if
                  else
                     if ( xc .le. dxRight .and. xc .ge. -dxLeft &
                          .and. yc .le. dyUp .and. yc .ge. -dyDown &
                          .and. xyDistance .ge. backRadiusArray(0) & 
                          .and. xyDistance .le. backRadiusArray(1) &
                          .and. outsideOfAvoidingCircle) then
                        sumBkg=sumBkg+psf(ix,iy)
                        areaBkg=areaBkg+1
                     endif
                  endif

 
                  ! calculation of the ratio between the cal-radius fraction of psf
                  ! and its photometry-radius fraction (sumIn2 and sumOut2)
                  if (xyDistance .lt. calApertureRadius) then
                     sumIn2=sumIn2+psf(ix,iy) ! For the PHOT to CAL ratio
                     areaIn2=areaIn2+1

                     ! Calculation of the ratio between the detected radius fraction of PSF
                     ! and CAL-radius fraction (sumIn3 and sumOut3)
                     if (xyDistance .le. radiusDetected) then 
                        sumIn3=sumIn3+psf(ix,iy)
                        areaIn3=areaIn3+1
                     else
                        sumOut3=sumOut3+psf(ix,iy)
                        areaOut3=areaOut3+1
                     endif
                  else
                     !---sumOut2 corresponds to photometryApertureRadius
                     if (xyDistance .lt. photometryApertureRadius) then
                        sumOut2=sumOut2+psf(ix,iy)
                        areaOut2=areaOut2+1
                     endif
                  end if
               end do
            end do

            !write(*,*) "L2522 sumIn=", sumIn, " areaIn=", areaIn, " areaOut=", areaOut
            !write(*,*) "L2523 sumIn2=", sumIn2, " areaIn2=", areaIn2, " calRadius=", calApertureRadius
            !write(*,*) "L2524 sumIn3=", sumIn3, " areaIn3=", areaIn3, " radiusDetected=", radiusDetected
            !write(*,*) "L2525 photometryApertureRadius=", photometryApertureRadius

            !write(*,*) "L2527 imagingBkgRate=", imagingBkgRate, " srcExtractionArea=", srcExtractionArea
            !write(*,*) "L2529 im2bkg=",im2bkg, " im2bkgErr=", im2bkgErr, " im2Aperture=", im2Aperture

            im2bkgArea=PI * im2Aperture *  im2Aperture
            !write(*,*) "l2531 im2BkgArea=", im2BkgArea
            ! scale the im2-background area to the source area

            
            
            ! Scale the imaging background counts to the srcExtractionArea (source extraction area)
            imagingBkgRate=imagingBkgRate*srcExtractionArea
            imagingBkgBin=imagingBkgBin*srcExtractionArea
            !imagingBkgRate=imagingBkgRate*areaIn
            imagingBkgRateErr=sqrt(imagingBkgRate)
            imagingBkgBinErr=sqrt(imagingBkgBin)
            imagingErr=imagingErr*srcExtractionArea
            !imagingErr=imagingErr*areaIn

            im2bkg=im2bkg*srcExtractionArea/im2BkgArea
            im2bkgErr=im2bkgErr*srcExtractionArea/im2BkgArea
            !write(*,*) "L2536 im2bkg=",im2bkg, " im2bkgErr=", im2bkgErr
            !*************************************************
            ! Replacing the imaging background with the value 
            ! corresponding to the imaging-chain photometry
            ! 2012-10-20
            !*************************************************
            !write(*,*) "L2552 before imagingBkgRate=", imagingBkgRate, " imagingBkgRateErr=", imagingBkgRateErr, &
            !     " imagingErr=", imagingErr
            imagingBkgRate=im2bkg ! im2bkg is the background rate from the source extraction region
            imagingBkgRateErr=im2bkgErr
            imagingErr=im2bkgErr 
            imagingBkgBin=im2bkg*timeDel
            imagingBkgBinErr=im2bkgErr*timeDel
            imagingErr=imagingBkgBinErr
            !***************************************************
            !write(*,*) "L2555 after imagingBkgBin=", imagingBkgBin, " imagingBkgBinErr=", imagingBkgBinErr, &
            !     " imagingErr=", imagingErr
            !write(*,*) "---"

            !write(*,*) "L2238 imagingBkgRate=", imagingBkgRate, " imagingBkgRateErr=", &
            !     imagingBkgRateErr, " imagingErr=", imagingErr

            !--- VNY 23.04.02 Calculation of the second value for each scaling factor
            ! Calculation of the PSF fractions within different areas
            sumInB=0.  ! for PSF integrated over the area of the source inside the window 
            sumOutB=0. ! the same, but outside the window
            sumIn2B=0. ! PSF integrated within CALapertureRadius
            sumOut2B=0. 
            sumIn3B=0.
            sumOut3B=0.
            sumBkgB=0.

            areaInB=0. ! Geometrical areas 
            areaOutB=0.
            areaIn2B=0.
            areaOut2B=0.
            areaIn3B=0.
            areaOut3B=0.
            areaBkgB=0.

            ! Calculate psf-fractions inside and outside OSW 
            do iy=0,2*int(psfRadius)
               do ix=0,2*int(psfRadius)
                  !---Checking if ix and iy fall into the avoiding circle
                  outsideOfAvoidingCircle = .true.
                  if (nAvoidingSources .gt. 0) then
                     ! Calculate the distance from the avoiding centre
                     do i = 1, nAvoidingSources
                        xc=real(ix)-psfRadius-avoidX(i)+psfShiftX
                        yc=real(iy)-psfRadius-avoidY(i)+psfShiftY
                        xyDistance=sqrt(xc*xc+yc*yc)
                        if (xyDistance .lt. avoidR(i)) then
                           outsideOfAvoidingCircle = .false.                            
                        end if
                     end do
                  end if
                  if (nBackAvoid .gt. 0) then
                     do i=1,nBackAvoid
                        dist=(real(ix)+psfShiftX-xBackAvoid(i))*(real(ix)+psfShiftX-xBackAvoid(i))+ &
                             (real(iy)+psfShiftY-yBackAvoid(i))*(real(iy)+psfShiftY-yBackAvoid(i))
                        dist=sqrt(dist)
                        if (dist .lt. rBackAvoid(i)) then
                           outsideOfAvoidingCircle=.false.
                        endif
                     enddo
                  endif ! nBackAvoid .gt. 0

                  xc=real(ix)-psfRadius+psfShiftX
                  yc=real(iy)-psfRadius+psfShiftY
                  xyDistance=xc*xc+yc*yc
                  xyDistance=sqrt(xyDistance)
 
                  ! calculation of the psf fractions falling inside 
                  ! and outside the OM science window
                  if (xyDistance .le. radiusDetected) then      
                     if (xc .le. dxRight .and. xc .ge. -dxLeft &
                          !  An additional condition is necessary here in order to take into account 
                          !  photon counts falling into the avoiding area of neighbouring sources
                          .and. yc .le. dyUp .and. yc .ge. -dyDown &
                          .and. outsideOfAvoidingCircle) then
                        sumInB=sumInB+psf(ix,iy)
                        areaInB=areaInB+1
                     else                          
                        sumOutB=sumOutB+psf(ix,iy)
                        areaOutB=areaOutB+1
                     end if
                  else
                     if ( xc .le. dxRight .and. xc .ge. -dxLeft &
                          .and. yc .le. dyUp .and. yc .ge. -dyDown &
                          .and. xyDistance .ge. backRadiusArray(0) &
                          .and. xyDistance .le. backRadiusArray(1) &
                          .and. outsideOfAvoidingCircle) then
                        sumBkgB=sumBkgB+psf(ix,iy)
                        areaBkgB=areaBkgB+1
                     endif
                  endif
                  !write(*,*) "L2315 sumBkgB=", sumBkgB, " areaBkgB=", areaBkgB, &
                  !     " areaInB=", areaInB, " areaOutB=", areaOutB
                   !calculation of the ratio between the cal-radius fraction of psf
                   !and its photometry-radius fraction (sumIn2 and sumOut2)
                  if (xyDistance .lt. calApertureRadius) then
                     sumIn2B=sumIn2B+psf(ix,iy) ! For the PHOT to CAL ratio
                     areaIn2B=areaIn2B+1

                     ! Calculation of the ratio between the detected radius fraction of PSF
                     ! and CAL-radius fraction (sumIn3 and sumOut3)
                     if (xyDistance .le. radiusDetected) then
                        sumIn3B=sumIn3B+psf(ix,iy)
                        areaIn3B=areaIn3B+1
                     else
                        sumOut3B=sumOut3B+psf(ix,iy)
                        areaOut3B=areaOut3B+1
                     endif
                  else
                     !---VNY 15.04.2002 sumOut2 is related to photometryApertureRadius
                     if (xyDistance .lt. photometryApertureRadius) then
                        sumOut2B=sumOut2B+psf(ix,iy)
                        areaOut2B=areaOut2B+1
                     endif
                  end if
               end do
            end do
            !write(*,*) "L2341 sumIn2B=", sumIn2B, " areaIn2B=", areaIn2B, &
            !     " sumOut2B=", sumOut2B, " areaOut2B=", areaOut2B
            areaOut=(areaOut+areaOutB)/2.0
            !write(*,*) "L2348 areaIn=", areaIn, " srcExtractionAea=", srcExtractionArea
            !---Using the values corresponding to the rectangular grid
            areaIn=srcExtractionArea
            areaBkg=bkgExtractionArea
            
            if (sumIn .gt. 0.0) then
               ratioBS=sumBkg/sumIn
            else
               ratioBS=0.0
            endif
            
            write(errstr,'(a,f6.1)') "Source extraction area (pixels): ",real(areaIn)
            call message(AppMsg, 2)

            write(errstr,'(a,f6.1)') "Background extraction area (pixels): ",bkgExtractionArea
            call message(AppMsg, 2)

            photcal=1.0 ! ratio between photometry-radius PSF fraction and its CAL-radius fraction
            caldet=1.0  ! ratio between CAL-radius PSF fraction and its detected-radius fraction
            countsRatio=1.0
            psfFrac=0.0
            if (sumIn .gt. 0.) then
               countsRatio=(sumIn+sumOut)/sumIn ! to scale src counts to the CAL-circular area 
               psfFrac=sumIn/(sumIn+sumOut)
            end if
            if (sumIn2 .gt. 0.) then
               photcal=(sumIn2+sumOut2)/sumIn2 ! to scale src counts to the Photometry circular area
            endif
            if (sumIn3 .gt. 0.) then
               caldet=(sumIn3+sumOut3)/sumIn3 ! to scale src counts from the detected to CAL circular areas
            endif

            !write(*,*) "L2745 countsRatio=", countsRatio, " photcal=", photcal, " caldet=", caldet

            psf1Correction(iLoop)=(caldet-1.0)*100.0
            psf2Correction(iLoop)=(photcal-1.0)*100.0 
            mgAver=0.0
            mgMin=1.E30
            mgMax=1.E-30
            nmgAver=0.0
            rateAver=0.0
            backErrAver=0.0
            rateErrAver=0.0
            rateMin=1.E30
            rateMax=-1.E30
            nmgBackAver=0.0
            rateBackAver=0.0
            rateBackAverIm=0.0
            mgBackAver=0.0

            nHighCoincidenceLoss=0
            nCountsOverLimit=0

            ! 
            rateAver2=0.0
            rateErrAver2=0.0
            nRateAver2=0.0
            rateAver3row=0.0
            nRateAver3row=0.0
            bkgRateAver3row=0.0
            nbkgRateAver3row=0.0

            ! 
            spb_coiloss(iLoop)=0.0
            nSpb_coiloss=0.0
            spb_coiloss_max=0.0
            spb_coiloss_min=999.9
            bk_coiloss(iLoop)=0.0
            nBk_coiloss=0.0

           
            snoCorrAver=0.0
            sclCorrAver=0.0
            nsCorrAver=0.0

            !avSrcExtrArea=0.0
            !avBkgExtrArea=0.0
            avSrcCLcorr=0.0
            avBkgCLcorr=0.0
            navSrcCLcorr=0.0
            navBkgCLcorr=0.0

            avMagCLcorr=0.0
            navMagCLcorr=0.0

            ! Loop all over the time-bins
            if (srcRateExists) then
               ! loop over the time line
               !write(*,*) "L2419 nBins=", nBins, " bkgRateExists=", bkgRateExists
               do i = 0, nBins - 1
                  !if (i .eq. 58) then
                  !   write(*,*) "--------------------------_"
                  !   write(*,*) "L2423 i=",i, " rateBackArray(i)=", rateBackArray(i)
                  !endif
                  ! find the position of this master time in the background timeseries
                  ! and interpolate to get the background rate
                  if (bkgRateExists) then
                     call dlocate(timeBack, nBack, timeOut(i), j)

                     !---protection against zero value of j
                     ja=0
                     if (j .lt. 0) then 
                        j=0
                        ja=1
                     endif
                     
                     background=0.0
                     if (backMax .gt. 0) then
                        backErr=medianBkgError
                     else
                        backErr=0.0
                     endif
                     
                     if (bkgFromImage) then
                        backErr=imagingErr
                     endif

                     !if (i .gt. 190) then
                     !   write(*,*) "..................."
                     !   write(*,*) "L2795 i=", i, " j=", j, " nBack=", nBack, &
                     !        " timeOut(i)=", timeOut(i), " timeBack(j)=", timeBack(j), &
                     !        " timeBack(j-1)=", timeBack(j-1)
                     !endif

                     ! For the last time-bin interpolation is impossible, as this time-bin
                     ! is usually shorter than the others. 
                     !   if (j .lt. nBack-1) then
                     ! There is a possibility for j to be zero. Then j-1 will cause a crash
                     if (nBack .gt. 2) then
                        if (j .lt. nBack-1 .and. j .gt. 0) then
                           prop = (timeOut(i)-timeBack(j)) / (timeBack(j)-timeBack(j-1))
                           !if (i .gt. 190) then
                           !   write(*,*) "L2807 prop=", prop, " imagingBkgBin=", imagingBkgBin
                           !endif

                           if (bkgFromImage) then
                              background = imagingBkgBin!*(prop-1.0)
                           else
                              background = rateBackArray(j-1) + &
                                   prop * (rateBackArray(j) - rateBackArray(j-1))
                           endif
                           if (background .lt. 0.0 .or. ja .eq. 1) background=0.0
                           if  (bkgFromImage) then
                              if (background .gt. medianBkg) then
                                 backErr = BackArrayError(j) + &
                                      prop * (BackArrayError(j+1) - BackArrayError(j))
                              endif ! background .gt. medianBkg
                           endif
                        else
                           if (j .eq. nBack-1 .and. j .gt. i) then
                              !write(*,*) "L2827 imagingBkgBin=", imagingBkgBin
                              if  (bkgFromImage) then
                                 background=imagingBkgBin
                              else
                                 background= rateBackArray(j-1)
                                 if (background .gt. medianBkg) then
                                    backErr=BackArrayError(j-1)
                                 endif ! background .gt. medianBkg
                              endif
                           else
                              if  (bkgFromImage) then
                                 !write(*,*) "L2838 imagingBkgBin=", imagingBkgBin
                                 background=imagingBkgBin
                              else   
                                 background= rateBackArray(j)
                                 if (background .gt. medianBkg) then
                                    backErr=BackArrayError(j)
                                 endif ! background .gt. medianBkg
                              endif ! bkgFromImage
                           endif ! j .eq.
                        endif ! j .lt. nBack
                     else ! nBack gt 2
                        if (j .lt. nBins-1 .and. j .gt. 0) then
                           prop = (timeOut(i)-timeBack(j)) / (timeBack(j)-timeBack(j-1))
                           if (bkgFromImage) then
                              !background = imagingBkgRate
                              background = imagingBkgBin
                           else
                              background = rateBackArray(j-1) + &
                                   prop * (rateBackArray(j) - rateBackArray(j-1))
                           endif
                           if (background .lt. 0.0 .or. ja .eq. 1) background=0.0
                           if  (bkgFromImage) then
                              if (background .gt. medianBkg) then
                                 backErr = BackArrayError(j) + &
                                      prop * (BackArrayError(j+1) - BackArrayError(j))
                              endif ! background .gt. medianBkg
                           endif
                        else
                           if (j .eq. nBins-1 .and. j .gt. i) then
                              if  (bkgFromImage) then
                                 !background=imagingBkgRate
                                 background=imagingBkgBin
                              else
                                 background= rateBackArray(j-1)
                                 if (background .gt. medianBkg) then
                                    backErr=BackArrayError(j-1)
                                 endif ! background .gt. medianBkg
                              endif
                           else
                              if  (bkgFromImage) then
                                 !background=imagingBkgRate
                                 background=imagingBkgBin
                              else   
                                 background= rateBackArray(j)
                                 if (background .gt. medianBkg) then
                                    backErr=BackArrayError(j)
                                 endif ! background .gt. medianBkg
                              endif ! bkgFromImage
                           endif ! j .eq.
                        endif ! j .lt. nBack
                        !***********************
                     endif ! nBAck gt 2
                  else ! bkgRateExists
                     if (bkgFromImage) then
                        !background=imagingBkgRate
                        background=imagingBkgBin
                        !backErr=imagingBkgRateErr
                        backErr=imagingBkgBinErr
                     else
                        background=0.0
                        if (backMax .gt. 0) then
                           backErr=medianBkg
                        else
                           backErr=0.0
                        endif
                     endif ! bkgFromImage
                  endif ! bkgRateExists

                  ! Adding the output TIME column
                  timeSourceOutArray(i)=timeOut(i)
                  ! (background: background counts collected from the backArea)         
                  source=rateSourceArray(i)
                  source_safe=source
                  sourceErr=SourceArrayError(i)
                  !if (i .eq. 195) then
                  !   write(*,*) "L2882 i=", i, " 1 source=", source, " bkg=", background
                  !endif
                  ! scaling background to the source area:
                  ! Source counts are collected from the source extraction area, that is sumIn
                  ! and the background counts are collected from the bkgExtractionArea
                  ! Scaling with the correction for PSF partly overlapping with the bkg-area

                  !************************************************
                  ! In the case of using the imaging data scale the 
                  ! background to the imaging area (already scaled)
                  ! otherwise solve the system of two equations
                  !************************************************
                  if (.not. bkgFromImage) then
                     testBkg=0.0
                     if (sumIn .ne. 0.0) then
                        testZnam=areaBkg-areaIn*sumBkg/sumIn
                        if (testZnam .ne. 0.0) then
                           testBkg=(background-source*sumBkg/sumIn)/testZnam
                           testSrc=(source-testBkg*areaIn)/sumIn
                           !---These two values are used if the result of solving two 
                           ! equations is taken into account
                           background=testBkg*areaIn
                           backErr=backErr*areaIn/bkgExtractionArea
                        endif ! testZnam .ne. 0.0
                     endif ! sumIn .ne. 0.0
                  endif ! not bkgFromImage

!---These two values are used when background is calculated without the system of two equations
!   background=background*real(areaIn)/bkgExtractionArea
!   backErr=backErr*real(areaIn)/bkgExtractionArea

                  !--- Now  the source counts could be corrected for asymmetrical position 
                  !    of the source within the OSW (ratio between the PSF fraction within the OSW
                  !    and outside of it), then the background counts should be scaled to the standard
                  !    photometry radius
                     
                  !***************************************************
                  ! Subtracting the background from the source counts 
                  ! in order to correct for the source PSF
                  !***************************************************
                  bgsubtracted=0 ! Indicator of whether background was subtracted or not
                  source = source - background
                  bgsubtracted=1

                  

                  !if (i .eq. 195) then
                  !   write(*,*) "L2927 i=", i, " 2 source=", source, " bkg=", background
                  !endif

                  !******************************************************
                  ! The source counts (with background substracted) 
                  ! should be corrected for the PSF fraction outside of the OSW
                  ! (do this only if background was substracted)
                  !******************************************************
                  source = source * countsRatio
                  sourceErr = sourceErr * countsRatio

                  !if (i .eq. 195) then
                  !   write(*,*) "L2941 i=", i, " 3 source=", source
                  !endif

                 

                  !Now the source is normalised to the circle having the detected 
                  ! radius: radiusDetectedArray(0)
                  !  Converting counts to rates, dividing by time bins
                  if (timeDel .gt. 0) then
                     ! normalising the last time-bin
                     !write(*,*) "L3055 normalising the last time bin"
                     !if (i .gt. 390) then
                     !   write(*,*) "L3057 tameDel=", timeDel, " bkgLastTimeDel=", bkgLastTimeDel, &
                     !        " lastTameDel=", lastTimeDel, " background=", background, " source=", source
                     !endif
                     if (i .eq. nBins-1) then
                        !write(*,*) "L3061 i=", i, " nBins-1=", nBins-1
                        !write(*,*) "L3062 using for backgroundbkgLastTimeDel=", bkgLastTimeDel
                        if (bkgFromImage) then
                           background = background / timeDel
                           backErr = backErr / timeDel
                        else 
                           if (bkgLastTimeDel .gt. 0.) then                        
                              background = background / bkgLastTimeDel
                              !background = background / (timeDel+bkgLastTimeDel)
                              backErr = backErr / bkgLastTimeDel
                              !backErr = backErr / (timeDel+bkgLastTimeDel)
                           endif
                        endif
                        
                        if (lastTimeDel .gt. 0.) then
                           !write(*,*) "L3075 using for source lastTameDel=", lastTimeDel
                           !write(*,*) "L3076 source/timeDel=", source/timeDel
                           !write(*,*) "L3077 source/(timeDel+lastTimeDel)=", source/(timeDel+lastTimeDel)
                           !write(*,*) "L3078 source/(lastTimeDel)=", source/(lastTimeDel)
                           source = source / lastTimeDel
                           !source = source / (timeDel+lastTimeDel)
                           sourceErr = sourceErr / lastTimeDel
                           !sourceErr = sourceErr / (timeDel+lastTimeDel)
                           !write(*,*) "L3083 source=", source, " sourceErr=", sourceErr
                        endif
                     else
                        !if (i .gt. 390) then
                        !   write(*,*) "L3086 using for src and bkg timeDel=", timeDel
                        !endif
                        background = background / timeDel
                        backErr = backErr / timeDel
                        
                        source = source / timeDel
                        sourceErr = sourceErr / timeDel
                     endif
                  end if
                  !if (i .gt. 398) then
                  !   write(*,*) "L3098 i=", i, " source=", source, " sourceErr=", sourceErr
                  !endif

                  ! test the row count rates
                  rateAver3row=rateAver3row+source
                  nRateAver3row=nRateAver3row+1.0

                  !avSrcExtrArea=avSrcExtrArea+source ! this is the same as rateAver3row

                  ! Now source has counts collected during timeDel from the area limited by radiusDetected
                  ! The background has counts collected during timeDel from source extraction area (areaIn)
                     
                  ! Scaling background to the detectedArea (which is areaIn+areaOut or sourceArea)
                  !if (areaIn .gt. 0) then
                  !   background=background*(real(areaIn)+real(areaOut))/real(areaIn)
                  !   backErr=backErr*(real(areaIn)+real(areaOut))/real(areaIn)
                  !endif
                  if (srcExtractionArea .gt. 0) then
                     background=background*detectedArea/srcExtractionArea
                     backErr=backErr*detectedArea/srcExtractionArea
                  endif

                  ! Now background is scaled to the sourceArea

                  !if (i .eq. 195) then
                  !   write(*,*) "L2972 i=", i, " 4 source=", source, " bkg=", background
                  !   write(*,*) "L2973       imagingBkgBin=", imagingBkgBin
                  !   write(*,*) "L2974        source_safe=", source_safe
                  !   write(*,*) "L2975 radiusDetected=", radiusDetected
                  !   write(*,*) "L2976 sourceArea=", sourceArea 
                  !endif

                  ! here the background is divided by time bin
                  
                  
                  !--- here source could be corrected for the coincidence losses
                  !*************************************************
                  ! Before the coincidence loss correction source 
                  ! and background should be scaled 
                  ! to the standard calApertureRadius
                  !*************************************************
                  ! (using 6-pixel detected radius in the case
                  ! of the mod-8 corrupted image)
                  !*************************************************
                  !
                  if (radiusDetected .gt. 0.0) then
                     corrRate = calApertureRadius * calApertureRadius
                     ! this is used temporarily for the calculation of the background scaling factor
                     
                     corrRate = corrRate / ((radiusDetected) * (radiusDetected))
                     source = source * caldet ! Scaling the source count rates to the calApertureRadius
                     sourceErr = sourceErr * caldet

                     source_safe=source_safe*caldet ! Scaling source_safe to the calApertureRadius

                     !if (i .gt. 190) then
                     !   write(*,*) "L3003 before scaling to calApertureRadius background=", background
                     !endif
                     background = background * corrRate ! Scaling background to the calApertureRadius
                     backErr = backErr * corrRate

                     ! here the background is scaled to the CAL 12-radius used for coincidence losses
                     !if (i .gt. 190) then
                     !   write(*,*) "L3005 corrRate=", corrRate, " caldet=", caldet
                     !   write(*,*) "L3006 calApertureRadius=", calApertureRadius, &
                     !        " radiusDetectedArray=", radiusDetectedArray(0), &
                     !        " radiusDetected=", radiusDetected
                     !   write(*,*) "L3009 i=", i, " 4 source=", source, " bkg=", background 
                     !   write(*,*) "   backErr=", backErr
                     !endif
                  end if


                  ! returning the background counts to the source in order to be able to apply
                  !  the coincidence loss correction
                  ! (the source_safe is not background-subtracted)
                  source = source + background

                  
                  !if (i .gt. 190) then
                  !   write(*,*) "L3022 i=", i, " 5 source=", source, " bkg=", background
                  !endif


                  !---The source errors are not increasing here, 
                  ! because we are restoring previously 
                  ! subtracted background (it was just scaling of the source 
                  ! counts for the missed photons outside 
                  ! the window)

                  !******************************************************
                  ! Need to correct the rates for coincident losses, etc-
                  !******************************************************
                  !---initialisation of corrRateErrMinus1 and corrRateErrPlus1
                  corrRateErrMinus1=0.0
                  corrRateErrPlus1=0.0
                                    
                  if (background .gt. 0.0) then
                     ! ***************************************************************************
                     ! Compute the coincidence loss correction using theoretical calibration curve
                     !****************************************************************************
                     !corrRate = CorrectedCountRate(dble(background), backErr, frameTime, deadFrac,  &
                     !     extend, corrRateErrMinus1, corrRateErrPlus1)
                     ! *********************************************************************
                     ! Compute the coincidence loss correction using empirical calibration
                     !**********************************************************************
                     empirical = .true.

                     corrRate = CorrectedCountRate(dble(background), backErr, frameTime, deadFrac,  &
                          extend, empirical, corrRateErrMinus1, corrRateErrPlus1)
                     !corrRate = CorrectedCountRateTest(dble(background), backErr, frameTime, deadFrac,  &
                     !     extend, empirical, corrRateErrMinus1, corrRateErrPlus1, i)
                     
                     bk_coiloss(iLoop)=bk_coiloss(iLoop)+corrRate/background
                     nBk_coiloss=nBk_coiloss+1.0

                     

                     !******* Test printing ********
                     !if (i .gt. 190) then
                     !   write(*,*) "L3059 background=", background, " corrRate=", corrRate
                     !endif
                     !*****************************
                     ! Saving the original value of background, which was initially
                     ! added to the source counts in order to calculate the 
                     ! coincidence losses for the source region. 
                     !****************************
                     bkg_safe=background ! bkg_safe contains the background which was not CL-corrected
                     background = corrRate ! not the variable background contains the CL-corrected bkg
                     
                     iloss=int(background+0.55)
                     
                     ! protection against the subscript of the array loss2corrRMS
                     if (iloss .gt. nAux) iLoss=nAux
                     if (iloss .gt. 0) then
                        corrRateErrMinus1=loss2corrRMS(iloss)
                        corrRateErrPlus1=loss2corrRMS(iloss)
                     endif

                     !*********************************************************
                     ! Time-dependent sensitivity correction
                     !********************************************************
                     if (degradationCorrection) then
                        background = background * tdpCorrFactor
                        corrRateErrMinus1=corrRateErrMinus1 * tdpCorrFactor
                        corrRateErrPlus1=corrRateErrPlus1 * tdpCorrFactor
                     endif
                     !if (i .gt. 190) then
                     !   write(*,*) "L3087 background=", background, " im2bkg=", im2bkg, " tdpCorrFactor=", tdpCorrFactor
                     !endif

                  else
                     nNegative=nNegative+1
                  endif ! background
                  
                  !---Initialisation of corrRateErrMinus and corrRateErrPlus
                  corrRateErrMinus=0.0
                  corrRateErrPlus=0.0
                  
                  if (source .gt. 0.0) then
                     ! Now the correction for the coincidence losses 
                     ! and dead fraction is made before
                     ! scaling the count rates to the photometric aperture radius
                     if (source .lt. maxCountsPerFrame) then

                        empirical = .true.


                        !if (i .eq. 58) then
                        !   write(*,*) "L2777 maxCountsPerFrame+", maxCountsPerFrame
                        !   write(*,*) "L2778 i=", i, " frameTime=", frameTime, " deadFrac=", deadFrac
                        !   write(*,*) "L2779 source=", source!, " corr source=", source*deadFracCorrection
                        !endif

                        !******************************************
                        ! 2009-10-11 applying the new experimental deadFracCorrection
                        !*******************************************
                        if (dtCorrection .gt. 0) then
                           source=source*deadFracCorrection
                        endif

                        !********************************************************
                        ! Tests 2009.10.09 of a new (experimental) small correction to the raw count
                        ! rates before applying the coincidence loss correction
                        !******************************************************
                        !corrRate = CorrectedCountRateTest(dble(source), sourceErr, frameTime, deadFrac, &
                        !     extend, empirical, corrRateErrMinus, corrRateErrPlus, i)
      
                        corrRate = CorrectedCountRate(dble(source), sourceErr, frameTime, deadFrac, &
                             extend, empirical, corrRateErrMinus, corrRateErrPlus)

                        !******* Test printing ********
                        !if (i .eq. 195) then
                        !   write(*,*) "L3131 source=", source, " corrRate=", corrRate
                        !endif
                        if (i .lt. nBins-1) then
                           snoCorrAver=snoCorrAver+source
                           sclCorrAver=sclCorrAver+corrRate
                           nsCorrAver=nsCorrAver+1.0
                        endif
                        !*****************************
                        iloss=int(source+0.5)
                        if (source .gt. 0.0) then
                           spb_coiloss(iLoop)=spb_coiloss(iLoop)+corrRate/source
                           nSpb_coiloss=nSpb_coiloss+1.0
                           if (corrRate/source .gt. spb_coiloss_max) spb_coiloss_max=corrRate/source
                           if (corrRate/source .lt. spb_coiloss_min) spb_coiloss_min=corrRate/source
                        endif

                        ! Protection against subscripts
                        if (iloss .gt. nAux) iloss=nAux
                        if (iloss .gt. 0) then 
                           corrRateErrMinus=loss2corrRMS(iloss)
                           corrRateErrPlus=loss2corrRMS(iloss)
                        endif
                        !*********************************************************
                        ! Time-dependent sensitivity correction
                        !********************************************************
                        if (degradationCorrection) then
                           corrRate = corrRate * tdpCorrFactor
                           corrRateErrMinus=corrRateErrMinus * tdpCorrFactor
                           corrRateErrPlus=corrRateErrPlus * tdpCorrFactor
                           !if (i .eq. 58) then
                           !   write(*,*) "L2833 tdpCorrFactor=", tdpCorrFactor, " corrRate=", corrRate
                           !endif
                        endif
                     else
                        empirical = .true.
                        corrRate = CorrectedCountRate(maxCountsPerFrame, sourceErr, &
                             frameTime, deadFrac, &
                             extend, empirical, corrRateErrMinus, corrRateErrPlus)
                        nCountsOverLimit=nCountsOverLimit+1

                        !if (i .eq. 58) then
                        !   write(*,*) "L2844 maxCountsPerFrame+", maxCountsPerFrame
                        !   write(*,*) "L2845 i=", i, " corrRate=", corrRate 
                        !endif

                     endif
                     ! find the formula-estimated errors
                     !if (i .eq. 20) then
                     !   write(*,*) " OMLCBUILD L1890 maxCounts=", maxCounts, " nAux=", nAux
                     !   write(*,*) "************************************************"
                     !   do iloss=1,maxCounts-1
                     !      if (lossCountsArray2(iloss) .lt. source .and. &
                     !           lossCountsArray2(iloss+1) .ge. source) then
                     !         write(*,*) " L1852 source=", source, " corrRate=", corrRate, &
                     !              " iloss=", iloss, " lossCountsArray2(iloss)=", lossCountsArray2(iloss)
                     !         write(*,*) " dispLeftArray(iloss)=", dispLeftArray(iloss), &
                     !              " dispRightArray(iloss)=", dispRightArray(iloss)
                     !         write(*,*) " corrRateErrMinus=", corrRateErrMinus, &
                     !             " corrRateErrPlus=", corrRateErrPlus
                     !      endif
                     !   enddo ! iloss
                     !   iloss=int(source)
                     !   write(*,*) " L1903 int(source)=", iloss, " loss2corr=", loss2corr(iloss), &
                     !        " errLeft=", loss2corrErrLeft(iloss), " errRight=", loss2corrErrRight(iloss), &
                     !        " rms=", loss2corrRMS(iloss)
                     !   write(*,*) "**************************************************"
                     !endif
                     
                     corrToSource=corrRate/source
                     if (corrToSource .gt. 2.0) nHighCoincidenceLoss=nHighCoincidenceLoss+1    
                     source = corrRate
                     !if (i .eq. 58) then
                     !   write(*,*) "L2875 source=", source
                     !endif
                  else     
                     nNegative=nNegative+1
                  endif ! Source .gt. 0.0
                  
                  !if (i .eq. 195) then
                  !   write(*,*) "L3212 before scaling the background background=", background
                  !endif
                  rateBackOutArray(i)=background ! corrected for coincidence losses
                  rateBackErrOutArray(i) = real(abs(max(corrRateErrMinus1, corrRateErrPlus1)))


                  !---Protection against too high values of corrRAteErrMinus1, 
                  ! corrRateErrPlus1
                  if (rateBackErrOutArray(i) .gt. backErr*15.0) then
                     rateBackErrOutArray(i)=backErr
                  endif
                  corrRateErr=real(abs(max(corrRateErrMinus, corrRateErrPlus)))
                  if (corrRateErr .gt. sourceErr*15.0) then
                     corrRateErr=sqrt(abs(source))
                  endif

!----------------------------------------------------------------------------------
! 06.07.2007 Suppress the background subtraction in the case of a corrupted source
!-----------------------------------------------------------------------------------
                  
                  if (mod8flag .ne. 1 .or. mod8corrupted .eq. 2) then
                     ! (mod8corrupted=2 is equivalent to the condition of ignoring mod8-noise)
                     !******************************************************
                     ! There is a possibility here that in the 
                     ! difference between the source and background coincidence
                     ! loss corrections is applied twice. So, the initial (uncorrected)
                     ! background value must be subtracted here in order not to
                     ! overestimate the coincidence losses
                     !*******************************************************
                     ! 2012-10-21 vny:Subtracting the CL-corrected background from the CL-corrected source+bkg
                     ! seems to be wrong. Perhaps we should use the same background that
                     ! was initially subtracted to make the source scaling
                     sourceSubtracted= source -  rateBackOutArray(i)
                     !sourceSubtracted= source - bkg_safe

                     !write(*,*) "L3359 i=", i, " source=", source, " bkg=",  rateBackOutArray(i), &
                     !     " sourceSubtracted=", sourceSubtracted 
                     !if (i .eq. 195) then
                     !   write(*,*) "L3243 sourceSubtracted=", sourceSubtracted, &
                     !        " rateBackOutArray=", rateBackOutArray(i), " bkg_safe=", bkg_safe, &
                     !        " im2bkg=", im2bkg
                     !endif
                     if (subtractBkg) then
                        rateSourceOutArray(i) = sourceSubtracted
                        bgsubtracted=1
                     else
                        rateSourceOutArray(i) = source
                        bgsubtracted=0
                     endif

                      
                     if (i .lt. nBins-1) then
                        bkgRateAver3row=bkgRateAver3row+bkg_safe
                        nbkgRateAver3row=nbkgRateAver3row+1.0
                     endif
                  else
                     sourceSubtracted=source
                     rateSourceOutArray(i) = source
                     bgsubtracted=0
                  endif


                  !******* Test printing ********
                  !if (i .gt. 190) then
                  !   !write(*,*) "............................"
                  !   write(*,*) "L3258 sourceSubtracted=", sourceSubtracted, &
                  !        " source-bkg_safe=", source-bkg_safe ! for uncrorrected background
                  !   !write(*,*) "............................"
                  !   write(*,*) "L3261 background=", background, " rateBackOutArray=", &
                  !        rateBackOutArray(i) 
                  !endif
                  !*****************************

                  !---Increasing the source error bars by about 2% because of 
                  !   a systematic non-poissonian
                  !   error for bright sources  after the coincidence loss correction 
                  rateSourceErrOutArray(i) = sqrt(corrRateErr**2 &
                       + rateBackErrOutArray(i)**2 + (0.02 * rateSourceOutArray(i))**2)

                  !--- Calculation of the magnitude
                  magnitude=1.e30
                  magnitudeErr=1.e30
                  
                  if (rateSourceOutArray(i) .gt. 0.0) then

                     !**************************************************
                     ! Using sourceSubtracted for the photometry in case 
                     ! the output counts were not bkg-subtracted
                     !**************************************************
                     
                     corrRate=sourceSubtracted

                     !---scaling corrRateErrors
                     corrRateErrMinus=rateSourceErrOutArray(i) * photcal
                     corrRateErrPlus=rateSourceErrOutArray(i) * photcal

                     ! Scaling source counts to the photometry aperture radius 
                     corrRate = corrRate * photcal

                     call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                          corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                     magnitude=real(corrMag)
                     magnitudeErr=real(abs(max(corrMagErrMinus, corrMagErrPlus)))

                     if (i .lt. nBins-1) then
                        mgAver=mgAver+magnitude
                        nmgAver=nmgAver+1.0
                     endif

                     ! average the CL-corrected source (bkg-subtracted)
                     if (i .lt. nBins-1) then
                        avSrcCLcorr=avSrcCLcorr+corrRate
                        navSrcCLcorr=navSrcCLcorr+1.0
                        
                     endif
                     
                     if (i .lt. nBins-1) then
                        rateAver=rateAver+sourceSubtracted
                        rateErrAver=rateErrAver+real(abs(max(corrRateErrMinus,corrRateErrPlus)))
                     endif

                     if (magnitude .lt. mgMin) then
                        mgMin=magnitude
                     end if

                     if (magnitude .gt. mgMax) then
                        mgMax=magnitude
                     end if

                     if (rateSourceOutArray(i) .lt. rateMin) then
                        rateMin=rateSourceOutArray(i) ! Only those rates are taken, which are > 0
                     end if
                     if (rateSourceOutArray(i) .gt. rateMax) then
                        rateMax=rateSourceOutArray(i)
                     end if

                  else   
                     nNegative=nNegative+1
                     ! checking whether the negative rate affects the resuls
                  endif ! rateSourceOutArray
                  !if (i .eq. 58) then
                  !   write(*,*) "L2899 rateSourceOutArray(58)=", rateSourceOutArray(i)
                  !endif
                  
                  !********************************************
                  ! The variable rateAver2 contain the avearage source counts
                  ! taking into account negative values (due to bkg subtraction)
                  if (i .lt. nBins-1) then
                     rateAver2=rateAver2+sourceSubtracted
                     nRateAver2=nRateAver2+1.0
                     rateErrAver2=rateErrAver2+rateSourceErrOutArray(i)
                  endif
                  !--- Calculation of the background magnitude
                  magnitudeBack=1.e30
                  magnitudeBackErr=1.e30

                  if (rateBackOutArray(i) .gt. 0.0) then
                     corrRate=rateBackOutArray(i)
                     !---scaling corrRateErrors
                     corrRateErrMinus=rateBackErrOutArray(i) 
                     corrRateErrPlus=rateBackErrOutArray(i) 

                     !if (i .eq. nBack-1) then
                     !   write(*,*) "L3415------------------"
                     !endif
                     !if (i .eq. 195) then
                     !   write(*,*) "L3375 before phot-scaling bkg=", corrRate, " phot/cal=", &
                     !        (photometryApertureRadius/calApertureRadius)**2  
                     !   write(*,*) " photometryApertureRadius=", photometryApertureRadius, &
                     !        " calApertureRadius=", calApertureRadius
                     !endif
                     ! Scaling background rates to the photometry aperture radius
                     if (calApertureRadius .gt. 0.0) then
                        corrRate = corrRate * (photometryApertureRadius/calApertureRadius)**2
                        corrRateErrMinus=corrRateErrMinus* (photometryApertureRadius/calApertureRadius)**2
                        corrRateErrPlus=corrRateErrPlus* (photometryApertureRadius/calApertureRadius)**2
                     endif
                     !if (i .eq. 195) then

                     !if (i .gt. 100) then  
                     !   write(*,*) "L3385  i=", i, " nBins=", nBins, " AFTER phot-scaling bkg=", corrRate, "***"
                     !endif

                     ! average the CL-corrected background
                     if (i .lt. nBins-1) then
                        avBkgCLcorr=avBkgCLcorr+corrRate
                        navBkgCLcorr=navBkgCLcorr+1.0
                     endif

                     backErrAver=backErrAver+real(abs(max(corrRateErrMinus,corrRateErrPlus)))
                     call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                          corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)

                     magnitudeBack=real(corrMag)
                     magnitudeBackErr=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
                     if (i .lt. nBins-1) then
                        mgBackAver=mgBackAver+corrMag
                        nmgBackAver=nmgBackAver+1
                        rateBackAver=rateBackAver+rateBackOutArray(i)
                     endif
                  endif ! ratebackOutArray
               enddo ! loop over all sources (i)

               if (nsCorrAver .gt. 1.0) then
                  snoCorrAver=snoCorrAver/nsCorrAver
                  sclCorrAver=sclCorrAver/nsCorrAver
               endif
               !write(*,*) "L3400 snoCorrAver=", snoCorrAver, " sclCorrAver=", sclCorrAver, " ***"

               if(nmgBackAver .gt. 0.0) then
                  rateBackAver=rateBackAver/nmgBackAver
                  
               end if
               if(navBkgCLcorr .gt. 0.0) then      
                  avBkgCLcorr=avBkgCLcorr/navBkgCLcorr ! Average background scaled to the photometry area
               endif

               if (subtractBkg) then
                  if (mod8flag .eq. 1 .and. mod8corrupted .ne. 2) then
                     ! (mod8corrupted=2 is equivalent to ignoreMod8noise=true)
                     call warning("bright source", &
                          "background is not subtracted since the source is likely to be corrupted")
                  endif
               endif

               percentage=0
               if (nBins .gt. 0) percentage=100*nHighCoincidenceLoss/nBins
               if (percentage .gt. 1) then
                  write (errstr, *) percentage, &
                       "% of the points exceed the coincidence loss factor of two"
                  call message(AppMsg, 2)
               endif
               clcComment1=" "
               if (percentage .gt. 10) then
                  write(clcComment1, '(i4, a)') percentage, &
                       "% of the points exceed the coincidence loss factor of two"
               endif

               percentage=0
               if (nBins .gt. 0) percentage=100*nCountsOverLimit/nBins
               if (percentage .gt. 1) then
                  write (errstr, *) percentage, &
                       "% of the points reach the limit of the coincidence loss correction"
                  call message(AppMsg, 2)
               endif
               clcComment2=" "
               if (percentage .gt. 10) then
                  write(clcComment2, '(i4, a)') percentage, &
                       "% of the points reach the limit of the coincidence loss correction"
               endif

               if (nmgAver .gt. 0.0) then
                  mgAver=mgAver/nmgAver
                  rateAver = rateAver/nmgAver
                  rateErrAver=rateErrAver/nmgAver
                  
               end if
               if (navSrcCLcorr .gt. 0.0) then
                  avSrcCLcorr=avSrcCLcorr/navSrcCLcorr
               endif



               ! 
               if (nRateAver2 .gt. 1.0) then
                  rateAver2=rateAver2/nRateAver2
                  rateErrAver2=rateErrAver2/nRateAver2
                  rateErrAver2=rateErrAver2/sqrt(nRateAver2)
               endif

               !write(*,*) "L3594 avSrcCLcorr=", avSrcCLcorr, " avBkgCLcorr=", avBkgCLcorr, " ***"
               !write(*,*) "L3595 rateAver=", rateAver, " rateAver2=", rateAver2, " mgAver=", mgAver
               !write(*,*) "L3596 photcal=", photcal, " rateAver2*photcal=", rateAver2*photcal
               !write(*,*) "L3597 rateAver2=", rateAver2
               write(errstr,'(a,f6.3)') "Average (corrected) count rate: ",rateAver2 * photcal
               call message(AppMsg, 2)


               if (nRateAver3row .gt. 1.0) then
                  rateAver3row=rateAver3row/nRateAver3row
               endif
               if (nbkgRateAver3row .gt. 1.0) then
                  bkgRateAver3row=bkgRateAver3row/nbkgRateAver3row
               endif

               if (.not. bkgFromImage) im2bkg=bkgRateAver3row
               !write(*,*) "-----------------------------------------"
               !write(*,*) "L3611 rateAver3row=", rateAver3row, " nRateAver3row=", nRateAver3row, " im2bkg=", im2bkg,  " ***"
               !write(*,*) "L3612 bkgRateAver3row=", bkgRateAver3row, &
               !     " nbkgRateAver3row=", nbkgRateAver3row

               !write(*,*) "L3615 timeExposure=", timeExposure
               !write(*,*) "L3616 calculated significance=", rateAver3row*timeExposure/sqrt(rateAver3row*timeExposure &
               !     +2.0*im2bkg*timeExposure)

               !**********************************
               ! modify the significance column
               !**********************************
               if (rateAver2+2.0*rateBackAver .gt. 0.0) then
                  !significance(iLoop)=rateAver2/sqrt(rateAver2+2.0*rateBackAver)
                  significance(iLoop)=rateAver3row*timeExposure/sqrt(rateAver3row*timeExposure &
                       +2.0*im2bkg*timeExposure)
               endif

               

               !---Calculation of the average magitude 
               !   on the basis of the average positive counts
               corrRate=rateAver
               !corrRate=avSrcCLcorr
               ! Scaling source counts to the photometry aperture radius
               ! avSrcCLcorr is already scaled to the photometry aperture
               corrRate = corrRate * photcal
               corrRateErrMinus=rateErrAver*photcal
               corrRateErrPlus=rateErrAver *photcal

               
               if (corrRate .gt. 0.0) then 
                 
                  call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                       corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                  magnitude=real(corrMag)
                  mag(iLoop)=magnitude
                  magnitudeErr=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
                  magErr(iLoop)=magnitudeErr
               else
                  
                  magnitude=0.0
                  mag(iLoop)=0.0
                  magnitudeErr=0.0
                  magErr(iLoop)=0.0
               endif

               !***************************************************************
               ! rateAver2 is the average of the CL-corrected count rate taking
               ! into account negative counts (due to the subtracted background)
               !***************************************************************
               !write(*,*) "L3661 rateAver=", rateAver, " rateAver2=", rateAver2, &
               !     " avSrcCLcorr=", avSrcCLcorr, " positive counts magnitude=", magnitude 
               !---Calculation of the average magitude with
               !   the negative counts taken into account
               corrRate=rateAver2
               !corrRate=avSrcCLcorr
               ! Scaling source counts to the photometry aperture radius 
               ! avSrcCLcorr is already scaled to the photometry aperture
               corrRate = corrRate * photcal
               corrRateErrMinus=rateErrAver2*photcal
               corrRateErrPlus=rateErrAver2 *photcal

               if (corrRate .gt. 0.0) then 
                  call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                       corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                  magnitude=real(corrMag)
                  mag(iLoop)=magnitude
                  magnitudeErr=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
                  magErr(iLoop)=magnitudeErr
               else
                  magnitude=0.0
                  mag(iLoop)=0.0
                  magnitudeErr=0.0
                  magErr(iLoop)=0.0
               endif

               !write(*,*) "L3687 negative count magnitude=", magnitude
               ! storing the rate and magnitude in the source list table
               ! finding whether rate(iLoop) has a NaN-value (using the logical inconsistensy og NaN)
               if (rate(iLoop) .ge. 0.0) then
                  ! the value is not NaN (do nothing)
               else
                  ! the rate is either NaN or <0
                  !rate(iLoop)=rateAver2
                  rate(iLoop)=rateAver3row
               endif
               if (rateErr(iLoop) .ge. 0.0) then
                  ! the rateErr is non NaN (do nothing)
               else
                  ! the rateErr is either NaN or <0 
                  rateErr(iLoop)=rateErrAver2
               endif

               corr_rate(iLoop)=rateAver2*photcal
               !corr_rate(iLoop)=avSrcCLcorr
               corr_rate_err(iLoop)=rateErrAver2*photcal
               nopixels(iLoop)=int(sourceArea)
               
               if (nSpb_coiloss .gt. 1.0) then
                  spb_coiloss(iLoop)=spb_coiloss(iLoop)/nSpb_coiloss
               endif

               if (nBk_coiloss .gt. 1.0) then
                  bk_coiloss(iLoop)=bk_coiloss(iLoop)/nBk_coiloss
               endif

               !----- Calculation of magnitudeMin and magnitudeMax
               !      minimal rate
               if (rateMin .gt. 0.0 .and. rateMin .lt. 1.0E+10) then
                  corrRate=rateMin
               else
                  corrRate=0.0
               endif
               ! Scaling source counts to the photometry aperture radius 
               
               corrRate = corrRate * photcal
               corrRateErrMinus=corrRateErrMinus * photcal * 0.1
               corrRateErrPlus=corrRateErrPlus * photcal * 0.1
               
               if (corrRate .gt. 0.0) then
                  
                  call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                       corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                  magnitude=real(corrMag)
                  magnitudeErr=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
               else
                  
                  magnitude=0.0
                  magnitudeErr=0.0
               endif

               ! maximal rate
               if (rateMax .gt. 0.0) then
                  
                  corrRate=rateMax
               else
                  
                  corrRate=0.0
               endif
               ! Scaling source counts to the photometry aperture radius 
               corrRate = corrRate * photcal

               
               if (corrRate .gt. 0.0) then
                  
                  call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                       corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                  magnitude=real(corrMag)
                  magnitudeErr=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
               else
                  
                  magnitude=0.0
                  magnitudeErr=0.0
               endif

               if(nmgBackAver .gt. 0.0) then
                  mgBackAver=mgBackAver/nmgBackAver
                  
                  backErrAver=backErrAver/nmgBackAver
                  backErrAver=backErrAver/sqrt(nmgBackAver)
               end if
               !*****************************************
               !write(*,*) "L3773 iLoop=", iLoop, " backgroundRate=", backgroundRate(iLoop), " rateBackaver=", rateBackAver, &
               !     " backErrAver=", backErrAver
               !write(*,*) "L3775 rate=", rate, " rateErr=", rateErr
               !write(*,*) "L3776 rateAver3row=", rateAver3row, " nRateAver3row=", nRateAver3row
               !*****************************************
               if (bkgRateAver3row .gt. 0.0) then 
                  if (bkgFromImage) then
                     backgroundRate(iLoop)=im2bkg
                  else
                     backgroundRate(iLoop)=bkgRateAver3row
                  endif
               endif

               if (rateAver3row .gt. 0.0) then
                  rate(iLoop)=rateAver3row 
               endif

               !backgroundRate(iLoop)=rateBackAver
               if (backgroundRate(iLoop) .ge. 0.0) then
                  ! the value is not NaN (do nothing)
               else
                  ! backgroundRate is either NaN or <0
                  if (bkgFromImage) then
                     backgroundRate(iLoop)=im2bkg
                     backgroundRate(iLoop)=rateBackAver
                  endif
               endif

               !corr_background_rate(iLoop)=rateBackAver
               corr_background_rate(iLoop)=avBkgCLcorr

               
               if (bkgFromImage) then
                  
                  corrRate=avBkgCLcorr
                  corrRateErrMinus=backErrAver
                  corrRateErrPlus=backErrAver
                  if (calApertureRadius .gt. 0.0) then
                     corrRateErrMinus = corrRateErrMinus * (photometryApertureRadius/calApertureRadius)**2
                     corrRateErrPlus = corrRateErrPlus * (photometryApertureRadius/calApertureRadius)**2
                  endif
                  call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                       corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                  magnitudeBack=real(corrMag)
                  magnitudeBackErr=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
               else
                  
                  if (rateBackAver .gt. 0.0) then
                     !write(*,*) "L3821 rateBackAver=", rateBackAver
                     !if (avBkgCLcorr .gt. 0.0) then                                   
                     corrRate=rateBackAver
                     
                     corrRateErrMinus=backErrAver
                     corrRateErrPlus=backErrAver
                     ! Scaling background rates to the photometry aperture radius
                     if (calApertureRadius .gt. 0.0) then
                        if (.not. bkgFromImage) then
                           corrRate = corrRate * (photometryApertureRadius/calApertureRadius)**2
                        endif
                        corrRateErrMinus = corrRateErrMinus * (photometryApertureRadius/calApertureRadius)**2
                        corrRateErrPlus = corrRateErrPlus * (photometryApertureRadius/calApertureRadius)**2
                     endif
                     call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                          corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                     magnitudeBack=real(corrMag)
                     magnitudeBackErr=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
                  endif ! ratebackOutArray
               endif
               !write(*,*) "L3841 magnitudeBack=", magnitudeBack
               nAux=0

               !---Median filtration of the source rates to make a warning 
               do i = 0, nBins - 10
                  ! Array for the median filter
                  do ix= 1, 9
                     rateAuxArray(ix)=rateSourceOutArray(i+ix)
                  end do
                  ! Ordering the auxiliar array
                  do ix=1,8
                     do iy=ix+1,9
                        if (rateAuxArray(iy) .gt. rateAuxArray(ix)) then
                           rateAux=rateAuxArray(ix)
                           rateAuxArray(ix)=rateAuxArray(iy)
                           rateAuxArray(iy)=rateAux
                        end if
                     end do
                  end do
                  rateAux=rateAuxArray(5)
                  if (rateSourceOutArray(i) .gt. rateAux*30.0 .and. rateAux .gt. 0.0) then
                     nAux=nAux+1
                     end if
                  end do
                  ! Repeat the analysis for the last 9 values
                  if (nBins .gt. 9) then
                     do i=nBins-9, nBins-1
                        if (rateSourceOutArray(i) .gt. rateAux*30.0 .and. rateAux .gt. 0.0) then
                           nAux=nAux+1
                           !call warning("sourceRate", "High value of the source rate")
                        end if
                     end do
                  endif
                  if (nAux .gt. 0) then
                     write (errstr, *) "High values of the source rates in ", nAux, " points"
                     call message(AppMsg, 2)
                     !call warning("sourceRate", "High value(values) of the source rate")
                  endif
               endif ! srcRateExists
               
               psfApp= .true.
               vignApp= .false.
               backApp= .true.
               if (bgsubtracted .eq. 0) backApp = .false.
               deadApp= .true.

!************************** 2009-03-11*****************
!----------- variability data
              
               allocate(Net_Rate(1, size(rateSourceOutArray(:))))
               
               Net_Rate(1,:)=rateSourceOutArray(:)
               !write(*,*) "L2576 lbound=", lbound(rateSourceOutArray,1), &
               !     " ubound=", ubound(rateSourceOutArray,1)
               iaux=ubound(rateSourceOutArray,1)
               !do i=1,iaux
               !   write(*,*) "L2580 i=", i, " rate=", rateSourceOutArray(i)
               !end do

               allocate(Net_Err(1, size(rateSourceErrOutArray(:)) ) )
               Net_Err(1,:)=rateSourceErrOutArray(:)

               allocate(Bkg_Rate(1, size(rateBackOutArray(:)) ) )
               Bkg_Rate(1,:)=rateBackOutArray(:)

               allocate(Bkg_Err(1, size(rateBackErrOutArray(:)) ) )
               Bkg_Err(1,:)=rateBackErrOutArray(:)

               call check_sizes (Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, Nlight, Nbin)
               !write(*,*) "L2663 Nlight=", Nlight, " Nbin=", Nbin

               !            do i=1,Nbin
               !               write(*,*) "L1549 i=", i, " Net_Rate=", Net_Rate(1,i)
               !            end do

               Bin_Width=timeDel
               Exposure_Time = timeExposure
               Time_Zero = 0
               

               allocate( Time_Vect(Nbin) )
               do i=1,Nbin
                  Time_Vect(i) = (i-0.5)*Bin_Width
               end do
               
               
               call delete_gap_data(Time_Vect, Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, &
                    Nlight, Nbin)
               
               
               Binning_Factor=1
               call rebin(Binning_Factor, Bin_Width, Time_Vect, Net_Rate, Net_Err, &
                    Bkg_Rate, Bkg_Err, Good_Bin, Nlight, Nbin, rateSourceOutCol, rateBackOutCol, gotnullvals)
               
               
               call recover_counts(Bin_Width*Binning_Factor, Net_Rate, Bkg_Rate, &
                    Net_Count, Net_Total, Good_Bin, Nlight, Nbin)
               
              
               deallocate(Net_Rate, Net_Err,Bkg_Rate,Bkg_Err, Time_Vect, Good_Bin)
               
               deallocate(gotnullvals)
         
               
               
               if (Nbin > 2) then
                  
                  call variability_tests(Bin_Width*Binning_Factor, Net_Count, Net_Total,&
                       Nlight, Nbin, Which_Tests, Light_Curve)
               else
                  call warning('OMLCBUILD','Not enough bins to process variability tests')
               end if
               
               deallocate(Net_Count, Net_Total)

               !write(*,*) "L2709 l.c.variance=",  Light_Curve(1)%Variance
               !write(*,*) "L2704 Chi_square stat==",Light_Curve(1)%Chi_Square%Stat
               !write(*,*) "L2705 chi prob=",Light_Curve(1)%Chi_Square%Prob
               !write(*,*) "L2706 Kolm.Sm.:=", Light_Curve(1)%K_S%Stat
               !write(*,*) "L2707 k-s prob=",Light_Curve(1)%K_S%Prob
               !write(*,*) "L2708 max deviation=",Light_Curve(1)%Max_Deviation

               chi2(iLoop)=Light_Curve(1)%Chi_Square%Stat
               chi2prob(iLoop)=Light_Curve(1)%Chi_Square%Prob
               ksStat(iLoop)=Light_Curve(1)%K_S%Stat
               ksStatProb(iLoop)=Light_Curve(1)%K_S%Prob
               maxDev(iLoop)=Light_Curve(1)%Max_Deviation
!-------------------------------------------

!**************************2009-03-11 ****************^^^
               
               !---Three new attributes have been added to the output table: MG_AVER, MG_MIN, MG_MAX
               call setAttribute(outTable, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
               call setAttribute(outTable, "MAGNITUD", mag(iLoop), "mg", "Average magnitude of the source")
               call setAttribute(outTable, "MAGN_MIN", mgMin, "mg", "Minimum magnitude in this time series")
               call setAttribute(outTable, "MAGN_MAX", mgMax, "mg", "Maximum magnitude in this time series")

              

               !---Attribute with the PSF-fraction within the OSW
               call setAttribute(outTable, "PSF_FRAC", psfFrac, " ", "Fraction of PSF inside OSW")
               call setAttribute(outTable, "VIGNET", vignet ,"Geometrical vignetting of the source")
               call setAttribute(outTable, "NPIXSOU", sourceArea, "pix", "Source area")
               call setAttribute(outTable, "NPIXBACK", backArea, "pix", "Background area")
               call setAttribute(outTable, "SRRADIUS", radiusDetectedArray(0), "pix", "Source extraction radius")
               call setAttribute(outTable, "CLRADIUS", calApertureRadius, "pix", "Coincidence loss correction radius")
               call setAttribute(outTable, "PHRADIUS", photometryApertureRadius,"pix", "Photometry aperture radius")
               call setAttribute(outTable, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
               call setAttribute(outTable, "DEADAPP", deadApp, " Dead time correction applied")
               call setAttribute(outTable, "BACKAPP", backApp,  "Background is subtracted")
               call setAttribute(outTable, "PSF_CAPP", psfApp, "PSF-fraction correction applied")
               !---
           
               call setAttribute(outTable, "HDUCLAS1", "LIGHTCURVE", "Light curve series")
               call setAttribute(outTable, "HDUCLAS2", "NET", "Background subtracted")
               call setAttribute(outTable, "HDUCLAS3", "RATE", "Exposure corrected")
               call setAttribute(outTable, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
               call setAttribute(outTable, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
               call setAttribute(outTable, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
               call setAttribute(outTable, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
               
               if (columnDataTypeFlag .eq. 2) then
                  call setAttribute(outTable, "QFLAG", sourceQF8(0), " ",  "Quality flag")
               else
                  call setAttribute(outTable, "QFLAG", sourceQF16(0), " ",  "Quality flag")
               endif
               call setAttribute(outTable, "CFLAG", sourceCF(0), " ",  "Confusion flag")
               !--- Adding the time-degradation coefficient as an attribute
               call setattribute(outTable, "TDS_CORR", tdpCorrFactor, "Sensitivity degradation correction fator")
               call addcomment(outTable, "TDS_CORR is the correction factor applied to the corrected count-rates") 
               call addcomment(outTable, "to account for the OM sensitivity degradation") 
               
               call addComment(outTable, " ")
               call addComment(outTable,  sourceShapeComment)
               call addComment(outTable, backShapeComment)
               if (clcComment1 .ne. " ") call addComment(outTable, clcComment1)
               if (clcComment2 .ne. " ") call addComment(outTable, clcComment2)

               write(errstr,'(a,f6.3)') "Sensitivity degradation correction factor: ",tdpCorrFactor
               call message(AppMsg, 2)

               if (minval(rateSourceOutArray) < 0.) then
                  call warning("negVals", "Background subtracted rate has negative values")
               endif
               
               ! release the dal handles
               call release(rateBackErrOutCol)
               call release(rateBackOutCol)
               call release(rateSourceErrOutCol)
               call release(rateSourceOutCol)
               call release(timeSourceOutCol)
               call release(outTable)
               call release(outSet)
               call release(backYPosCol)
               call release(backXPosCol)
               call release(backRadiusCol)
               call release(backShapeCol)
               call release(backRegionTable)
               if (bkgRateExists) then
                  call release(rateBackCol)
                  call release(backTable)
               endif
               call release(backSet)
            else ! hasBlock(sourceSet, "RATE")
               call warning("Rate", "Block RATE does not exist in the source rates file")
            endif ! hasBlock(sourceSet, "RATE")
            call release(sourceYPosCol)
            call release(sourceXPosCol)
            call release(sourceRadiusCol)
            call release(sourceShapeCol)
            call release(regionTable)
            call release(rateSourceCol)
            call release(sourceTable)
            call release(sourceSet)

            deallocate(rateSourceArray)
            deallocate(timeSource)
            if (bkgRateExists) then
               deallocate(rateBackArray)
               deallocate(timeBack)
            endif
            deallocate(timeOut)
            deallocate(sourceArrayError)
            deallocate(backArrayError)         

            
            !---Adding columns to SRC list
            !****************************************************************
            ! Add the psf1Correction, psf2Correction, and magnitudes
            !**************************************************************** 
            if (inColumnFlag(16) .eq. 0) then
               ! check if CORR_BACKGROUND_RATE column exists
               if (inColumnFlag(17) .eq. 1) then
                  position = columnNumber(inTableH, "CORR_BACKGROUND_RATE")+1
               else
                  if (inColumnFlag(20) .eq. 1) then
                     position = columnNumber(inTableH, "CORR_RATE_ERR")+1
                  else
                     if (inColumnFlag(19) .eq. 1) then
                        position = columnNumber(inTableH, "CORR_RATE")+1
                     else
                        if (inColumnFlag(18) .eq. 1) then
                           position = columnNumber(inTableH, "SIGNIFICANCE")+1
                        else
                           position=11
                        endif
                     endif
                  endif
               endif
               
               call AddReal32Column(inTableH, psf1Correction, nSources, "PSF1_CORR", &
                    "%", "PSF1 correction (extraction radius to 12 pixels)", nullValue, position)
            endif
            
            position = columnNumber(inTableH, "PSF1_CORR")+1     
            call AddReal32Column(inTableH, psf2Correction, nSources, "PSF2_CORR", &
                 "%", "PSF2 correction (12 pixels to cal radius)", nullValue, position )

           
            ! protection from the absence of the COINCLOSS_CORR column
            if (hasColumn(inTableH, "COINCLOSS_CORR")) then
               position=columnNumber(inTableH, "COINCLOSS_CORR")+1
               
            else
               position=columnNumber(inTableH, "PSF2_CORR")+1
               
            endif        
            
            call AddReal32Column(inTableH, mag, nSources, "MAG", &
                 "mag", "Instrumental magnitude", nullValue, position)
 
            !****************************************************************
            ! Add the magnitude errors
            !**************************************************************** 
            call AddReal32Column(inTableH, magErr, nSources, "MAGERR", &
                 "mag", "Instrumental magnitude error", nullValue, position + 1)
            
            !****************************************************************
            ! Add chi-squared 
            !**************************************************************** 
            call AddReal32Column(inTableH, chi2, nSources, "CHI2", &
                 "", "Chi-square variability test", nullValue, position + 2)
            call AddReal32Column(inTableH, chi2prob, nSources, "CHI2_PROB", &
                 "", "Chi-square variability probability", nullValue, position + 3)
            !****************************************************************
            ! Add Kolmogorov-Smirnov test statistics 
            !**************************************************************** 
            
            call AddReal32Column(inTableH, ksStat, nSources, "K-S", &
                 "", "Kolmogorov-Smirnov variability test", nullValue, position + 4)
            call AddReal32Column(inTableH, ksStatProb, nSources, "K-S_PROB", &
                 "", "Kolmogorov-Smirnov variability probability", nullValue, position + 5)
            ! Add maximal deviation column
             call AddReal32Column(inTableH, maxDev, nSources, "MAXDEV", &
                 "sigma", "Maximal deviation from the mean", nullValue, position + 6)           
             
            if (inColumnFlag(3) .eq. 0) then
               deallocate(mag)
            endif
            if (inColumnFlag(4) .eq. 0) then
               deallocate(magErr)
            endif
            if (inColumnFlag(16) .eq. 0) then
               deallocate(psf1Correction)
            endif
            if (inColumnFlag(24) .eq. 0) then
               deallocate(chi2)
            endif
            if (inColumnFlag(25) .eq. 0) then
               deallocate(chi2prob)
            endif
            if (inColumnFlag(26) .eq. 0) then
               deallocate(ksStat)
            endif
            if (inColumnFlag(27) .eq. 0) then
               deallocate(ksStatProb)
            endif
            if (inColumnFlag(28) .eq. 0) then
               deallocate(maxDev)
            endif

            deallocate(psf2Correction)
            
            !******
            !*************************
            ! Release handles on data
            !*************************
            do i=1, noColumns
               if (inColumnFlag(i) .eq. 1) then
                  call release(inColumnH(i))
               endif
            end do
            ! VNY 2009.05.18 NB! Here is the task can finish silently 
            ! without generating an output (which is wrong)
         else
            
!*********************************************************************
! VNY 2009.05.18
!*********************************************************************
            write (errstr, *) "The number of sources (",nSources,  ") is too large (no output files produced)"
            call message(AppMsg, 2)
!*********************************************************************
! VNY 2009.05.18
!*********************************************************************
         endif ! nSources>0 .an. nSource .lt. 2000
         !if (allocated(backgroundRate)) then
         !   deallocate(backgroundRate)
         !endif
         
         if (allocated(backgroundRateSafe)) deallocate(backgroundRateSafe)

         call release(inTableH)
      else ! hasBlock(inDataSetH)
         !************************************
         ! Incompatible file - can't continue
         !************************************
         write (errstr, *) "Could not find any compatibile blocks in set ", sourceList
         call message(AppMsg, 2)

      endif ! hasBlock(inDataSetH)

      call release(inDataSetH)
      call release(sourceRegSet)
      call release(backRegSet)
!---VNY test file 26.02.2004
!    close(2)
!----------
    end subroutine omlcbuild

!*******************************************************************
! 2009-10-07 Copying the C++ routinefluxCorrect in  OmFluxConverter.cc from cal/src
!*******************************************************************
type(ValueAsymmetricErrorT) function test_omPhotoNatural(rawRate,frameTime, &
     deadFrac, countLoss, extended, empirical) result(res)
  type(ValueAsymmetricErrorT), intent(in) :: rawRate
  real(kind=double), intent(in) :: frameTime
  real(kind=double), intent(in) :: deadFrac
  real(kind=double), intent(in) :: countLoss
  logical, intent(in) :: extended
  logical, intent(in) :: empirical

  real(kind=double) :: rate,f, v, x, phdFrac
  !type(ValueAsymmetricErrorT) :: res

  phdFrac=1.0
  rate=rawRate%value
  res=rawRate

  if (.not. extended) then
     if (frameTime .eq. 0.0) then
        !write(*,*) "L3022 illegal frame time"
        write(errstr,'(a)') "Illegal frame time (0) "
        call message(AppMsg, 2)
     else
        if (rate .lt. 1.0/frameTime) then
           f = 1.0/(1.0-rate*frameTime)
           rate = -log(1.0-rate*frameTime)/frameTime
           res%errorPlus = f*rawRate%errorPlus
           res%errorMinus = f*rawRate%errorMinus
        else
           res%value=1.0d20
           res%errorPlus = 1.0d20 
           res%errorMinus = 1.0d20
           !write(*,*) "L3033 warning: Illegal count rate"
           write(errstr,'(a)') "Illegal count rate "
           call message(AppMsg, 2)
        endif ! rate .lt. 1.0/frameTime
        !	apply empirical correction if requested
        if (empirical) then
           x=rawRate%value*frameTime
           !*********************************************************
           ! The c++ algorithm reproduces the formula:
           ! 1.0+0.075673*x-0.143986*x**2+0.638497*x**3-0.570184*x**4
           !*********************************************************
           v=-0.570184
           v=v*x+0.638497
           v=v*x-0.143986
           v=v*x+0.075673
           v=v*x+1.0
           rate=rate*v
        endif ! empirical
     endif ! frameTime eq 0.0
  endif ! not extended
  !**************************************
  ! Correct for deadtime
  !**************************************
  if (deadFrac .eq. 1.0) then
     !write(*,*) "L3056 Illegal dead fraction"
     write(errstr,'(a)') "Illegal dead fraction (1.0)  "
     call message(AppMsg, 2)
  else
     rate=rate/(1.0-deadFrac)
  end if ! deadFrac eq 1.0
!**********************************************
! 	correction for average detector aging
!**********************************************
  if (countLoss .ne. 0.0) then
     if (countLoss .eq. 1.0) then
        !write(*,*) "L3065 Illegal event loss rate"
        write(errstr,'(a)') "Illegal event loss rate (1.0) "
        call message(AppMsg, 2)
     else
        rate=rate/(1.0-countLoss)*phdFrac
     endif ! countLoss eq 1
  endif ! countLoss .not. eq 0

  res%value=rate


end function test_omPhotoNatural

!*******************************************************************
! 2009-10-07 Copying the C++ routinefluxCorrect in  OmFluxConverter.cc from cal/src
! 2009-10-08 Trying to reduce the data jumping by dividing deadFrac by
! (rawCount*frameTime)**3
!*******************************************************************
type(ValueAsymmetricErrorT) function test2_omPhotoNatural(rawRate,frameTime, &
     deadFrac, countLoss, extended, empirical) result(res)
  type(ValueAsymmetricErrorT), intent(in) :: rawRate
  real(kind=double), intent(in) :: frameTime
  real(kind=double), intent(in) :: deadFrac
  real(kind=double), intent(in) :: countLoss
  logical, intent(in) :: extended
  logical, intent(in) :: empirical

  real(kind=double) :: rate,f, v, x, phdFrac, x3
  !type(ValueAsymmetricErrorT) :: res

  phdFrac=1.0
  rate=rawRate%value
  res=rawRate

  x=rawRate%value*frameTime
  !x3=x*x*x
  !x3=1.0-x*x*x*x
  x3=1.0+2.8*(deadFrac-0.017)
  !write(*,*) "L3102 x3=", x3
  if (.not. extended) then
     if (frameTime .eq. 0.0) then
        !write(*,*) "L3022 illegal frame time"
        write(errstr,'(a)') "Illegal frame time "
        call message(AppMsg, 2)
     else
        if (rate .lt. 1.0/frameTime) then
           f = 1.0/(1.0-rate*frameTime)
           !************************************************************
           ! Correction 2009-10-09 to avoid exposure-toexposure jumping
           !************************************************************
           !rate=rate*x3
 
           rate = -log(1.0-rate*frameTime)/frameTime

           res%errorPlus = f*rawRate%errorPlus
           res%errorMinus = f*rawRate%errorMinus
        else
           res%value=1.0d20
           res%errorPlus = 1.0d20 
           res%errorMinus = 1.0d20
           !write(*,*) "L3033 warning: Illegal count rate"
           write(errstr,'(a)') "Ileegal count rate  "
           call message(AppMsg, 2)
        endif ! rate .lt. 1.0/frameTime
        !	apply empirical correction if requested

        if (empirical) then
           !*********************************************************
           ! The c++ algorithm reproduces the formula:
           ! 1.0+0.075673*x-0.143986*x**2+0.638497*x**3-0.570184*x**4
           !*********************************************************
           v=-0.570184
           v=v*x+0.638497
           v=v*x-0.143986
           v=v*x+0.075673
           v=v*x+1.0
           rate=rate*v
        endif ! empirical
     endif ! frameTime eq 0.0
  endif ! not extended
  !**************************************
  ! Correct for deadtime
  !**************************************
  if (deadFrac .eq. 1.0) then
     !write(*,*) "L3056 Illegal dead fraction"
     write(errstr,'(a)') "Illegal dead fraction (1.0)  "
     call message(AppMsg, 2)
  else
     rate=rate/(1.0-deadFrac)
     !if (x3 .gt. 0.0) then
     !   !rate=rate/(1.0-deadFrac/x3)
     !   rate=rate/(1.0-deadFrac*(1.0+x3))
     !else
     !   rate=rate/(1.0-deadFrac)
     !endif
  end if ! deadFrac eq 1.0
!**********************************************
! 	correction for average detector aging
!**********************************************
  if (countLoss .ne. 0.0) then
     if (countLoss .eq. 1.0) then
        !write(*,*) "L3065 Illegal event loss rate"
        write(errstr,'(a)') "Illegal event loss rate  "
        call message(AppMsg, 2)
     else
        rate=rate/(1.0-countLoss)*phdFrac
     endif ! countLoss eq 1
  endif ! countLoss .not. eq 0

  res%value=rate


end function test2_omPhotoNatural

!*********************************************************
! Function CorrectedCountRate
! Corrects the raw count rate for coincidence losses, etc  
!*********************************************************
real(kind=double) function CorrectedCountRate(rawCountRate, & 
                  countRateError, frameTime, deadFrac, extended, &
                  empirical, countRateErrorMinus, countRateErrorPlus)
    real(kind=double), intent(in)    :: rawCountRate
    real(kind=double), intent(in) :: countRateError
    real(kind=double), intent(out)    :: countRateErrorMinus
    real(kind=double), intent(out)    :: countRateErrorPlus
 
    real(kind=double), intent(in) :: frameTime
    real(kind=double), intent(in) :: deadFrac
    logical, intent(in) :: extended
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: correctedRate
    logical, intent(in) :: empirical

    rawRate = ValueAsymmetricErrorT(rawCountRate, dble(countRateError), dble(countRateError))
    !correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, 0.0d0, extended)
    
    correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, 0.0d0, extended, empirical)

    ! VNY 24.04.2008 temporarily using the parameter eventLoss (previously set to 0.0d0) 
    ! as a flag for test printing. The flag value "true" will be 0.01d0
    !correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, 0.01d0, extended, empirical)

    !****************************************************************
    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    !*****************************************************************
    if (CorrectedRate%value .lt. 1e10) then
        CorrectedCountRate = CorrectedRate%value
        countRateErrorMinus = correctedRate%errorMinus
        countRateErrorPlus = correctedRate%errorPlus
    else
       !***************************************************************
       !---VNY 01.11.2001 For the case of CAL_omPhotoNatural would fail
       !     the count rate are left without correction
       !***************************************************************
       if (rawCountRate .lt. 1e10) then
          CorrectedCountRate = rawCountRate
          countRateErrorMinus = countRateError
          countRateErrorPlus = countRateError
       else
          CorrectedCountRate = -1.0d0
          countRateErrorMinus = -1.0d0
          countRateErrorPlus = -1.0d0
       end if
    end if
end function CorrectedCountRate

real(kind=double) function CorrectedCountRateTest(rawCountRate, & 
                  countRateError, frameTime, deadFrac, extended, empirical, &
                  countRateErrorMinus, countRateErrorPlus, iflag)
    real(kind=double), intent(in)    :: rawCountRate
    real(kind=double), intent(in) :: countRateError
    real(kind=double), intent(out)    :: countRateErrorMinus
    real(kind=double), intent(out)    :: countRateErrorPlus
 
    real(kind=double), intent(in) :: frameTime
    real(kind=double), intent(in) :: deadFrac
    logical, intent(in) :: extended
    logical, intent(in) :: empirical
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: correctedRate
    real(kind=double) :: countLoss
    integer, intent(in) :: iflag

    rawRate = ValueAsymmetricErrorT(rawCountRate, dble(countRateError), dble(countRateError))
    countLoss=0.0d0
    !if (iflag .eq. 49) then
    !   countLoss=0.01d0
    !endif

    !correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, countLoss, extended, &
    !     empirical)

    !correctedRate = test_omPhotoNatural(rawRate, frameTime, deadFrac, countLoss, extended, &
    !     empirical)
    correctedRate = test2_omPhotoNatural(rawRate, frameTime, deadFrac, countLoss, extended, &
         empirical)

    !correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, 0.0d0, extended, &
    !     empirical)

    !******* Test printing ********
    !if (iflag .eq. 49) then
    !   write(*,*) "L2682 rawRate=", rawRate, " correctedRate%value=", correctedRate%value, &
    !        " frameTime=", frameTime
    !endif
    !*****************************

    !****************************************************************
    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    !*****************************************************************
    if (CorrectedRate%value .lt. 1e10) then
        CorrectedCountRateTest = CorrectedRate%value
        countRateErrorMinus = correctedRate%errorMinus
        countRateErrorPlus = correctedRate%errorPlus
    else
       !***************************************************************
       !---VNY 01.11.2001 For the case of CAL_omPhotoNatural would fail
       !     the count rate are left without correction
       !************************s***************************************
       if (rawCountRate .lt. 1e10) then
          CorrectedCountRateTest = rawCountRate
          countRateErrorMinus = countRateError
          countRateErrorPlus = countRateError
       else
          CorrectedCountRateTest = -1.0d0
          countRateErrorMinus = -1.0d0
          countRateErrorPlus = -1.0d0
       end if
    end if
end function CorrectedCountRateTest

subroutine dlocate(xx, n, x, j)
! VNY 25.01.2002; If the dimension of the array xx is 1 or 0
! this subroutin can crash due to subscript. Protection 
! against this crash has been added
   integer(kind=int32), intent(in) :: n
   real(kind=double)   :: xx(0 : n - 1)  
   real(kind=double)   :: x
   integer(kind=int32) :: j
   integer(kind=int32) :: ju
   integer(kind=int32) :: jl
   integer(kind=int32) :: jm

   jl = 0
   ju = n + 1

   ! VNY 25.01.2002 Protection against a one-element array
   if (n .le. 1) then
      j = 0
      return
   endif

   if (x .ge. xx(n - 1)) then
       j = n - 1
       return
   end if 
      
   if (x .le. xx(0)) then
      j = 0
      return
   endif

   do while(ju - jl > 1)
      jm = (ju + jl) / 2
      if (x .gt. xx(jm - 1)) then
         jl = jm
      else
         ju = jm
      end if
   end do

   j = max(jl - 1, 0)
   
end subroutine dlocate

!****************************************************************
! 2012-10-18 excluding this subroutine as it is defined also in 
! the module omskybackground_mod
!*****************************************************************
!subroutine AddReal32Column(tableSet, array, size, heading, &
!        units, comment, nullValue, position)
!    type(TableT), intent(in) :: tableSet
!    character(len=*), intent(in) :: heading, units, comment
!    integer, intent(in) :: size
!    real(kind=single), intent(in), dimension(0 : size - 1) :: array
!    real(kind=single), intent(in) :: nullValue
!    integer(kind=int32), intent(in), optional :: position
!    real(kind=single), pointer :: real32Col(:)
!    type(ColumnT) :: pColumn
!    integer i
!    integer :: ndims(1) = (/1/)!
!
!     if (.not. hasColumn(tableSet, trim(heading))) then
!        if (present(position)) then
!            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment), position)
!        else 
!            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment))
!        end if
!    else
!        pColumn = column(tableSet, trim(heading), modify)
!    end if
!   
!    real32Col => real32Data(pcolumn)
!
!    do i = 0 , size - 1
!        if (array(i) .eq. nullValue) then
!            call setToNull(pcolumn, i)
!        else
!            real32Col(i) = array(i)
!        end if
!    enddo
!    call release(pcolumn)!
!
!end subroutine AddReal32Column

!------------------------------------------------
! Simulate count rate
! 25.03.2004
!------------------------------------------------
!subroutine simulateCountRate(frameTime, timeBin, trueCountRate, &
!     lossCountRate, simulatedCountRate, dispersionLeft, dispersionRight)
!  real, intent(in) :: frameTime
!  real, intent(in) :: timeBin
!  real, intent(in) :: trueCountRate ! counts per second
!  real, intent(out) :: lossCountRate
!  real, intent(out) :: simulatedCountRate
!  real, intent(out) :: dispersionLeft
!  real, intent(out) :: dispersionRight!

!  real :: trueFluxPerFrame
!  integer :: nFrames
!  integer :: seedSize
!  integer, allocatable :: seedArray(:)
!  integer, allocatable :: trueCountsArray(:)
!  integer, allocatable :: lossCountsArray(:)
!  integer :: i, ival
!  logical first
!  real :: trueMean, lossMean, nMean
!  real :: nMeanLeft, nMeanRight
!  real :: diff
!  real :: trueRate, lossRate!

!  nFrames=int(timeBin/frameTime)
!  allocate(trueCountsArray(nFrames))
!  allocate(lossCountsArray(nFrames))


!!  write(*,*) " L2356 nFrames=", nFrames
!  ! VNY 26,03,2004
!  ! test Poissonian generator
!!  write(*,*) " Test Poissonian generator"
!!  write(*,*) "--------------------------"
!!  trueFluxPerFrame=100
!!  first=.true.
!!  ival=random_Poisson(trueFluxPerFrame, first)
!!  first=.false.
!!  trueCountsArray=0
!!  trueMean=0.0
!!  nMean=0.0
!!  do i=1, nFrames
!!     ival=random_Poisson(trueFluxPerFrame, first)
!!     trueCountsArray(i)=ival
!!     trueMean=trueMean+ival
!!     nMean=nMean+1.0
!!  enddo
!!  trueRate=trueMean/nMean
!!  dispersionLeft=0.0
!!  nMeanLeft=0.0
!!  do i=1, nFrames
!!     diff=trueCountsArray(i)-trueRate
!!     dispersionLeft=dispersionLeft+diff*diff
!!     nMeanLeft=nMeanLeft+1.0
!!  enddo
!!  dispersionLeft=sqrt(dispersionLeft/(nMeanLeft-1))
!!  write(*,*) " nFrames=", nFrames, " meanRate=", trueRate, " meanDisp=", dispersionLeft
!!  write(*,*) "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
!!----
!  trueCountsArray=0
!  lossCountsArray=0
!  trueMean=0.0
!  lossMean=0.0
!  nMean=0.0
!  trueFluxPerFrame=trueCountRate*frameTime
!  first=.true.
!  ival=random_Poisson(trueFluxPerFrame, first)
!  first=.false. 
!  do i=1, nFrames
!     ival=random_Poisson(trueFluxPerFrame, first)
 !    trueCountsArray(i)=ival
!     trueMean=trueMean+real(ival)
!     if (ival .gt. 0) then
!        lossCountsArray(i)=1
!        lossMean=lossMean+1.0
!     endif
!     nMean=nMean+1.0
!  enddo
!  trueRate=trueMean/timeBin
!  lossRate=lossMean/timeBin
!!  write(*,*) " L2387 trueRate=", trueRate, " lossRate=", lossRate 
!  if (nMean .gt. 0.0) then
!     trueMean=trueMean/nMean
!     lossMean=lossMean/nMean
!  endif
!!  write(*,*) " L2392 nMean=", nMean, " trueMean=", trueMean, &
!!       " lossMean=", lossMean
!!  simulatedCountRate=trueMean
!!  lossCountRate=lossMean
!  simulatedCountRate=trueRate
!  lossCountRate=lossRate
!  dispersionLeft=0.0
!  dispersionRight=0.0
!  nMeanLeft=0.0
!  nMeanRight=0.0
!  ! compute dispersion (to the left and to the right from the maximum
!  do i=1, nFrames
!     diff=trueCountsArray(i)/frameTime-trueRate
!     !write(*,*) " i=", i, " true=", trueCountsArray(i)/frameTime, &
!     !     " loss=", lossCountsArray(i)/frameTime, " diff=", diff
!     if (diff .lt. 0.0) then
!        dispersionLeft=dispersionLeft+diff*diff
!        nMeanLeft=nMeanLeft+1.0
!     endif
!     if (diff .gt. 0.0) then
!        dispersionRight=dispersionRight+diff*diff
!        nMeanRight=nMeanRight+1.0
!     endif
!  enddo
!  if (nMeanLeft .gt. 0.0) then
!     dispersionLeft=sqrt(dispersionLeft/(nMeanLeft-1.0))
!  else
!     dispersionLeft=sqrt(dispersionLeft)
!  endif
!  if (nMeanRight .gt. 0.0) then
!     dispersionRight=sqrt(dispersionRight/(nMeanRight-1.0))
!  else
!     dispersionRight=sqrt(dispersionRight)
!  endif!
!
!  deallocate(trueCountsArray)
!  deallocate(lossCountsArray)!

!end subroutine simulateCountRate

!---------------------------------------
! Simon's coincidence loss test routine
! 27.05.2004
!---------------------------------------
subroutine coincLossAsymErr(frameTime, timeBin, trueCountRate, &
     lossCountRate, dispersionLeft, dispersionRight, rmsValue, verbFlag)
  real, intent(in) :: frameTime
  real, intent(in) :: timeBin
  real, intent(in) :: trueCountRate ! counts per second
  real, intent(out) :: lossCountRate
  real, intent(out) :: dispersionLeft
  real, intent(out) :: dispersionRight
  real, intent(out) :: rmsValue
  integer, intent(in) :: verbFlag
  integer :: nFrames

!*****************************************************************
!  character(len=50) fout

  real(kind=double) :: pi, mu, mu_eff, ln_p, p
  real(kind=double) :: ln_k_fact, ln_m_fact, ln_k_m_fact, sum_p
!  real(kind=double) :: p_poiss, ln_p_poiss, sum_p_poiss, mu_poiss
  real(kind=double) :: mu_poiss
  !real(kind=double) :: rms, mean, r2, errl, errh, devl, devh, cl_factor
  real(kind=double) :: rms, mean, r2, errl, errh, devl, devh
  !real(kind=double) :: cl_formula, mean_poiss, sigma_poiss, sum_p_l, sum_p_h
  real(kind=double) :: cl_formula, sum_p_l, sum_p_h
  real(kind=double) :: ln_fixed_term, crfrr
  real(kind=double) :: ec(5), emp_corr, fco

!  integer(kind=int32) :: k, m, i, max
  integer(kind=int32) :: k, m, i

  real(kind=double) :: ci, t ! incident CR (cts/s) and frame rate (s)
  real(kind=double) :: co_best, diff, ci_pred_best
!  real(kind=double) :: ci_low, ci_hi, co_low
  real(kind=double) :: co, ci_pred, muo
  logical  use_empirical

! VNY 14.06.2004 from Simon's code:
! insert empirical coincidence loss coefficients, from 
! ~/sasccf/OM_PHOTTONAT_0003.CCF. Coefficients work on the co*t (raw CRFR 
! ratio).
  ec(1)=1.0d0
  ec(2)= 0.0756730d0
  ec(3)=-0.1439860d0
  ec(4)= 0.6384970d0
  ec(5)=-0.5701840d0

  pi=4.d0*atan(1.d0)

  nFrames=int(timeBin/frameTime)

  ci=trueCountRate
  t=frameTime

  mu_poiss=ci ! incident counts during one second
  mu=ci*t ! incident counts during one frame
  crfrr=ci*t ! INCIDENT CRFR ratio
  cl_formula=(1.d0-exp(-mu))/frameTime !t
  k=nFrames ! total number of frames per exposure time (timeBin)


! There is a PROBLEM to use the empirical curve directly here in this 
! test code because the empirical correction uses a polynomial based on 
! the observed, not incident count rate so have to iteratively find what 
! observed CR gives incident count rate provided by user. Quick solution 
! here where we have ci, not co, initially, is to simply compute ci for 
! a range of co*T (=muo =CRFRobserved values) and find value of co which 
! gives Ci (empirical) closest to input value

      diff=1.e10
      ci_pred_best=-1.0

      do i=1,100000

         muo=(i-1)/100000.0
         co=muo/t

         emp_corr=ec(1) + ec(2)*muo + ec(2)*muo*muo + ec(3)*muo**3 + & 
          ec(4)*muo**4 + ec(5)*muo**5


         ci_pred=-log(1.d0-muo)/t
         ci_pred=ci_pred*emp_corr


         if(abs(ci_pred - ci).lt.diff)then
            diff=abs(ci_pred-ci)
            co_best=co
            ci_pred_best=ci_pred
         end if

      end do

! compute correction factor f(Co) that is thus used to convert this Co to Ci
! Will be used lower down if empirical correction is applied. 

      muo=co_best*t
      fco=ec(1) + ec(2)*muo + ec(2)*muo*muo + ec(3)*muo**3 + & 
        ec(4)*muo**4 + ec(5)*muo**5
      !if (verbFlag .eq. 1) then 
      !   write(*,'(/,"Based on empirical CL function, trials show that an &
      !        &observed rate ",/," of Co= ",f10.3," would be observed for an &
      !        &incident rate of Ci= ",f10.3,/," compared to the user specified &
      !        &rate ci(input) = ",f10.3)')co_best,ci_pred_best,ci
      !   if(ci_pred_best.lt.0.0)then
      !      write(*,*)' NO BEST EMPIRICAL SOLN FOUND'
      !      !call exit
      !   end if
      !endif

! First compute the probability distribution of the incident count rate, for 
! reference
!
!!  sum_p_poiss=0.d0
!!  mean_poiss=0.d0
!!  sigma_poiss=0.d0

!!  max=int(5*Ci)
!!  if (ci.lt.10) max=30
!!  do i=0,max
!!     m=i
!!     ln_m_fact=calc_ln_fact(m)
!!     ln_p_poiss = m*log(mu_poiss) - mu_poiss - ln_m_fact
!!     p_poiss = exp(ln_p_poiss)
!!     if(p_poiss.lt.1.d-30)p_poiss=0.d0
!!     sum_p_poiss = sum_p_poiss + p_poiss
!!     mean_poiss=mean_poiss+p_poiss*m
!!  end do
!!  if (verbFlag .eq. 1) write(*,*) " L2556 mean_poiss=", mean_poiss
!
! As a check, compute sigma using sqrt[sum(p*(xi-X)**2] rather than 
! sqrt[bar(r**2) - (bar(r))**2]
!
  devl=0.d0
  devh=0.d0
  sum_p_l=0.d0
  sum_p_h=0.d0

!!  do i=0,max
!!     m=i
!!     ln_m_fact=calc_ln_fact(m)
!!     ln_p_poiss = m*log(mu_poiss) - mu_poiss - ln_m_fact
!!     p_poiss = exp(ln_p_poiss)
!!     if(p_poiss.lt.1.d-30)p_poiss=0.d0
!!     sigma_poiss=sigma_poiss + p_poiss * (1.d0*m-mean_poiss)**2
!!     if(m.le.mean_poiss)then
!!        devl=devl + p_poiss*(m-mean_poiss)**2
!!        sum_p_l=sum_p_l+p_poiss
!!     end if
!!     if(m.gt.mean_poiss)then
!!        devh=devh + p_poiss*(m-mean_poiss)**2
!!        sum_p_h=sum_p_h+p_poiss
!!     end if
!!  end do

!      write(1,'(//)')

!  if (verbFlag .eq. 1) then
!     write(*,*) " For input Poisson distribution "
!     write(*,*) " ------------------------------ "
!     write(*,*) " L2587 mean_poiss=", mean_poiss, " sqrt(sigma_poiss)=", &
!          sqrt(sigma_poiss)
!     write(*,*) " Sum of P(n,mu)=", sum_p_poiss, " (should be 1.0)"
!     write(*,*) " devl=", devl, " devh=", devh, " sum_p_l=", sum_p_l, &
!          " sum_p_h=", sum_p_h
!     write(*,*) " sqrt(devl/sum_p_l)=", sqrt(devl/sum_p_l), &
!          " sqrt(devh/sum_p_h)=", sqrt(devh/sum_p_h)
!  endif

!  compute the probability of getting m=i counts if we run k frames with an 
!  incident rate of Ci counts/s and a frame time of t seconds. This 
!  probability is based on kCm * P(0)**(k-m) * P(>=1)**m 
!
!  where
!
!  kCm = k!/(m!)(k-m)!
!
!

!
!  switch for  using empirical correction 
!
  use_empirical=.true.

  sum_p=0.d0
  rms=0.d0
  r2=0.d0
  mean=0.d0

!----------- 15.06.2004
  mu_eff=mu
  if(use_empirical)then
     mu_eff=mu/fco   !compute effective incident mu
     cl_formula=co_best
  end if
  ln_fixed_term=log(1.d0-exp(-mu_eff)) !this term if fixed for each k below
!-----------

  ln_k_fact=calc_ln_fact(k)        ! constant for given no of frames

  do i=0,k  ! m can not exceed k
     m=i        
     ln_m_fact=calc_ln_fact(m)
     ln_k_m_fact=calc_ln_fact(k-m)
     ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact
!     ln_p = ln_p - (k-m)*mu + m*log(1.d0-exp(-mu))
     ln_p = ln_p - (k-m)*mu_eff + m*ln_fixed_term 
     p=exp(ln_p)
     if(p.lt.1.d-30)p=0.d0    ! set to zero if P is very small. 
     mean=mean+p*m
     r2=r2 + p*m*m
     sum_p=sum_p+p
  end do

!--- 15.06.2004
! rounding errors and Stirlings approx give rise to value of sum_p which is 
! very slightly > 1.000. Thus, normalise sum(r2) amd mean by sum_p as if to 
! make them equivalent to having come from a distribution with sum_p=1.0. 
! This appears to remove potential problem of sum(r2)-mean*mean being -ve. 

      r2=r2/sum_p         
      mean=mean/sum_p     

      rms=r2-mean*mean    ! rms (square of)
      !if (verbFlag .eq. 1) then
      !   if(use_empirical)then
      !      write(*,'(" For the coincidence loss affected (truncated) &
      !           &distribution (WITH EMPIRICAL CORRECTION INCLUDED)",/)')
      !   else
      !      write(*,'(" For the coincidence loss affected (truncated) &
      !           &distribution",/)')
      !   end if!
      ! 
      !   write(*,'("Sum of P(k,m,mu) (should be 1.0) &
      !        & = ",f12.8)')sum_p
      !   write(*,'("Mean of observed cts = ",f12.4," cf cl formula &
      !        &mean = ",f12.4)')mean, cl_formula*k*t
      !endif
!----

!!  rms=r2-mean*mean    ! rms (square of)
!!  if(rms.ge.0.d0)then
!!     rms=sqrt(rms)
!!  else
!!     if (verbFlag .eq. 1) write(*,*)'r2-mean*mean < 0 !'
!!  end if
!!  cl_factor=ci*frameTime*k/mean
!
! now compute asymmetric errors - For asymmetric errors need to know 
! what mean is first, so must compute above loop first and then repeat to 
! get deviations from mean, weighted by P
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  errl=0.d0
  errh=0.d0
  devl=0.d0
  devh=0.d0
  rms=0.d0
  do i=0,k     ! m can not exceed k
     m=i
     ln_m_fact=calc_ln_fact(m)
     ln_k_m_fact=calc_ln_fact(k-m)
     ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact
     ln_p = ln_p - (k-m)*mu + m*log(1.d0-exp(-mu))
     p=exp(ln_p)
     if (p.lt.1.d-30) p=0.d0
     rms=rms+p*(m-mean)**2
     if (m.le.mean) devl=devl + p*(m-mean)**2
     if (m.gt.mean) devh=devh + p*(m-mean)**2
  end do
  errl=sqrt(devl)
  errh=sqrt(devh)
  !if (verbFlag .eq. 1) write(*,*) "L2765  devl=", devl, "errl=", errl, " devh=", devh, " errh=", errh, " rms=", rms
  if (rms .gt. 0.d0) then
     rms=sqrt(rms)
  else
     if (verbFlag .eq. 1) write(*,*) " L2771 rms is negative"
  endif

  rmsValue=rms/timeBin
  dispersionLeft=errl/timeBin
  dispersionRight=errh/timeBin
  lossCountRate=mean/timeBin
end subroutine coincLossAsymErr

!-------------------------------------
! log of factorial
!-------------------------------------
real(kind=double) function calc_ln_fact(i)      
  integer(kind=int32) :: i, j
  real(kind=double) :: ln_i_fact, pi, i_fact
  pi=4.d0*atan(1.d0)    !3.141592654d0
  if(i.le.50)then
     j=1
     i_fact=1.d0      ! set for 0 factorial
     do while(j.le.i)
        i_fact=i_fact*j
        j=j+1
     end do
     ln_i_fact=log(i_fact)   ! convert to a log
  else   !use Stirling's approximation
     ln_i_fact=0.5d0*log(2.d0*pi*i) + 1.d0*i*(log(1.d0*i)-1.d0)
  end if
  calc_ln_fact=ln_i_fact
  return
end function calc_ln_fact
! End of Simon's code
!*****************************************************************

!--------------------------------------
! Initialise the random number generator
!--------------------------------------
subroutine initRandomSequence(initNumber)
  integer, intent(in) :: initNumber
  integer :: seedSize, i
  integer, allocatable :: seedArray(:)
  call random_seed(SIZE=seedSize)
  allocate(seedArray(seedSize))
  do i=1,seedSize
     seedArray(i)=i*initNumber
  enddo
  call random_seed(PUT=seedArray)
  deallocate(seedArray)
end subroutine initRandomSequence

!*******************************************************
! ModifiedJulianDate - Returns the modified Julian date
! from the string date representation
!*******************************************************
!****************************************************************
! 2012-10-18 excluding this subroutine as it is defined also in 
! the module omskybackground_mod
!*****************************************************************
!real(double) function ModifiedJulianDate(dateString)
!    character(*), intent(in) :: dateString
!
!    integer :: year            ! Stores the year from the date string
!    integer :: month           ! Stores the month from the date string
!    integer :: day             ! Stores the day from the date string  
!    integer :: hour            ! Stores the hour from the date string
!    integer :: minute          ! Stores the minute from the date string  
!    integer :: second          ! Stores the second from the date string
!    character(len = 1) char    ! Used for date and time extraction 
!    ! from string representation
!    real(double) JulianDay    
!    integer :: flag
!
!    read(dateString, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
!         year, char, month, char, day, char, hour, &
!         char, minute, char, second     
!
!    ! Get the modified Julian date for this date
!    call sla_CLDJ(year, month, day, JulianDay, flag)
!
!    ! Need to add on the fraction of a day corresponding to the time
!    ModifiedJulianDate = JulianDay + (dble(hour) + &
!         (dble(minute) + dble(second) / 60.0d0) / 60.0d0) / 24.0d0
!end function ModifiedJulianDate


!**********************************************************************************
!
! Returns the time-dependent sensitivity correction factor
!
! The correction factor is Y = a + b * mjd, where mjd is the modified Julian date
! The corrected count-rates are given by corrected rate = measured rate * Y
!**********************************************************************************

real(single) function TDPCorrectionFactorOld(mjd, filter)
    real(double), intent(in) :: mjd  ! Modified Julian day
    character(*), intent(in) :: filter
    real(single) :: coeffs(2)

    select case (trim(filter))
        case ("U")
            coeffs = (/-0.313603, 2.51781e-05/)
        case ("B")
            coeffs = (/-0.484320, 2.84507e-05/)
        case ("V")
            coeffs = (/-1.25429, 4.32118e-05/)
        case ("UVW1")
            coeffs = (/-1.16448, 4.14899e-05/)

        case ("UVW2")
            coeffs = (/-2.72723, 7.14543e-05/)

        case ("UVM2")
            coeffs = (/-2.91093, 7.49769e-05/)
        case default
            coeffs = (/1.0, 0.0/)

     end select 
       
     !TDPCorrectionFactor = coeffs(1) + coeffs(2) * mjd
     TDPCorrectionFactorOld = coeffs(1) + coeffs(2) * mjd
     !write(*,*) "L3246 filter=", trim(filter)
     !write(*,*) "L3247 coeffs=", coeffs
    
end function TDPCorrectionFactorOld

  
  !-------------------------------------------------------------------------------------
  subroutine variability_tests(Bin_size, Net_Count, Net_total, Nlight, Nbin, &
       Which_Tests, Light_Curve)
  !-------------------------------------------------------------------------------------
  ! Performs many algorithms to determine how varies the total source. 
  ! Results are provided on the same output device than plots.

    real(kind=single),dimension(:,:),pointer :: Net_Count, Net_total
    integer,intent(in) :: Nlight , Nbin
    real,intent(in) :: Bin_size
    real :: Bin_size2

    type(data_tests),dimension(Nlight) :: Light_Curve
    type(data_tests),dimension(Nlight) :: Light_Curve_total

    integer :: Light

!    character(len=20) :: String
!    character(len=20) :: String, String_t, String_b
    character(len=12) :: Which_Tests
    character(len=12) :: Which_Tests_aux

!    integer :: Length , Length_t , Length_b , i
!    integer :: Length, i
    integer :: i
    real(kind=single) :: sigma1, maxDev, aux1

    !real :: Character_Size , Height

    Which_Tests_aux=Which_Tests

    
    !if (Nlight>1) then
    !   Height = 1.0 / (Nlight + 0.5 )
    !else
    !   Height = 1.0 / 2.5
    !end if
    !Character_Size = Height

    !write(*,*) "L3548 Height=", Height

    !call pgpage
    !call pgsvp(0.0 , 1.0 , 1.0-0.5*Height , 1.0)
    !call pgsch(7.5*Character_Size) ; call pgsci(3)
    !call pgmtxt('T' , -1. , 0.5 , 0.5 , 'VARIABILITY TESTS')
    !call pgsch(4.*Character_Size) ; call pgsci(5)
    !call string_of_real(real(Nbin) , String , Length)
    !call string_of_real(Bin_size , String_b , Length_b)
    !call pgmtxt('T' , -4. , 0.5 , 0.5 , 'NUMBER OF BINS : '&
    !     //String(1:Length)//' BIN SIZE : '//String_b(1:Length_b) &
    !     //' sec. ')
    !call pg sci(2)
    Bin_size2=  Bin_size  
    do Light=1,Nlight
      

       Light_Curve(Light)%Mean_Count = sum(Net_Count(Light,:)) / real(Nbin)
      

       sigma1=0.0
       maxDev=0.0
       do i=1,Nbin
          sigma1=sigma1+(Net_Count(Light,i)-Light_Curve(Light)%Mean_Count)* &
               (Net_Count(Light,i)-Light_Curve(Light)%Mean_Count)
          aux1=abs(Net_Count(Light,i)-Light_Curve(Light)%Mean_Count)
          if (aux1 .gt. maxDev) maxDev=aux1
       enddo
       sigma1=sigma1/real(Nbin-1)
       sigma1=sqrt(sigma1)
       !write(*,*) "L5198 NET sigma1=", sigma1, "maxDev=", maxDev, " .....", maxDev/sigma1
       
       if (sigma1 .gt. 1.0) maxDev=maxDev/sigma1 ! expressed in the number of sigmas
       Light_Curve(Light)%Max_Deviation=maxDev
       !-----------

       Light_Curve_total(Light)%Mean_Count = sum(Net_total(Light,:)) / real(Nbin)
       sigma1=0.0
       maxDev=0.0
       do i=1,Nbin
          sigma1=sigma1+(Net_total(Light,i)-Light_Curve_total(Light)%Mean_Count)* &
               (Net_total(Light,i)-Light_Curve_total(Light)%Mean_Count)
          aux1=abs(Net_total(Light,i)-Light_Curve_total(Light)%Mean_Count)
          if (aux1 .gt. maxDev) maxDev=aux1
       enddo
       sigma1=sigma1/real(Nbin-1)
       sigma1=sqrt(sigma1)
       !write(*,*) "L5215 TOT sigma1=", sigma1, "maxDev=", maxDev, " .....", maxDev/sigma1
       
       if (sigma1 .gt. 1.0) maxDev=maxDev/sigma1 ! expressed in the number of sigmas
       Light_Curve_total(Light)%Max_Deviation=maxDev

       Light_Curve(Light)%Variance = 0.
       Light_Curve_total(Light)%Variance = 0.
       !write(*,*) "L5222  Light_Curve(Light) Mean_Count=",  Light_Curve(Light)%Mean_Count
       !write(*,*) "L5223  Light_Curve_total(Light) Mean_Count=", Light_Curve_total(Light)%Mean_Count
 
       do i=1,Nbin
          Light_Curve(Light)%Variance = Light_Curve(Light)%Variance + &
               (Net_Count(Light,i) - Light_Curve(Light)%Mean_Count)**2.0
          Light_Curve_total(Light)%Variance=Light_Curve_total(Light)%Variance &
              + (Net_total(Light,i) - Light_Curve_total(Light)%Mean_Count)**2.0
       end do
       

       Light_Curve(Light)%Variance = Light_Curve(Light)%Variance / real(Nbin-1)
       Light_Curve_total(Light)%Variance = Light_Curve_total(Light)%Variance /&
                                                                real(Nbin-1)
       !write(*,*) "L5236 variance=",Light_Curve(Light)%Variance
       !write(*,*) "L5237 total variance=",  Light_Curve_total(Light)%Variance

       call chi_square_test( Net_total(Light,:), Nbin, &
            Light_Curve_total(Light) )
       
       
       call kolmogorov_smirnov_test( Net_total(Light,:), Nbin, &
            Light_Curve_total(Light) )

       

       call chi_square_test( Net_Count(Light,:), Nbin, &
            Light_Curve(Light) )
       
       call kolmogorov_smirnov_test( Net_Count(Light,:), Nbin, &
            Light_Curve(Light) )
       
       !write(*,*) "L5254 RATES:"
       !write(*,*) "L5255 Total=",Light_Curve_total(Light)%Mean_Count/Bin_size 
       !write(*,*) "L5256   Net=",Light_Curve(Light)%Mean_Count/Bin_size

       !write(*,*) "L3535 VARIANCE:"
       !write(*,*) "L3536 Total=",Light_Curve_total(Light)%Variance/(Bin_size**2.0)
       !write(*,*) "L3537   Net=",Light_Curve(Light)%Variance/(Bin_size**2.0)

       !write(*,*) "L3550 Chi: Tot=",Light_Curve_total(Light)%Chi_Square%Stat
       !write(*,*) "L3551 dof:", real(Nbin-1)
       !write(*,*) "L3552 Prob=",Light_Curve_total(Light)%Chi_Square%Prob


       !write(*,*) "L3555 Kolm.Sm.: Tot=", Light_Curve_total(Light)%K_S%Stat
       !write(*,*) "L3556 prob=",Light_Curve_total(Light)%K_S%Prob

    end do

  end subroutine variability_tests

  !-------------------------------------------------------------------------------------
  subroutine chi_square_test(Count , Nbin , Lcurve)
  !-------------------------------------------------------------------------------------
  ! Returns the chi square statistic and probability of the hypothesis non variable
  ! source. So, the Degree of freedom (Df) is equal to Nbin-1.

    real(kind=single),dimension(:) :: Count
    integer,intent(in) :: Nbin
    type(data_tests),intent(inout) :: Lcurve

    real(kind=single) :: Stat 
    integer :: Df , i

    Df = Nbin - 1
    Stat = 0.
    do i=1,Nbin
       Stat = Stat + ( Count(i) - Lcurve%Mean_Count )**2
    end do
    Stat = Stat / Lcurve%Mean_Count
    Lcurve%Chi_Square%Stat = Stat
    Lcurve%Chi_Square%Prob = gamma_function_Q(0.5*Df , 0.5*Stat)

  end subroutine chi_square_test

  !-------------------------------------------------------------------------------------
  subroutine kolmogorov_smirnov_test(Count , Nbin , Lcurve)
  !-------------------------------------------------------------------------------------
  ! Compares the cumulative rate distribution with the cumulative expected poisson
  ! distribution. Returns the corresponding K_S statistic D, and the signifiance level
  ! Prob. Great values of Prob significate there is no variability in the input
  ! distribution.

    real(kind=single),dimension(:) :: Count
    integer,intent(in) :: Nbin
    type(data_tests),intent(inout) :: Lcurve

    !real(kind=single) :: D, en, Nbin_Real, ir, Cumulative_Poisson_Value, fn, f0, KSprob
    real(kind=single) :: D, en, Nbin_Real, ir, fn, f0, KSprob
    !integer :: i, j
    integer :: j

    !write(*,*) "L5315 Nbin=", Nbin
    !write(*,*) "L5316 size(Count)=", size(Count)
    !write(*,*) "L5316 lbound=", lbound(Count,1), &
    !                " ubound=", ubound(Count,1)
    D = 0.0;  f0=0.0
    call sort(Count)
    

    Nbin_Real = real(Nbin)
    en=Nbin_Real
    
!!    en = (sqrt((Nbin_Real*Nbin_Real)/(Nbin_Real+Nbin_Real)))
!!    en = sqrt(Nbin_Real)
    ir = 0.
    
    do j=1,Nbin_Real
      
        fn=j/en  
        ir=amax1(abs(f0-Count(j)),abs(fn-Count(j)))
        if (ir > D) D=ir
        f0=fn
    end do
   Lcurve%K_S%Stat = D
!   print*,sqrt(en)*D
   Lcurve%K_S%Prob = kolmogorov_smirnov_prob(sqrt(en)*D)
   KSprob = Lcurve%K_S%Prob
   
!    D = 0.
!    call sort(Count)
!    Nbin_Real = real(Nbin)
!    en = sqrt(Nbin_Real)
!    ir = 0.
!    do i=1,Nbin
!       Cumulative_Poisson_Value = 0.5*( 1. - Gauss_Integral_Function( &
!            (Lcurve%Mean_Count-Count(i)) / sqrt(2.*Lcurve%Mean_Count) ) )
!       D = max( D , abs(ir/Nbin_Real - Cumulative_Poisson_Value ) )
!       ir = real(i)
!       D = max( D , abs(ir/Nbin_Real - Cumulative_Poisson_Value) )
!    end do
!    Lcurve%K_S%Stat = D
!    Lcurve%K_S%Prob = kolmogorov_smirnov_prob((en+0.12+0.11/en) * D)

  end subroutine kolmogorov_smirnov_test


   !-------------------------------------------------------------------------------------
  function kolmogorov_smirnov_prob(lambda)
  !-------------------------------------------------------------------------------------
  ! Kolmogorov-Smirnov probability function.

    real(kind=single) :: kolmogorov_smirnov_prob
    real(kind=single),intent(in) :: lambda
    real(kind=single),parameter :: eps1=0.001 , eps2=1.e-8
    integer :: j
    real(kind=single) :: lambda2 , term , last_term , fac

    lambda2 = -2.*lambda**2.0
    fac = 2.
    j = 1
    last_term = 0.
    term = fac*exp(lambda2*j**2.0)
    kolmogorov_smirnov_prob = term
    do while ( (abs(term) > eps1*last_term) .and. &
         (abs(term) > eps2*kolmogorov_smirnov_prob) .and. (j < 100) )
       last_term = abs(term)
       j = j + 1
       fac = -fac
       term = fac*exp(lambda2*j**2.0)
       kolmogorov_smirnov_prob = kolmogorov_smirnov_prob + term
    end do
    if (j >= 100) kolmogorov_smirnov_prob = 1.
  end function kolmogorov_smirnov_prob


  !-------------------------------------------------------------------------------------
  function gamma_function_P(a,x)
  !-------------------------------------------------------------------------------------
  ! Returns the incomplete gamma function P(a,x) for a > 0. and x >= 0.

    real(kind=single),intent(in) :: a , x
    real(kind=single) :: gamma_function_P , gammcf , gamser , gln

    if (x < a+1.) then
       call series_gamma_func(gamser,a,x,gln)       ! Uses the serie representation
       gamma_function_P = gamser
    else
       call Fraction_gamma_function(gammcf,a,x,gln) ! Uses the complement of the
       gamma_function_P = 1. - gammcf               ! continued fraction representation
    end if
  end function gamma_function_P

  !-------------------------------------------------------------------------------------
  function gamma_function_Q(a,x)
  !-------------------------------------------------------------------------------------
  ! Returns the incomplete gamma function Q(a,x)=1-P(a,x) for a > 0. and x >= 0.

    real(kind=single),intent(in) :: a , x
    real(kind=single) :: gamma_function_Q , gammcf , gamser , gln

    if (x < a+1.) then
       call series_gamma_func(gamser,a,x,gln)         ! Uses the complement of the serie
       gamma_function_Q = 1. - gamser                 ! representation
    else
       call Fraction_gamma_function(gammcf,a,x,gln)   ! Uses the continued fraction
       gamma_function_Q = gammcf                      ! representation
    end if
  end function gamma_function_Q

  !-------------------------------------------------------------------------------------
  subroutine series_gamma_func(gamser ,a , x , gln)
  !-------------------------------------------------------------------------------------
  ! Returns the incomplete gamma function P(a,x) evaluated by its series representation
  ! as gamser. Also returns Ln(gamma(a) as gln. Warning : this routine works only if 
  ! a > 0. and x >= 0.

    real(kind=single),intent(in) :: a , x
    real(kind=single),intent(out) :: gamser , gln
    real(kind=single),parameter :: eps = 3.e-7
    integer,parameter :: imax = 100
    integer :: i
    real(kind=single) :: del , sum

    gln = gamma_log(a)
    if ( x == 0. ) then
       gamser = 0.
    else
       del = 1./a
       sum = del
       i = 1
       do while ( (i <= imax) .and. (abs(del) >= abs(sum)*eps) )
          del = del * x/(a+i)
          sum = sum + del
          i = i + 1
       end do
       gamser = sum * exp(-x + a*log(x) - gln)
    end if
  end subroutine series_gamma_func

  !-------------------------------------------------------------------------------------
  subroutine fraction_gamma_function(gammcf ,a , x , gln)
  !-------------------------------------------------------------------------------------
  ! Returns the incomplete gamma function Q(a,x) evaluated by its continued fraction
  ! representation as gammcf. Also returns Ln(gamma(a)) as gln. Warning : this routine
  ! works only if x >0. and a > 0.

    real(kind=single),intent(in) :: a , x
    real(kind=single),intent(out) :: gammcf , gln
    real(kind=single),parameter :: eps = 3.e-7 , min_float = 1.e-30
    integer,parameter :: imax = 100
    integer :: i
    real(kind=single) :: an , b , c , d , del , h

    gln = gamma_log(a)
    b = x + 1. - a
    c = 1./min_float
    d = 1./b
    h = d
    i = 1
    del = 2.                     ! Non significative value use for first loop iteration.
    do while ( (i <= imax) .and. (abs(del-1.) >= eps) )
       an = -i*(i-a)
       b = b + 2.
       d = an*d + b
       if ( abs(d) < min_float ) d = min_float
       c = b + an/c
       if ( abs(c) < min_float ) c = min_float
       d = 1./d
       del = d * c
       h = h * del
    end do
    gammcf = exp(-x + a*log(x) - gln) * h
  end subroutine fraction_gamma_function

  !-------------------------------------------------------------------------------------
  function gamma_log(xx)
  !-------------------------------------------------------------------------------------
  ! Returns Lanczos approximation of the value Ln(gamma(xx)) for xx > 0.

    real(kind=single),intent(in) :: xx
    real(kind=single) :: gamma_log
    real(kind=double) :: ser
    real(kind=double),dimension(6) :: cof = (/ 76.18009172947146d0 , &
         -86.50532032941677d0 , 24.01409824083091d0 , -1.231739572450155d0 , &
         .1208650973866179d-2 , -0.5395239384953d-5 /)
    real(kind=double) :: stp = 2.5066282746310005d0
    real(kind=double) :: x , y , tmp
    integer :: j

    ser = 1.000000000190015d0
    x = xx
    y = x
    tmp = x + 5.5d0
    tmp = (x+0.5d0) * log(tmp) - tmp
    do j=1,6
       y = y + 1.0d0
       ser =  ser + cof(j)/y
    end do
    gamma_log = real(tmp + log(stp * ser/x))
  end function gamma_log

  !-------------------------------------------------------------------------------------
  recursive subroutine sort(arr)
  !-------------------------------------------------------------------------------------
  ! Sorts an array arr(1:n) into ascending numerical order. arr is replaced on output
  ! by its sorted rearrangement

    real(kind=single),dimension(:),intent(inout) :: arr
    integer :: begin , end
    real(kind=single) :: elt , tmp
    integer :: i , j , med
    logical :: isNaN
    integer :: nNans

    isNaN= .false.

    begin = 1 ; end = size(arr)
     
    ! check whether the input values are NaN    
    if (end > begin) then
       
       ! check whether the input values are NaN
       nNans=0
       do i= begin,end
          elt=arr(i)
          ! make two checks of elt being NaN
          if (( (elt > 0.0) .EQV. (elt <= 0.0)) .or. (elt/=elt) ) then 
             isNaN=.true.
             nNans=nNans+1
          endif
       enddo
       

       ! make sort only if there are no NaN values
       if (.not. isNan) then
          elt = arr(begin)

          i = begin + 1 ; j = end
          do while (i < j)
             
             do while ( (arr(i) < elt) .and. (i < j) )
                i = i + 1 ; arr(i) = arr(i)
             end do
             do while ( (arr(j) >= elt) .and. (j > i) )
                j = j - 1 ; arr(j) = arr(j)
             end do
             if (i < j) then
                tmp = arr(i)
                arr(i) = arr(j)
                arr(j) = tmp
             end if
          end do
          if (arr(i) < elt) then
             med = i
          else
             med = i - 1
          end if
          tmp = arr(begin)
          arr(begin) = arr(med)
          arr(med) = tmp
          if (med-begin > 1) call sort(arr(begin : med-1))
          if (end-med > 1) call sort(arr(med+1 : end))
       else 
          ! there were NaN values detected
          write(errstr,*) "Detected NaN values:  ", nNans
          call message(AppMsg, 2)
       end if
    end if
  end subroutine sort

  !----------------------------------------------------------------------------------
  subroutine recover_counts(Time_Length, Net_Rate, Bkg_Rate, Net_Count, &
       Net_Total, Good_Bin, Nlight, Nbin)
  !----------------------------------------------------------------------------------
  ! Recovers total source counts (net*time)

    real(kind=single),intent(in) :: Time_Length
    real(kind=single),dimension(:,:),pointer :: Net_Rate , Net_Count
    real(kind=single),dimension(:,:),pointer :: Bkg_Rate , Net_Total
    real(kind=single),dimension(:,:),pointer :: Net_Count_Temp, Net_Total_Temp
    logical,dimension(:),pointer :: Good_Bin
    integer,intent(in) :: Nlight
    integer,intent(inout) :: Nbin
    integer :: i, n

    allocate(Net_Count_Temp(Nlight,Nbin))
    allocate(Net_Total_Temp(Nlight,Nbin))
    n=0
    do i=1,Nbin
       if ( Good_Bin(i) ) n = n + 1
    end do
    !write(*,*) "L3958 Good_Bin n=", n, " Nbin=", Nbin, " Nlight=", Nlight

    !do i=1,Nbin
    !   write(*,*) "L3961 i=", i, " Net_Rate=", Net_Rate(:,i), &
    !        " Bkg_Rate=", Bkg_Rate(:,i)
    !enddo
    !write(*,*) "-------"
    !write(*,*) "L3965 Time_Length=", Time_Length

    do i=1,Nbin
       if ( Good_Bin(i) ) then
          Net_Count_Temp(:,i) = Net_Rate(:,i) * Time_Length
          Net_Total_Temp(:,i) = (Net_Rate(:,i)+Bkg_Rate(:,i)) * Time_Length
       endif
    end do
    Nbin = n
    allocate(Net_Count(Nlight,Nbin))
    allocate(Net_Total(Nlight,Nbin))
    Net_Count = Net_Count_Temp(:,1:Nbin)
    Net_Total = Net_Total_Temp(:,1:Nbin)

    !do i=1,Nbin
    !   write(*,*) "L3973 i=", i, " Net_Count=", Net_Count(:,i), &
    !        " Net_Total=", Net_Total(:,i)
    !enddo

    deallocate(Net_Count_Temp)
    deallocate(Net_Total_Temp)

  end subroutine recover_counts


  !-------------------------------------------------------------------------------------
  subroutine check_sizes(Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, Nlight, Nbin)
  !-------------------------------------------------------------------------------------
  ! Checks if all these vectors have same dimensions : Nlight * Nbin.

    real(kind=single),dimension(:,:),pointer :: Net_Rate , Net_Err , Bkg_Rate , Bkg_Err
    integer,intent(out) :: Nlight, Nbin

    Nlight = size(Net_Rate(:,1))
    if ( size(Net_Err(:,1)) /= Nlight ) call error('OMLCBUILD', &
         'Different number of columns between RATE, ERROR, BACKV and BACKE')
    if ( size(Bkg_Rate(:,1)) /= Nlight ) call error('OMLCBUILD', &
         'Different number of columns between RATE, ERROR, BACKV and BACKE')
    if ( size(Bkg_Err(:,1)) /= Nlight ) call error('OMLCBUILD', &
         'Different number of columns between RATE, ERROR, BACKV and BACKE')
    if ( Nlight > 9 ) call error('OMLCBUILD', &
         'Number of light-curves contained in a file must not exceed 9')
    if (.not. ( size(Net_Rate(1,:)) == size(Net_Err(1,:)) &
         .and. size(Bkg_Rate(1,:)) == size(Bkg_Err(1,:)) &
         .and. size(Net_Rate(1,:)) == size(Bkg_Rate(1,:)) )) call error('OMLCBUILD' ,&
         'Different number of rows in columns RATE, ERROR, BACKV and BACKE')
    Nbin = size( Net_Rate(1,:) )
    if ( Nbin <= 0 ) call error('OMLCBUILD','No data point received')

  end subroutine check_sizes 

  !-------------------------------------------------------------------------------------
  subroutine delete_gap_data(Time_Vect, Net_Rate, Net_Err, Bkg_Rate, Bkg_Err, &
       Nlight, Nbin)
  !-------------------------------------------------------------------------------------
  ! Deletes gap (when net and background rates are both null)

    real(kind=single),dimension(:),pointer :: Time_Vect
    real(kind=single),dimension(:,:),pointer :: Net_Rate, Net_Err, Bkg_Rate 
    real(kind=single),dimension(:,:),pointer :: Bkg_Err
    integer,intent(inout) :: Nbin
    integer,intent(in) :: Nlight

    real(kind=single),dimension(Nbin) :: Time_Vect_Temp
    real(kind=single),dimension(Nlight,Nbin) :: Net_Rate_Temp , Net_Err_Temp
    real(kind=single),dimension(Nlight,Nbin) :: Bkg_Rate_Temp , Bkg_Err_Temp
    real(kind=single) :: Eps
    integer :: Nbin_Temp , Light , i
    logical :: Good_Value

    Eps = epsilon(0.)
    Nbin_Temp = 0
    do i=1,Nbin
       Good_Value = .true.
       do Light=1,Nlight
!          Good_Value = Good_Value .and. ( ( Net_Rate(Light,i) > Eps ) .or. &
!
!               ( Bkg_Rate(Light,i) > Eps ) )
         Good_Value = Good_Value 
       end do
       if (Good_Value) then
          Nbin_Temp = Nbin_Temp + 1
          Time_Vect_Temp(Nbin_Temp) = Time_Vect(i)
          Net_Rate_Temp(:,Nbin_Temp) = Net_Rate(:,i)
          Net_Err_Temp(:,Nbin_Temp) = Net_Err(:,i)
          Bkg_Rate_Temp(:,Nbin_Temp) = Bkg_Rate(:,i)
          Bkg_Err_Temp(:,Nbin_Temp) = Bkg_Err(:,i)
       end if
    end do
    deallocate(Time_Vect)
    deallocate(Net_Rate , Net_Err , Bkg_Rate , Bkg_Err)
    Nbin = Nbin_Temp
    allocate( Time_Vect(Nbin) )
    Time_Vect = Time_Vect_Temp
    allocate( Net_Rate(Nlight,Nbin) , Net_Err(Nlight,Nbin) )
    allocate( Bkg_Rate(Nlight,Nbin) , Bkg_Err(Nlight,Nbin) )
    Net_Rate = Net_Rate_Temp ; Net_Err = Net_Err_Temp
    Bkg_Rate = Bkg_Rate_Temp ; Bkg_Err = Bkg_Err_Temp
    if ( minval(Net_Rate) < 0.0 .or. minval(Bkg_Rate)  < 0.0 ) &
         call warning('OMLCBUILD','Count rate is not positive')
    if ( minval(Net_Err) < 0.0 .or. minval(Bkg_Err) < 0.0 ) &
         call warning('OMLCBUILD','Count rate errors must be positive')

    if (size(Time_Vect) < 1) then
         call error('OMLCBUILD','No good data points')
    end if

  end subroutine delete_gap_data

  !-------------------------------------------------------------------------------------
  subroutine rebin(Binning_Factor, Bin_Width, Time_Vect, Net_Rate, Net_Err, &
  Bkg_Rate, Bkg_Err, Good_Bin, Nlight, Nbin, Col_Rate, Col_Backe, gotnullvals)
  !-------------------------------------------------------------------------------------
  ! Rebins Net_Rate, Net_Err, Bkg_Rate, Bkg_Err by at least a factor five 
  ! to ensure good signal-to-noise and to limit the number of plotted points 
  ! to 500.  Also modifies Time_Vect.

    integer,intent(inout) :: Binning_Factor
    real(kind=single),dimension(:,:),pointer :: Net_Rate, Net_Err, Bkg_Rate, Bkg_Err
    integer,intent(in) :: Nlight
    integer,intent(inout) :: Nbin
    real(kind=single),intent(in) :: Bin_Width
    real(kind=single),dimension(:),pointer :: Time_Vect
    logical,dimension(:),pointer :: Good_Bin
    type( ColumnT ) :: Col_Rate, Col_Backe

    real(kind=single),dimension(:,:),pointer :: Net_Rate_Reb, Net_Err_Reb
    real(kind=single),dimension(:,:),pointer :: Bkg_Rate_Reb, Bkg_Err_Reb
    real(kind=single),dimension(:),pointer :: Time_Reb
    !real(kind=single),dimension(Nbin) :: Time_Vect_Temp
    !real(kind=single),dimension(Nlight,Nbin) :: Net_Rate_Temp, Net_Err_Temp
    
    !real(kind=single),dimension(Nlight,Nbin) :: Bkg_Rate_Temp, Bkg_Err_Temp 
    
    logical(boolean),dimension(:),pointer  :: gotnullvals
    !integer :: Nbin_Reb ,  i , j, Nbin_Temp, Light, k
    integer :: Nbin_Reb ,  i , j, k
    !logical :: No_Gap, Good_Value
    logical :: No_Gap
    !real :: Eps, averr
    real :: Eps

    ! Sets Binning_Factor to correct value
    if (Binning_Factor < 1) then
       Binning_Factor = 1
       call warning('OMLCBUILD','Binning_Factor must be positive, set to default: 1')
    end if
    if ( int(Nbin/Binning_Factor)>500 ) then
       Binning_Factor = ((int(Nbin/Binning_Factor)-1)/512 + 1)*Binning_Factor
       write(errstr,*) 'Data point to plot must not exceed 500. '//&
                  'Binning Factor is increased from input value to : ',Binning_Factor
       call warning('OMLCBUILD')
    end if

    ! Rebinning

    allocate(Time_Reb(Nbin))
    allocate(Good_Bin(Nbin))
    allocate(gotnullvals(Nbin))
    allocate(Net_Rate_Reb(Nlight,Nbin))
    allocate(Net_Err_Reb(Nlight,Nbin))
    allocate(Bkg_Rate_Reb(Nlight,Nbin))      
    allocate(Bkg_Err_Reb(Nlight,Nbin))      

    Eps = Bin_Width/10.
    Nbin_Reb = 0
    i = 0
    do while (i < Nbin)
       i = i + 1  
       Nbin_Reb                 =  Nbin_Reb + 1
       Time_Reb(Nbin_Reb)       =  Time_Vect(i)
       Net_Rate_Reb(:,Nbin_Reb) =  Net_Rate(:,i)
       if ( isNull ( Col_Rate , i-1 ) ) then
          Net_Err_Reb(:,Nbin_Reb)  =  0.0
          gotnullvals(Nbin_Reb)    = .true.
       else
         Net_Err_Reb(:,Nbin_Reb)  =  Net_Err(:,i)**2.0
         gotnullvals(Nbin_Reb)    = .false.
       end if
       Bkg_Rate_Reb(:,Nbin_Reb) =  Bkg_Rate(:,i)
       if ( isNull ( Col_Backe , i-1 ) ) then
         Bkg_Err_Reb(:,Nbin_Reb)  =  0.0
       else 
         Bkg_Err_Reb(:,Nbin_Reb)  =  Bkg_Err(:,i)**2.0
       end if
       j = 1; k = 1
!       No_Gap=(abs(Time_Vect(min(i,Nbin))-Time_Vect(i-1)-Bin_Width) <= Eps)
       grouploop: do while ( ( (i < Nbin) .and. (k < Binning_Factor) ))
          i=i+1; k=k+1
          Time_Reb(Nbin_Reb) = Time_Reb(Nbin_Reb) + Time_Vect(i)
        if (isNull( Col_Rate , i-1 ) .and. .not. isNull( Col_Rate , i-2 )) then
          Net_Rate_Reb(:,Nbin_Reb) = Net_Rate_Reb(:,Nbin_Reb) + 0.0 
          Net_Err_Reb(:,Nbin_Reb) =  Net_Err_Reb(:,Nbin_Reb) + 0.0 
          Bkg_Rate_Reb(:,Nbin_Reb) =  Bkg_Rate_Reb(:,Nbin_Reb) + 0.0 
          Bkg_Err_Reb(:,Nbin_Reb) = Bkg_Err_Reb(:,Nbin_Reb) + 0.0 
          gotnullvals(Nbin_Reb)   = .true.
        else if (isNull( Col_Rate , i-2 ) .and. .not. isNull( Col_Rate , i-1 )) then
          Net_Rate_Reb(:,Nbin_Reb) = 0.0 + Net_Rate(:,i)
          Net_Err_Reb(:,Nbin_Reb) = 0.0 + Net_Err(:,i)**2.0
          Bkg_Rate_Reb(:,Nbin_Reb) = 0.0 + Bkg_Rate(:,i)
          Bkg_Err_Reb(:,Nbin_Reb) = 0.0 + Bkg_Err(:,i)**2.0
          gotnullvals(Nbin_Reb)   = .true.
        else if (isNull( Col_Rate , i-1 ) .and. isNull( Col_Rate , i-2 )) then
          Net_Rate_Reb(:,Nbin_Reb) =   1.0e10
          Net_Err_Reb(:,Nbin_Reb) =  0.0
          Bkg_Rate_Reb(:,Nbin_Reb) =  1.0e10
          Bkg_Err_Reb(:,Nbin_Reb) =   0.0
          gotnullvals(Nbin_Reb)   = .true.
        else 
          Net_Rate_Reb(:,Nbin_Reb) = Net_Rate_Reb(:,Nbin_Reb) + Net_Rate(:,i)
          Net_Err_Reb(:,Nbin_Reb) = Net_Err_Reb(:,Nbin_Reb) + Net_Err(:,i)**2.0
          Bkg_Rate_Reb(:,Nbin_Reb) = Bkg_Rate_Reb(:,Nbin_Reb) + Bkg_Rate(:,i)
          Bkg_Err_Reb(:,Nbin_Reb) = Bkg_Err_Reb(:,Nbin_Reb) + Bkg_Err(:,i)**2.0
          gotnullvals(Nbin_Reb)   = .false.
          j = j + 1 
        endif
          No_Gap=(abs(Time_Vect(min(i,Nbin))-Time_Vect(i-1)-Bin_Width) <= Eps)
       end do grouploop
       Time_Reb(Nbin_Reb) = Time_Reb(Nbin_Reb)/real(k)
       if (.not. gotnullvals(Nbin_Reb)) then
         Net_Rate_Reb(:,Nbin_Reb)  =  Net_Rate_Reb(:,Nbin_Reb) / real(j)
         Net_Err_Reb(:,Nbin_Reb)   =  sqrt(Net_Err_Reb(:,Nbin_Reb)) / real(j)
         Bkg_Rate_Reb(:,Nbin_Reb)  =  Bkg_Rate_Reb(:,Nbin_Reb) / real(j)
         Bkg_Err_Reb(:,Nbin_Reb)   =  sqrt(Bkg_Err_Reb(:,Nbin_Reb)) / real(j)
         Good_Bin(Nbin_Reb)        =  ( j == Binning_Factor )
       end if 
    end do 
    deallocate(Time_Vect)
    deallocate(Net_Rate , Net_Err , Bkg_Rate , Bkg_Err)
    Nbin = Nbin_Reb
    allocate(Time_Vect(Nbin))
    allocate(Net_Rate(Nlight,Nbin), Net_Err(Nlight,Nbin))
    allocate(Bkg_Rate(Nlight,Nbin), Bkg_Err(Nlight,Nbin))
    Time_Vect  =  Time_Reb(1:Nbin)
    Net_Rate   =  Net_Rate_Reb(:,1:Nbin)
    Net_Err    =  Net_Err_Reb(:,1:Nbin)
    Bkg_Rate   =  Bkg_Rate_Reb(:,1:Nbin)
    Bkg_Err    =  Bkg_Err_Reb(:,1:Nbin)

    deallocate(Net_Rate_Reb, Net_Err_Reb, Bkg_Rate_Reb, Bkg_Err_Reb, Time_Reb)

  end subroutine rebin


  !***************************************************************
  !
  !**************************************************************
  subroutine createBackgroundImage2(image, qualityImage, width, imbinX, &
       averageBkg, averageSigma, levelImage,  &
       backgroundImage, backgroundsigmaImage, maxheight)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
    
  
    integer(int16), intent(in) :: width
    integer(kind=int32), intent(in) :: imbinX
    real(single), intent(out) :: averageBkg
    real(single), intent(out) :: averageSigma
    integer(int16), pointer :: levelImage(:, :)
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundImage0(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    real(single), intent(in), optional :: maxHeight


    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
    !integer(int16), pointer :: levelImage(:, :) => null()
    logical, pointer :: pixelOkay(:, :) => null()
  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
   
    logical, pointer :: computed(:, :) => null()

    !--------------------------------------------------------
    !! variables for saving testimage77.fit
    character(len=255) :: outSetCheck
    integer(int32) :: outDimCheck(2) 
    type(DataSetT) :: outDataSetCheckH 
    type(ArrayT) :: outArrayCheckH 
    real(single), dimension(:, :), pointer :: outOSWcheckP  
    integer :: count1 
    !--------------------------------------------------------
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)

    averageBkg=-1.0
    averageSigma=-1.0

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = 50
    dw1 = dw - 1
    
    !write(*,*) "L5381 width=", width, " dx=", dx, " dy=", dy

    if (.not. associated(levelImage)) then
       levelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "levelimage")       
    end if

    !levelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "levelimage")
    
    !write(*,*) "L5384 allocated levelImage" ! the level image is not used?
    mode = ComputeBackgroundMode4(image, levelImage, qualityImage, &
               nx + 1_int16, ny + 1_int16, 2.0, sigma)


    !write(*,*) "L5424 after computeBackgroundMode4 mode=", mode, " sigma=", sigma
    averageBkg=mode
    averageSigma=sigma

    if (present(maxHeight)) then
        globalMedian = mode + maxheight
    else
        globalMedian = mode + 10.0 * max(1.0, sigma)
    end if
    !write(*,*) "L5422 globalMedian=", globalMedian

    !deallocate(LevelImage)
    !nullify(levelImage)
     
    if (.not. associated(backgroundimage)) then
    
        backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if

    
    
    backgroundImage0 => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    


    if (.not. associated(backgroundsigmaimage)) then
 
        backgroundSigmaImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if



    backgroundImage0 = image


    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    !write(*,*) "L5452 globalMedian=", globalMedian

    where (image .lt. globalMedian .and. .not. qualityImage)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 

    count1=0  
    do i=0, nx-1
       do j=0, ny-1
          if (pixelOkay(i,j)) then
             count1=count1+1
          endif
       enddo
    enddo

    globalMedian = mode - 5.0 * max(1.0, sigma)

    where(image .lt. globalMedian) pixelOkay = .false.


    !write(*,*) "L5474 globalMedian=", globalMedian, " count1=", count1


    allocate(computed( 0 : nx, 0 : ny) )
    computed = .false.


    !write(*,*) "L5481 nx=", nx, " dw=", dw
    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************
    backgroundImage0=-1.0
    backgroundsigmaImage=1000.0
    do i = 0, nx, dw
       iu = min(i + dw1, nx)
       do j = 0, ny, dw
          ju = min(j + dw1, ny)            
          call CreateBackgroundStrip(i, iu, j, ju, image, &
               backgroundImage0, backgroundsigmaImage, pixelOkay, computed, &
               dx, dy, dr, mode)
       end do
    end do



    deallocate(pixelOkay)
    nullify(pixelOkay)


    !----------------------
    !write(*,*) "L5504 creating testb1image77.fit"
    !outSetCheck="testb1image77.fit"
    !outDimCheck(1)=nx+1 !imWindowX1-imWindowX0+1
    !outDimCheck(2)=ny+1 !imWindowY1-imWindowY0+1
    !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
    !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
    !     real32, outDimCheck, "Counts", "IMAGE")!             
    !outOSWcheckP => real32Array2Data(outArrayCheckH)
    !outOSWcheckP = backgroundImage0  
    !! Release handles on data.
    !call release(outArrayCheckH)        
    !call release(outDataSetCheckH) 
    !
    !write(*,*) "L5517 testb1image77 saved:", trim(outSetCheck)!
    !--------------------------------

    backgroundImage=-1.0
    !****************************************************************************
    ! Fill in the missing background pixels by interpolation
    !***************************************************************************
    call FillInBackgroundPixels( backgroundimage0, backgroundImage, backgroundsigmaImage, computed)

    
    !----------------------
    !write(*,*) "L5518 creating testb2image77.fit"
    !outSetCheck="testb2image77.fit"
    !outDimCheck(1)=nx+1 !imWindowX1-imWindowX0+1
    !outDimCheck(2)=ny+1 !imWindowY1-imWindowY0+1
    !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
    !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
    !     real32, outDimCheck, "Counts", "IMAGE")!             
    !outOSWcheckP => real32Array2Data(outArrayCheckH)
    !outOSWcheckP = backgroundImage  
    !! Release handles on data.
    !call release(outArrayCheckH)        
    !call release(outDataSetCheckH) 
    !
    !write(*,*) "L5531 testb2image77 saved:", trim(outSetCheck)!
    !--------------------------------


    deallocate(computed)
    nullify(computed)



    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  
    !write(*,*) "L5484 nx=", nx, " ny=", ny, " imbinX=", imbinX
    call smoothimage( backgroundImage, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / imbinX, int32), int(20 / imbinX, int32))

    !write(*,*) "L5488 after smoothing"

    !----------------------
    !write(*,*) "L5533 creating testb3image77.fit"
    !outSetCheck="testb3image77.fit"
    !outDimCheck(1)=nx+1 !imWindowX1-imWindowX0+1
    !outDimCheck(2)=ny+1 !imWindowY1-imWindowY0+1
    !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
    !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
    !     real32, outDimCheck, "Counts", "IMAGE")!             
    !outOSWcheckP => real32Array2Data(outArrayCheckH)
    !outOSWcheckP = backgroundImage  
    !! Release handles on data.
    !call release(outArrayCheckH)        
    !call release(outDataSetCheckH) 
    ! 
    ! write(*,*) "L5546 testb3image77 saved:", trim(outSetCheck)!
    !--------------------------------


    deallocate(backgroundImage0)

    !    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
    !call fatal("stopped displaying modeimage.fit")  
    !    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

  end subroutine createBackgroundImage2


  !****************************************************************************
  ! Fill in the missing background pixels by interpolation
  !***************************************************************************
  subroutine FillInBackgroundPixels( image, imageFilled, sigmaImage, computed)
    real(single), pointer, dimension(:, :) :: image, imageFilled, sigmaImage 
    logical, pointer :: computed(:, :)
    type(PixelStruct), pointer :: pixel(:) => null()

    integer(int8) :: dl
    integer(int16) :: i, j, nx, ny

    dl = 3
    allocate( pixel( (2 * dl + 1) * (2 * dl + 1) ))

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    !write(*,*) "L5569 nx=", nx, " ny=", ny

    do j = 0, ny        
       do i = 0, nx 
          
          if ( .not. computed(i, j) ) then
             imageFilled(i, j) = pixelAverage(i, j,   &
                  image,  sigmaImage,  &
                  pixel, dl, computed, &
                  sigmaImage(i, j) )                 
             
          end if

       end do
    end do

    deallocate(pixel)
    nullify(pixel)
    

  end subroutine FillInBackgroundPixels


  !************************************************
  !
  !***********************************************
  real(single) Function PixelAverage(i, j, image1, image2, pixel, dx, computed, fb)
    integer(int16), intent(in) :: i, j
    real(single), pointer, dimension(:, :) :: image1, image2
    type(pixelStruct), pointer, dimension(:) :: pixel

    logical, pointer, dimension(:, :) :: computed
  

    integer(int8), intent(in) :: dx
    real(single), intent(out) :: fb


    ! Local variables     
    integer(int16) :: nx, ny, il, iu, jl, ju, i1, j1
    integer(int32) :: k   
    real :: xPos, yPos

    nx = ubound(image1, 1)
    ny = ubound(image1, 2)
    
    il = max(0_int16, i - dx)
    iu = min(nx, i + dx)
    jl = max(0_int16, j - dx)
    ju = min(ny, j + dx)


    k = 0
   
    do i1 = il, iu
       do j1 = jl, ju            
          if (i1 .eq. i .and. j1 .eq. j) cycle
          if (.not. computed(i1, j1)) cycle 
          if (image1(i1, j1) .lt. 0.0) cycle
          k = k + 1
          pixel(k) = pixelStruct( real(i1), real(j1), 0, image1(i1, j1), image2(i1, j1), 0)  
           
       end do
    end do
  

    !***************************************************************************
    ! Interpolate to get the background counts
    !**************************************************************************

    xPos = real(i)
    yPos = real(j)
  
    PixelAverage = neighbourInterpolate( xPos, yPos, k, real(pixel%x), real(pixel%y), pixel%counts ) 
            
    !***************************************************************************
    ! Interpolate to get the background counts
    !**************************************************************************
    fb = neighbourInterpolate( xPos, yPos, k, real(pixel%x), real(pixel%y), pixel%backgroundcounts ) 
          
  end Function PixelAverage


  !***************************************************************
  !
  !**************************************************************
  subroutine CreateBackgroundStrip(il, iu, jl, ju, image, &
       backgroundImage, backgroundsigmaImage, pixelOkay, &
       computed, dx, dy, dr, mode)
    
    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :)
    
    
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    logical, pointer, dimension(:, :) :: pixelOkay, computed
    integer(int16), intent(in) :: dx, dy, dr
    real(single), intent(inout) :: mode

    integer(int16) :: i, j, l, k
    integer(int16) :: il1, iu1, jl1, ju1, nx, ny
    integer(int32) :: count, pixelSize, n   

    logical, pointer :: usepixel(:) => null()
    real(single) :: sigma, mode1 
    real(single), pointer :: pixelValue(:) => null()
    integer, pointer :: pixelIndex(:, :) => null()
  
    integer(int8) :: skip = 2

    integer :: count2, count2a

    !write(*, *) "L5682 dx=", dx, " dy=", dy, " mode=", mode, " dr=", dr
    !***************************************************************
    !
    !
    !**************************************************************
    nx = ubound(image, 1)
    ny = ubound(image, 2)

 
    il1 = il - dr
    if (il1 < 0) then
        il1 = 0
    end if
    iu1 = iu + dr
    if (iu1 > nx) then
        iu1 = nx
         
    end if 
  

    jl1 = jl - dr
    if (jl1 < 0) then
        jl1 = 0
    end if
    ju1 = ju + dr
    if (ju1 > ny) then
        ju1 = ny
    

    end if

    !write(*, *) "L5714 il1=", il1, " iu1=", iu1, " jl1=", jl1, " ju1=", ju1

    call SetUpIndices(il1, iu1, jl1, ju1, image, pixelOkay, pixelValue, pixelIndex)

    !write(*, *) "Line 901"

    pixelSize = size(pixelValue)
    !write(*,*) "L5723 pixelValue=", pixelValue, " pixelSize=", pixelSize

    allocate(usePixel(1 : pixelSize))

    pixelSize = size(pixelValue)
    allocate(usePixel(1 : pixelSize))

    usePixel = .false.

    mode1 = mode !+ 5.0 * sqrt(mode)
   
    !write(*, *) "Line 914 createbackgroudnstrip mode1=", mode1, &
    !            " il=", il, " iu=", iu, " jl=", jl, " ju=", ju



    count2=0
    count2a=0

    ! Create the background image
    do j = jl, ju, skip        
       
       jl1 = j - dy
       if (jl1 < 0) then
          jl1 = 0
          ju1 = min(ny, dr)
       else 
          ju1 = j + dy
          if (ju1 > ny) then
             ju1 = ny
             jl1 = max(0_int16, ny - dr)
          end if
          
       end if
       
       do i = il, iu, skip
          
          if (image(i, j) .lt. 0.0) then ! Ignore negative image values
             
             backgroundimage(i, j) = -1.0
             backgroundsigmaimage(i, j) = 1000.0
             computed(i, j ) = .true.
             
             cycle
          end if
          computed(i, j ) = .true.
          
          il1 = i - dx
          if (il1 < 0) then
             il1 = 0
             iu1 = min(nx, dr)
          else 
             iu1 = i + dx
             if (iu1 > nx) then
                iu1 = nx
                il1 = max(0_int16, nx - dr)
             end if
          end if
          
          count = 0           
     
          do l = jl1, ju1
             
             do k = il1, iu1
                
                if (pixelOkay(k, l)) then
                   count = count + 1
                   
                   n = pixelIndex(k, l)
                   usePixel(n) = .true.
                   
                end if
             end do
          end do


          if (count .gt. 10) then  ! Was 10
             
             backgroundimage(i, j) = BackgroundMode11(count, usePixel, pixelValue, &
                  0.2, 2.0, sigma) 
 
             mode1 = backgroundimage(i, j)        
            
             count2=count2+1

             !if (i .lt. 10 .and. j .lt. 10) then
             !   write(*,*) "L5810 i=", i, " j=", j, " mode1=", mode1, " sigma=", sigma
             !endif
             
             backgroundsigmaimage(i, j) = max(1.0, sigma)
             !if (abs(i - 464) < 20 .and. abs(j - 553) < 20) then
             !       write(*, *) "i=", i, " j=", j, " mode1=", mode1, " count=", count
             !end if
          else

             backgroundimage(i, j) = mode1 ! image(i, j)
             backgroundsigmaimage(i, j) = max(1.0, sqrt(mode1) )
             usePixel = .false.
             
             count2a=count2a+1
             

             !write(*, *) "i=", i, " j=", j, " sqrtmode=", sqrtMode
             !call fatal("stopped at line 1013")
             
             
          end if
          !write(*, *) "Line 973 count=", count, " backgroundimage=", backgroundimage(i, j)
          !write(*, *) "il1=", il1, " iu1=", iu1, " jl1=", jl1, " ju1=", ju1
             
    
       end do




    end do


    !write(*,*) "L5851 count2=", count2, " count2a=", count2a


    ! call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "backgroundimage.fit", "background image")
    
    !call saveImage( backgroundsigmaimage, nx + 1_int16, ny + 1_int16, "sigmaimage.fit", "background image")
    
       
    !  call fatal("stopped at line 1090 displaying backgroundimage.fit and sigmaimage.fit")  
       


    mode = mode1
       
    deallocate(usePixel, pixelValue, pixelIndex)
    nullify(usePixel, pixelValue, pixelIndex)
       
  
 
  end subroutine CreateBackgroundStrip


  !*****************************************
  ! Function BackgroundMode
  !*****************************************
  function BackgroundMode11(nData, usePixel, pixelValue, &
       tolerance, nSigma, sigma) result(mode)
    integer(int32), intent(in) :: nData
    
    logical :: usePixel(:)
    real(single) :: pixelValue(: )
    
    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma
    
    real(single) :: data(1 : nData) 
    integer(int32) :: nlist
    
    real(single)    :: mode
    integer(int32) :: i, k
    real(single)   :: median
    real(double)   :: variance, testSigma, diff
    real(single)   :: highValue, lowValue
    integer(int32) :: m, nmin
    

    k = 0
    
    do i = 1, size(usePixel)
       if (usePixel(i)) then
          k = k + 1
          
          data(k) = pixelValue(i)
          
          usePixel(i) = .false.
          
       end if
    end do
    
    
    nList = nData
    
    nmin = ishft(ndata, -1)
    
    
    ! reject data at more than nSigma from data
    
    testSigma = 0.0
    
    sigmaClip: do i = 1, 20   !  was 20
       
       !********************************************************************
       ! Get the median
       !*******************************************************************
       k = ishft(nlist, -1)
       if (ishft(k, 1) .eq. nlist) then   ! Even number
          median = (data(k) + data(k + 1)) / 2.0
       else
          median = data(k + 1)
       end if
       
       !*******************************
       ! Now do sigma clipping
       ! Get variance and std deviation
       !********************************
       
       variance = 0.0_double
       do k = 1, nList
          diff = data(k) - median
          variance = variance + diff * diff 
          
       end do
       sigma = sqrt(variance / dble(nlist - 1))
       
       
       highValue = median + nSigma * sigma
       
       lowValue = median - nsigma * sigma
       
       
       m = 0  
       do k = 1, nlist
          if (data(k) .lt. lowValue) cycle ! Put in on Monday
          if(data(k) .gt. highValue) exit
          
          m = m + 1
          data(m) = data(k)
       end do
       
       
       if(m .lt. nmin .or. sigma .lt. 0.1) exit sigmaClip
       
       if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip
       
       
       
       nList = m
       
       testSigma = sigma
       
    end do sigmaClip
    
       
       
    mode = sum(data(1 : nList)) / dble(nList)
    
    variance = 0.0_double
    do k = 1, nList
       diff = data(k) - mode
       variance = variance + diff * diff 
       
    end do
    sigma = sqrt(variance / dble(nlist - 1))
    
    
  end function BackgroundMode11
  


     !********************************************************************
     ! subroutine SetUpIndices
     ! 
     !
     !********************************************************************
     
     subroutine SetUpIndices(il, iu, jl, ju, image, usePixel, pixelValue, pixelIndex)
       integer(int16), intent(in) :: il, iu, jl, ju
       real(single), pointer :: image(:, :)
       logical, pointer :: usePixel(:, :)
       real(single), pointer :: pixelValue(:) 
       integer, pointer :: pixelIndex(:, :)
       
       integer(int32), pointer, dimension(:) :: arrayIndex => null()

       integer :: i, j, k
       type(PixelStruct), pointer :: pixel(:) => null()
 
       k = count(usePixel(il : iu, jl : ju)) ! The number of pixels to be checked
       !write(*, *) "L6029 k=", k  
       arrayIndex => allocateInt32(1, k, "arrayIndex")
       
       if (associated(pixelValue)) deallocate(pixelValue)
       if (associated(pixelIndex)) deallocate(pixelIndex)
       
       pixelValue => allocateSingle(1, k, "pixelValue")
       

       allocate(pixelIndex(il : iu, jl : ju)) 
       allocate(pixel(1 : k))
       
       k = 0 
       
       do j = jl, ju
          do i = il, iu
             if (usePixel(i, j)) then
                k = k + 1
                pixel(k)%x = i
                pixel(k)%y = j
                pixel(k)%counts = image(i, j)
                
             end if
          end do
       end do
       !write(*, *) "L6054 k=", k  
       
       
       ! Index sort the pixel values- pixelValue will be sorted into 
       ! increasing pixel value
       
       call IndexSort(pixel%counts, arrayIndex)
       
       
       do i = 1, k
          j = arrayIndex(i)
          pixelValue(i) = pixel(j)%counts
          pixelIndex(pixel(j)%x, pixel(j)%y) = i
       end do
       

       deallocate(pixel, arrayIndex)
       nullify(pixel, arrayIndex)
       
     end subroutine SetUpIndices

     !********************************************************************
     ! subroutine SetUpIndices
     ! 
     !
     !********************************************************************
     !subroutine SetUpIndices(image, tempmaskImage, iIndex, jIndex, k)
     !  real(single), pointer :: image(:, :)
     !  logical(boolean), pointer :: tempmaskImage(:, :)
     !  integer(int16), pointer, dimension(:) :: iIndex, jIndex
     !  integer(int32), intent(out) :: k
     !  integer(int32), pointer, dimension(:) :: arrayIndex => null(), tempIndex => null()
     !  
     !  integer(int16) :: nx, ny, i, j
     !  integer(int32) ::  i1, j1
     !  
     !
     !  real(single), pointer :: pixelValue(:) => null()
     !  integer(int16) :: lbx, lby
     !  
     !  lbx = lbound(tempmaskImage, 1)
     !  lby = lbound(tempmaskImage, 2)
     !
     !  nx = ubound(tempmaskImage, 1)
     !  ny = ubound(tempmaskImage, 2)
     !  
     !
     !  k = count(.not. tempmaskImage) 
     !  arrayIndex => allocateInt32(1, k, "arrayIndex")
     !  iIndex => allocateInt16(1, k, "iIndex")
     !  jIndex => allocateInt16(1, k, "jIndex")
     !  pixelValue => allocateSingle(1, k, "pixelValue")
     !
       
!
!       !if (MaskImage(ipixel, jpixel)) then
!       
!       !write(*, *) "omdetectdetection setupindices line 2466 tempmaskimage ipixel, jpixel is true"
!       !end if!
!
!
!       !if (binx .eq. 2) then
!       !   detectParams%minNoPixels = 1
!       !else
!       !   detectParams%minNoPixels = 2
!       !end if
!
!       !**********************************************************
!       !
!       !
!       !**********************************************************
!       !minHeight = 3.0  ! was 5.0
!
!       k = 0 
!       do j = lby, ny
!          do i = lbx, nx
!             if ( .not. tempmaskImage(i, j)) then
!                
!                
!                k = k + 1
!                pixelValue(k) = image(i, j)  ! Record the pixel value
!                iIndex(k) = i                 
!                jIndex(k) = j
!                
!             end if
!          end do
!       end do
!       
!
!       !    minHeight = 7.0
!       
!       ! Index sort the pixel values 
!       call IndexSort(pixelValue(1 : k), arrayIndex)
!       
!       
!       deallocate(pixelValue)
!       nullify(pixelValue)!
!
!
!       ! Reorder the arrays into order of decreasing pixel value
!       tempIndex => allocateInt32(1, k, "tempIndex")
!       tempIndex(1 : k) = iIndex(1 : k)
!       
!       do i1 = 1, k
!          j1 = arrayIndex(i1)
!          
!          iIndex(k + 1 - i1) = tempIndex(j1)
!       end do
!       
!       tempIndex(1 : k) = jIndex(1 : k)
!       
!       do i1 = 1, k
!          j1 = arrayIndex(i1)    
!          jIndex(k + 1 - i1) = tempIndex(j1)
!       end do
!       
!
!       deallocate(arrayIndex, tempIndex)
!       nullify(arrayIndex, tempIndex)
!       !
!
!     end subroutine SetUpIndices


end module omlcbuild_mod








